/*
**    Copyright 2014-2020 Ivan Filippov
**    Copyright 2010-2020 Alexey Shutov
**    Copyright 2010-2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef ADC64MAINWINDOW_H
#define ADC64MAINWINDOW_H

#include <optional>
#include <QElapsedTimer>

#include "AdcGuiConfig.h"
#include "adc64-common/Adc64StartupOptions.h"
#include "adc64-common/DeviceMetaInfo.h"

#include "fsm/FsmState.h"
#include "lib-common/ReadoutAppMainWindow.h"

namespace Ui {
    class Adc64MainWindow;
}

class AdcConfigWidget;
class AdcStatus;
class AdcStatusWidget;
class ChannelsSetupForm;
class ConfigKey;
class DeviceTableWidget;
struct CompleteEvent;
class LemoMuxDialog;
class LldpInfo;
class MonitorEventCombiner;
class PreamplifierDialog;
class QDialog;
class QLabel;
class QwtPlotMarker;
class DeviceSettingsDialog;
class DeviceChModel;
class DeviceChTreeView;
class AdcCalibrationDialog;
struct DeserializersControlBuildParams;

class Adc64MainWindow : public ReadoutAppMainWindow
{
    Q_OBJECT

public:
    explicit Adc64MainWindow(const Adc64StartupOptions &startup_options, QWidget *parent = nullptr);
    ~Adc64MainWindow() override;

    DeviceStatusTableViewFancy *getDeviceStatusTableView() override;
    AdcConfigWidget *getAdcConfigWidget();
    LemoMuxDialog *getLemoMuxDialog();
    PreamplifierDialog *getPreamplifierDialog();
    DeviceChModel * getDeviceTreeModel() const;
    DeviceChTreeView * getDeviceChTreeView() const;

Q_SIGNALS:
    void destroyAll() const;

    // proxy signals to adcStatusWidget
    void setAdcStatus(const AdcStatus &);
    // other proxy signals
    void performLocalStart() const;
    void performLocalStop() const;
    void printFsmInput() const;

public slots:
    void setConfig(const ConfigKey &key, const BaseDeviceAppConfig &_config) override;
    void setDeserializersControlBuildParams(DeviceIndex index, const DeserializersControlBuildParams &bp);
    void onMonitorEvent(const CompleteEvent &eventData);
    void onFsmStateChange(const FsmState &state) override;

protected:
    void changeEvent(QEvent *e) override;
    virtual void addMenuToolsConfigureActions() override;
    virtual void saveAppGuiSettings(QSettings &) const override;
    virtual void loadAppGuiSettings(const QSettings &) override;
    virtual void updateControlState() override;

private Q_SLOTS:
    //--Menu bar slots
    //options
    void on_actionBaseline_RMS_check_triggered();
    void on_actionShow_configuration_widget_toggled(bool val);
    void on_actionShow_PCB_ch_toggled(bool val);
    //about
    void on_actionDocs_triggered();
    void on_actionChannel_map_triggered();

    // Other slots
    void channelsSetupFormChanged();
    void deviceUpdated(const DeviceIndex &index, const LldpInfo &dd);
    void updateSparseMarkers();

    void on_pushButtonChannelsSetup_clicked();
    void on_comboBoxOctalADC_currentIndexChanged(int);

    void setShowSparseMarkers(bool val);

private:
    void DrawScope(const CompleteEvent &eventData, bool force = false);
    void setupChanged();
    void updateDisplayedChannels();
    void updateAdcChGroups();

    Ui::Adc64MainWindow * const ui;
    AdcGuiConfig guiConfig;
    QLabel * const evLabel;
    CompleteEvent * const lastEventData;
    QElapsedTimer scopeDisplayTime;
    ChannelsSetupForm * const channelsSetupForm;
    DeviceSettingsDialog * const deviceSettingsDialog;
    PreamplifierDialog * preamplifier_dialog;
    LemoMuxDialog * lemoMuxDialog;
    QDialog * channelMapDialog = nullptr;
    AdcCalibrationDialog * const calibDialog;
    std::optional<QPair<DeviceIndex, QSet<int>>> displayEn; // which channels we should draw
    QVector<QwtPlotMarker *> sparseMarkers;
    bool pcbSwitchingEnabled;
    QMap<DeviceIndex, DeviceMetaInfo> devicesMetaInfo;

    MonitorEventCombiner * const monEvCombiner;
};

#endif // ADC64MAINWINDOW_H
