#include "RegOpModel.h"

#include <QColor>
#include <QFont>

RegOpModel::RegOpModel(QObject *parent)
    : QAbstractTableModel(parent)
{
}

QVariant RegOpModel::data(const QModelIndex &index, int role) const
{
    if(!index.isValid())
        return QVariant();
    switch (role) {
    case Qt::BackgroundRole:
        return backgroundData(index);
    case Qt::CheckStateRole:
        return checkStateData(index);
    case Qt::DisplayRole:
        return displayData(index);
    case Qt::EditRole:
        return editData(index);
    case Qt::FontRole:
        return fontData(index);
    case Qt::TextAlignmentRole:
        return textAlignmentData(index);
    }
    return QVariant();
}

bool RegOpModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
    if (role != Qt::EditRole && role != Qt::CheckStateRole)
        return false;

    if(config.regOpActions.size() <= index.row())
        return false;
    auto &regOp = config.regOpActions[index.row()];

    auto col = static_cast<RegOpColumns>(index.column());

    switch (role) {
    case Qt::CheckStateRole:
    {
        bool checked = value == Qt::Checked;
        if(col == RegOpColumns::COL_READ_WRITE) {
            if(regOp.isWrite == checked)
                return false;
            regOp.isWrite = checked;
        }
        break;
    }
    case Qt::EditRole:
    {
        switch(col) {
        case RegOpColumns::COL_SDB_ID:
        {
            auto list = value.toList();
            if(list.size() != 2)
                return false;
            auto sdbId = list.first().toULongLong();
            auto sdbSeq = list.last().toULongLong();
            if(regOp.sdbId == sdbId && regOp.sdbSeq == sdbSeq)
                return false;
            regOp.sdbId = sdbId;
            regOp.sdbSeq = sdbSeq;
            break;
        }
        case RegOpColumns::COL_REG_OFFSET:
        {
            if(regOp.regOffset == value.toUInt())
                return false;
            regOp.regOffset = value.toUInt();
            break;
        }
        case RegOpColumns::COL_REG_TYPE:
        {
            if(regOp.regType == value.toUInt())
                return false;
            regOp.regType = static_cast<REG_OP_ACTION_TYPE>(value.toUInt());
            break;
        }
        case RegOpColumns::COL_COMMENT:
        {
            if(regOp.comment == value.toString())
                return false;
            regOp.comment = value.toString();
            break;
        }
        case RegOpColumns::COL_WRITE_VAL:
        {
            if(regOp.writeVal == value.toString())
                return false;
            regOp.writeVal = value.toString();
            break;
        }
        default:
            return false;
        }
        break;
    }
    default:
        return false;
    }

    emit dataChanged(index, index);
    emit configChanged(config);
    return true;
}

Qt::ItemFlags RegOpModel::flags(const QModelIndex &index) const
{
    if(!index.isValid())
        return Qt::NoItemFlags;
    auto f = QAbstractItemModel::flags(index);
    auto col = static_cast<RegOpColumns>(index.column());
    switch (col) {
    case RegOpColumns::COL_READ_WRITE:
        f |= Qt::ItemIsUserCheckable | Qt::ItemIsEnabled/* |  Qt::ItemIsEditable*/;
        break;
    case RegOpColumns::COL_SDB_ID:
    case RegOpColumns::COL_REG_TYPE:
    case RegOpColumns::COL_WRITE_VAL:
    case RegOpColumns::COL_REG_OFFSET:
    case RegOpColumns::COL_COMMENT:
        f |= Qt::ItemIsEditable;
        break;
    case RegOpColumns::COL_READ_VAL_DEC:
    case RegOpColumns::COL_READ_VAL_HEX:
    case RegOpColumns::COL_REMOVE:
    default:
        break;
    }
    if(!enable)
        f = f & (~Qt::ItemIsEnabled);
    return f;
}

QVariant RegOpModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if(orientation != Qt::Horizontal)
        return {};
    if(role != Qt::DisplayRole)
        return {};
    switch (static_cast<RegOpColumns>(section)) {
    case RegOpColumns::COL_SDB_ID: return "SDB Id";
    case RegOpColumns::COL_REG_TYPE: return "Reg type";
    case RegOpColumns::COL_REG_OFFSET: return "RegOffset";
    case RegOpColumns::COL_READ_WRITE: return "isWrite";
    case RegOpColumns::COL_WRITE_VAL: return "Write val";
    case RegOpColumns::COL_READ_VAL_DEC: return "Read val (dec)";
    case RegOpColumns::COL_READ_VAL_HEX: return "Read val (hex)";
    case RegOpColumns::COL_REMOVE: return "Remove";
    case RegOpColumns::COL_COMMENT: return "Comment";
    default:
        return {};
    }
}

int RegOpModel::rowCount(const QModelIndex &parent) const
{
    return config.regOpActions.count();
}

void RegOpModel::setEnabled(bool en)
{
    if(enable == en)
        return;

    enable = en;
    auto top = index(0,0);
    auto lastDev = index(rowCount(top)-1, columnCount(top)-1, top);
    auto lastCh = index(rowCount(lastDev)-1, columnCount(lastDev)-1, lastDev);
    emit dataChanged(top, lastCh);
}

void RegOpModel::setConfig(const DreConfig &newConfig)
{
    if(config == newConfig)
        return;
    beginResetModel();
    config = newConfig;
    endResetModel();
}

void RegOpModel::setRegResults(const QVector<quint64> &newResults)
{
    if(actionResults == newResults)
        return;
    actionResults = newResults;

    auto rows = rowCount();
    if(rows) {
        auto start = index(0, static_cast<int>(RegOpColumns::COL_READ_VAL_DEC));
        auto end = index(rows, static_cast<int>(RegOpColumns::COL_READ_VAL_HEX));
        emit dataChanged(start, end);
    }
}

QVariant RegOpModel::backgroundData(const QModelIndex &index) const
{
    auto regOp = getRegOpAction(index.row());
    switch (static_cast<RegOpColumns>(index.column())) {
    case RegOpColumns::COL_WRITE_VAL:
        if(!regOp->getWriteVal()) {
            return QColor(Qt::red);
        }
    default:;
    }
    return {};
}

QVariant RegOpModel::checkStateData(const QModelIndex &index) const
{
    auto regOp = getRegOpAction(index.row());
    switch (static_cast<RegOpColumns>(index.column())) {
    case RegOpColumns::COL_READ_WRITE:
        return regOp->isWrite ? Qt::Checked : Qt::Unchecked;
    default:
        return {};
    }
}

QVariant RegOpModel::displayData(const QModelIndex &index) const
{
    auto regOp = getRegOpAction(index.row());
    auto regRes = getRegResult(index.row());
    switch (static_cast<RegOpColumns>(index.column())) {
    case RegOpColumns::COL_SDB_ID: return QString("0x%1").arg(regOp->sdbId, 0, 16);
    case RegOpColumns::COL_REG_TYPE: return regOp->regType;
    case RegOpColumns::COL_REG_OFFSET: return QString("0x%1").arg(regOp->regOffset, 0, 16);
    case RegOpColumns::COL_READ_WRITE: return regOp->isWrite ? "W" : "R";
    case RegOpColumns::COL_WRITE_VAL:
        return regOp->writeVal;
    case RegOpColumns::COL_READ_VAL_DEC:
        return regRes ? QString("%L1").arg(*regRes) : "-";
    case RegOpColumns::COL_READ_VAL_HEX:
        return regRes ? QString("0x%1").arg(*regRes, 0, 16, QChar('0')) : "-";
    case RegOpColumns::COL_COMMENT: return regOp->comment;
    default:
        break;
    }
    return {};
}

QVariant RegOpModel::editData(const QModelIndex &index) const
{
    auto regOp = getRegOpAction(index.row());
    switch (static_cast<RegOpColumns>(index.column())) {
    case RegOpColumns::COL_SDB_ID: return QList<QVariant>{regOp->sdbId, regOp->sdbSeq};
    case RegOpColumns::COL_REG_TYPE: return regOp->regType;
    case RegOpColumns::COL_REG_OFFSET: return regOp->regOffset;
    case RegOpColumns::COL_READ_WRITE: return regOp->isWrite;
    case RegOpColumns::COL_WRITE_VAL: return regOp->writeVal;
    case RegOpColumns::COL_COMMENT: return regOp->comment;
    default:
        break;
    }
    return {};
}

QVariant RegOpModel::fontData(const QModelIndex &index) const
{
    switch (static_cast<RegOpColumns>(index.column())) {
    case RegOpColumns::COL_READ_VAL_DEC:
    case RegOpColumns::COL_READ_VAL_HEX:
    {
        QFont font;
        font.setFamily("Monospace");
        return font;
    }
    default:
        return {};
    }
}

QVariant RegOpModel::textAlignmentData(const QModelIndex &index) const
{
    switch (static_cast<RegOpColumns>(index.column())) {
    case RegOpColumns::COL_WRITE_VAL:
    case RegOpColumns::COL_READ_VAL_DEC:
    case RegOpColumns::COL_READ_VAL_HEX:
        return Qt::AlignRight;
    default:;
    }

    return {};
}

std::optional<RegOpAction> RegOpModel::getRegOpAction(int i) const
{
    if(i < config.regOpActions.size())
        return config.regOpActions[i];
    return {};
}

std::optional<quint64> RegOpModel::getRegResult(int i) const
{
    if(i < actionResults.size())
        return actionResults[i];
    return {};
}
