#include "AssignIndexesDialog.h"
#include "base/AbstractConfigConverter.h"
#include "ui_AssignIndexesDialog.h"
#include "mregdev/qtmregdevice.h"
#include <QPushButton>
#include <QComboBox>
#include <QHostAddress>

#include "base/DeviceId.h"

namespace {
const QString ITEM_NONE = "none";
}

AssignIndexesDialog::AssignIndexesDialog(DeviceIndexEnabledMap d, QMap<quint32, QString> _hostMap,
                                         QWidget *parent) :
    QDialog(parent),
    ui(new Ui::AssignIndexesDialog),
    devices(d),
    hostMap(_hostMap)
{
    AbstractConfigConverter<DeviceIndex>::registerTypes();

    ui->setupUi(this);
    setWindowTitle("Assign indexes");
    connect(ui->buttonBox->button(QDialogButtonBox::Reset), SIGNAL(clicked()),
            SLOT(setDefaultIndex()));
}

AssignIndexesDialog::~AssignIndexesDialog()
{
    delete ui;
}

void AssignIndexesDialog::init(HrbAppConfig ss, bool isHrbType)
{
    ui->spinBox->setValue(0);
    config = ss;
    hrbType = isHrbType;
    setDevices();
}

void AssignIndexesDialog::setDevices()
{
    int devNumber = 0;
    labelsMap.clear();
    for(const auto &die : devices) {
        if(skipDevice(die))
            continue;
        const auto &index = die.index;
        const auto &lldp = lldpCache.value(index);
        labelsMap[index] = QString("%1 (%2)")
                .arg(index.getSerialStr())
                .arg(hostMap.value(lldp.common.ip_addr, QHostAddress(lldp.common.ip_addr).toString()));
        ++devNumber;
    }
    if(devNumber) {
        // if there at lease one enabled device
         //todo
//        if(hrbType)
//            for(const auto &cfg : config.hrb.values())
//                devNumber = qMax(devNumber, cfg.cfgIndex+1);
//        else
//            for(const auto &cfg : config.ttb.values())
//                devNumber = qMax(devNumber, cfg.cfgIndex+1);
    }

    ui->spinBox->setValue(devNumber);
    setDefaultIndex();
}

HrbAppConfig AssignIndexesDialog::getSystemSetup()
{
    if(hrbType) {
        config.hrb = getHrbSetups();
    } else {
        config.ttb = getTtbSetups();
    }

    return config;
}

void AssignIndexesDialog::on_spinBox_valueChanged(int newMax)
{
    int oldMax = boxes.size();
    if(oldMax==newMax)
        return;

    for(int i=oldMax;i<newMax;++i) {
        QComboBox *box = new QComboBox(this);
        boxes.push_back(box);
        box->insertItem(0, ITEM_NONE, 0);
        for(auto i : labelsMap.keys())
            box->insertItem(box->count(), labelsMap[i], QVariant::fromValue(i));

        connect(box, SIGNAL(currentIndexChanged(int)), SLOT(comboBoxChanged(int)));
        ui->formLayout->addRow(QString(hrbType?"HRB-%1" : "TTB-%1").arg(i),box);
    }

    for(int i=newMax;i<oldMax;++i) {
        QComboBox *box = boxes.takeLast();
        ui->formLayout->removeRow(box);
    }
    Q_ASSERT(boxes.size() == newMax);

    refreshLabels();
    isValid();
}

void AssignIndexesDialog::comboBoxChanged(int)
{
    QComboBox *s = qobject_cast<QComboBox*>(sender());
    Q_ASSERT(s);
    int cfgId = boxes.indexOf(s);
    quint64 serial = s->currentData().toULongLong();
    Q_ASSERT(cfgId != -1);
    // Remove from others cfgId (exclude dublicate)
    if(serial) {
        foreach (QComboBox *box, boxes) {
            if(box!=s && box->currentData() == serial){
                box->setCurrentIndex(0);
            }
        }
    }

    refreshLabels();
    isValid();
}

void AssignIndexesDialog::setDefaultIndex()
{
    foreach (QComboBox *b, boxes)
        b->setCurrentIndex(0);

    for(auto die : devices) {
        if(skipDevice(die))
            continue;
        const auto &index = die.index;
        int cfgId=-1;
        if(hrbType) {
            for(auto s : config.hrb) {
                if(s.index == index) {
//                    cfgId = s.cfgIndex; //todo
                    break;
                }
            }
        } else {
            for(auto s : config.ttb) {
                if(s.index == index) {
//                    cfgId = s.cfgIndex; //todo
                    break;
                }
            }
        }

        if(cfgId!=-1) {
            if(boxes.size()>cfgId) {
                QComboBox *box = boxes[cfgId];
                int itemIndex = box->findData(QVariant::fromValue(index));
                if(itemIndex>=0)
                    box->setCurrentIndex(itemIndex);
                else
                    Q_ASSERT(itemIndex>=0);
            } else
                Q_ASSERT(boxes.size()>cfgId);
        } else
            qDebug()<<"Can't detect cfg-index to HRB "<<index.getSerialStr();
    }

    isValid();
}

bool AssignIndexesDialog::isValid()
{
    bool isOk = true;
    QMap<DeviceIndex, int> devMap = getDevMap();
    for(const auto &die : devices) {
        if(skipDevice(die))
            continue;
        const auto &index = die.index;
        bool found=devMap.contains(index);
        if(found){
            devMap.remove(index);
        } else {
            isOk = false;
            break;
        }
    }
    isOk &= devMap.isEmpty();

    ui->buttonBox->button(QDialogButtonBox::Ok)->setEnabled(isOk);
    return isOk;
}

bool AssignIndexesDialog::skipDevice(const DeviceIndexEnabled &die) const
{
    return !die.enabled ||
            (die.index.getDevId() != (hrbType ? DEVICE_ID_HRB6ASD : DEVICE_ID_TTB9V));
}

QMap<DeviceIndex, int> AssignIndexesDialog::getDevMap() const
{
    QMap<DeviceIndex, int> map;
    for(QComboBox *box : boxes) {
        const auto &index = box->currentData().value<DeviceIndex>();
        if(!index.isNull())
            map[index] = boxes.indexOf(box);
    }
    return map;
}

void AssignIndexesDialog::refreshLabels() const
{
    // Get used serials
    QSet<quint64> usedSerials;
    foreach (QComboBox *box, boxes) {
        quint64 s = box->currentData().toULongLong();
        if(s)
            usedSerials << s;
    }

    // Update labels
    for(QComboBox *box : boxes) {
        const auto &curIndex = box->currentData().value<DeviceIndex>();
        for(int i=0; i<box->count(); ++i) {
            const auto &index = box->itemData(i).value<DeviceIndex>();
            QString curTxt = box->itemText(i);
            QString txt = labelsMap[index];

            if(curIndex!=index && usedSerials.contains(i))
                txt.append("*");
            if(txt != curTxt)
                box->setItemText(i, txt);
        }
    }
}

QMap<DeviceIndex, ModularDeviceConfig> AssignIndexesDialog::getHrbSetups()
{
    QMap<DeviceIndex, ModularDeviceConfig> hrbMap;
    QVector<ModularDeviceConfig> disabledHrb;
    int maxCfgIndex=-1;
    for(const auto &die: devices) {
        const auto &i = die.index;
        if(i.getDevId() != DeviceId::DEVICE_ID_HRB6ASD)
            continue;
        int index = -1;
        for(QComboBox *box : boxes) {
            if(box->currentData().value<DeviceIndex>() == i) {
                index = boxes.indexOf(box);
                break;
            }
        }
        Q_ASSERT(die.enabled ? index>=0 : index==-1);

        bool found=false;
        ModularDeviceConfig hrbCfg;
        for(const auto &cfg : config.hrb) {
            if(cfg.index == die.index) {
                hrbCfg = cfg;
                found = true;
                break;
            }
        }
        if(!found) {
            hrbCfg = config.defaultHrb;
            hrbCfg.index = die.index;
        }

        hrbCfg.enabled = die.enabled;
        if(die.enabled) {
            maxCfgIndex = qMax(maxCfgIndex, index);
            hrbMap[hrbCfg.index] = hrbCfg;
        } else
            disabledHrb.push_back(hrbCfg);
    }

    // fill disabled
    for(int i=0; i<disabledHrb.size(); ++i) {
        auto &s = disabledHrb[i];
        int cfgIndex = -1;
        //todo
//        for(const HrbConfig &prevSetup : config.hrb) {
//            if(prevSetup.index == s.index)
//                cfgIndex = prevSetup.cfgIndex; // todo
//        }
//        if(cfgIndex==-1 || hrbMap.contains(cfgIndex))
//            cfgIndex = ++maxCfgIndex;
//        else
//            maxCfgIndex = qMax(maxCfgIndex, cfgIndex);

//        s.cfgIndex = cfgIndex;
//        hrbMap[s.cfgIndex] = s;
    }

    return hrbMap;
}

QMap<DeviceIndex, ModularDeviceConfig> AssignIndexesDialog::getTtbSetups()
{
    QMap<DeviceIndex, ModularDeviceConfig> ttbMap;
    QVector<ModularDeviceConfig> disabledTtb;
    int maxCfgIndex=-1;
    for(const auto &die : devices) {
        if(die.index.getDevId() != DEVICE_ID_TTB9V)
            continue;
        int index = -1;
        for(QComboBox *box : boxes) {
            if(box->currentData().value<DeviceIndex>() == die.index) {
                index = boxes.indexOf(box);
                break;
            }
        }
        Q_ASSERT(die.enabled ? index>=0 : index==-1);

        ModularDeviceConfig setup;
        bool found=false;
        for(const auto &s : config.ttb) {
            if(s.index == die.index){
                setup = s;
                found = true;
                break;
            }
        }
        if(!found)
            setup.index = die.index;
        setup.enabled = die.enabled;
        //todo
//        setup.host.clear(); // should be through MainWindow::updateDeviceInfo()
//        setup.cfgIndex = index;

        if(die.enabled)
            ttbMap[setup.index] = setup;
        else
            disabledTtb.push_back(setup);
    }

    // fill disabled
    for(int i=0; i<disabledTtb.size(); ++i) {
        auto &s = disabledTtb[i];
        int cfgIndex = -1;
        //todo
//        for(const TtbConfig &prevSetup : config.ttb) {
//            if(prevSetup.index == s.index)
//                cfgIndex = prevSetup.cfgIndex;
//        }
//        if(cfgIndex==-1 || ttbMap.contains(cfgIndex))
//            cfgIndex = ++maxCfgIndex;
//        else
//            maxCfgIndex = qMax(maxCfgIndex, cfgIndex);

//        s.cfgIndex = cfgIndex;
//        ttbMap[s.cfgIndex] = s;
    }
    return ttbMap;
}

void AssignIndexesDialog::on_pushButtonAutoFill_clicked()
{
    QVector<DeviceIndexEnabled> unsetDevices;
    int devCnt=0;

    for(const auto &die : devices) {
        if(skipDevice(die))
            continue;
        bool found=false;
        for(QComboBox *box : boxes) {
            if(box->currentData().value<DeviceIndex>()==die.index) {
                found = true;
                break;
            }
        }
        if(!found)
            unsetDevices.push_back(die);
        ++devCnt;
    }

    if(boxes.size() < devCnt)
        ui->spinBox->setValue(devCnt);

    for(QComboBox *box : boxes) {
        if(box->currentData().toULongLong()==0){
            auto die = unsetDevices.takeAt(0);
            int itemIndex = box->findData(QVariant::fromValue(die.index));
            if(itemIndex!=-1)
                box->setCurrentIndex(itemIndex);
            else
                Q_ASSERT(itemIndex!=-1);
        }
        if(unsetDevices.isEmpty())
            break;
    }
    Q_ASSERT(unsetDevices.isEmpty());

    isValid();
}
