/*
**    Copyright 2013-2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "HrbDevice.h"

#include <cmath>
#include <QDebug>
#include <QFile>
#include <QTextStream>

#include "HRBRegs.h"
#include "LegacyDeviceModule.h"
#include "lib-common/logDefinesPrefix.h"
#include "regio/regop.h"

namespace {
const int CH_COUNT = ASD8_COUNT * 16;
const fw_version_t REV_CNT_LOCK = fw_version_t{1,0,22709};
const fw_version_t REV_MSTREAM_22 = fw_version_t{1,0,24000};
}


HrbDevice::HrbDevice(const DeviceIndex &i, QObject *parent)
    : ModularMregDevice(i, status, parent)
    , status(i)
    , legacyDevice(createModule<LegacyDeviceModule>())
{}

void HrbDevice::writeThresholds(const QVector<int> &v)
{
/*
    RegOpVector op;
    for (int i=0; i<v.size(); i++) {
        int thr = v[i];
        qDebug() << "set asd8 board" << i << "threshold" << thr;
        op += HRBDeviceOps::opDacWrite(i, thr);
    }
    regOpExec(op);
*/
//    for (int j=0; j<10; j++) {
//    r.RegWrite16(REG_THR_DAC, 0x803F); Unbuffered, reference Vcc
    if(!isConnected())
        return;
    RegOpVector r;
    r.RegWrite16(REG_THR_DAC, 0x800C); // Buffered, reference external
    for (int i=0; i<ASD8_COUNT; i++) {
        int data = v[i];
        if (data < 0) data = 0;
        if (data > 4095) data = 4095;
        int ch = 6 - i; // DAC channels: 6,5,4,3,2,1
        r.RegWrite16(REG_THR_DAC, (((ch&0x7)<<12) | (data&0xFFF)));
    }
    regOpExec(r);
//    }
}

bool HrbDevice::writeConfig()
{
    bool ok = ModularMregDevice::writeConfig();
    if (!isConnected())
        return false;

    if(!isConfigCacheValid() || configCache != config) {

        QVector<int> thrRaw;
        bool asdEqual=true;
        for (int i=0; i<ASD8_COUNT; i++){
            thrRaw.push_back(thresholdConverter(config.hrb.thr[i]));
            if (configCache.hrb.thr[i] != config.hrb.thr[0]) asdEqual = false;
        }
        if(asdEqual){
            logInfo << QString("set asd8 threshold %1 (%2 V)").arg(thrRaw[0]).arg(config.hrb.thr[0]);
        } else {
            for (int i=0; i<ASD8_COUNT; i++)
                logInfo << QString("set asd8 board %1 threshold %2 (%3 V)").arg(i).arg(thrRaw[i]).arg(config.hrb.thr[i]);
        }
        writeThresholds(thrRaw);

        RegOpVector r;
        // Setup Pulser
        for (int i=0; i<ASD8_COUNT; i++)
            r.RegWrite32(REG_PULSER_PER_0 + 2*i, 125000); // period in clocks, 1 kHz
        r.RegWrite16(REG_PULSER_PER_EN,  0x0000); // periodic pulser
        r.RegWrite16(REG_PULSER_RAND_EN, 0x0000); // random pulser
        r.RegWrite16(REG_PULSER_EXT_EN,  0x003F); // external pulser (TTL0)

        // Setup trigger
        const auto &trigSetup = config.hrb.trigSetup;
        quint16 trig_src = 0;
        if(config.isRun()) {
            if(trigSetup.lvdsSpill) trig_src |= TRIG_SRC_BIT_LVDS_SPILL_EN;
            if(trigSetup.lvdsTrigger) trig_src |= TRIG_SRC_BIT_LVDS_TRIG_EN;
            if(trigSetup.ttlSpill) trig_src |= TRIG_SRC_BIT_TTL_SPILL_EN;
            if(trigSetup.ttlTrigger) trig_src |= TRIG_SRC_BIT_TTL_TRIG_EN;
            if(trigSetup.internalSpill) trig_src |= TRIG_SRC_BIT_TIMER_SPILL_EN;
            if(trigSetup.internalTrigger) trig_src |= TRIG_SRC_BIT_TIMER_TRIG_EN;
        }

        r.RegWrite16(REG_TIMER_TRIG_PERIOD, trigSetup.interval);
        r.RegWrite16(REG_TRIG_MATCH, config.hrb.window);
        r.RegWrite16(REG_TRIG_LATENCY, config.hrb.latency);
        r.RegWrite16(REG32_MIN_HIT_DURATION, config.hrb.minHitDuration);
        r.RegWrite16(REG_TRIG_SRC, trig_src);
        r.RegWrite16(REG_SPILL_CTRL, trigSetup.softSpill ? (1<<15) : 0);

        bool periodicTimestampReset = false;
        quint16 reg_hrb_core = 0;
        if (periodicTimestampReset) reg_hrb_core |= 1; // reset timestamp every ~1.4s
        //        if (setup.streamGlobal || setup.stream)
        reg_hrb_core |= 2; // Enable stream readout
        r.RegWrite16(REG_CORE0, reg_hrb_core);
        bool ok = true;
        if (!legacyDevice->regOpExecRebased(r))
            ok = false;
        configCache.hrb.trigSetup = trigSetup;
        if (ok) {
            logInfo << QString("set trig_src=%2; latency=%3(%4ns); window=%5(%6ns); minHitDuration=%7(%8ns)")
                       .arg(trig_src, 4, 16, QChar('0'))
                       .arg(config.hrb.latency).arg(config.hrb.latency*8)
                       .arg(config.hrb.window).arg(config.hrb.window*8)
                       .arg(config.hrb.minHitDuration).arg(config.hrb.minHitDuration*2);
        } else {
            logCritical << getIdent() << "Failed to write setup";
        }
    }
    clearStatistics();

    //    r.RegWrite16(REG_SPILL_CTRL, 0); // spill off
    //    r.RegWrite16(REG_SPILL_CTRL, 0x8000); // spill on
    return true;
}

void HrbDevice::softReset()
{
    qWarning() << __PRETTY_FUNCTION__<<"todo";
}

void HrbDevice::clearStatistics()
{
    RegOpVector r;
    r.RegWrite16(REG_CONTROL, 1);
    r.RegWrite16(REG_CONTROL, 0);
    if (!regOpExec(r))
        logDebug << "clearStatistics";
    else
        logCritical << "failed to clear statistic";
}

bool HrbDevice::readStatistics()
{
    if(!isConnected()) return false;

//    QVector<quint32> counters = readHitCounts();
//    readHit2DHistogram();
//    updateCountersFile(counters);
    bool ok = readTTCLinkErrors();
    return ok;
}

bool HrbDevice::readHrbStatus()
{
//    qDebug() << "\nStatus for board" << hex << dev.getDeviceSerialNumber();
    // Read data
    RegOpVector op;
    if (getDeviceFwVersion() >= REV_CNT_LOCK) {
        op.RegWrite16(REG_CONTROL, 0x2);
    }
    op.RegRead16(REG_INSTATE_0, &status.inState[0]);
    op.RegRead16(REG_INSTATE_1, &status.inState[1]);
    op.RegRead16(REG_INSTATE_2, &status.inState[2]);
    op.RegRead16(REG_INSTATE_3, &status.inState[3]);
    op.RegRead16(REG_INSTATE_4, &status.inState[4]);
    op.RegRead16(REG_INSTATE_5, &status.inState[5]);
    op.RegRead16(REG_INSTATE_6, &status.inState[6]);
    op.RegRead16(REG_INSTATE_7, &status.inState[7]);

    op.RegRead16(REG_RSTATUS, &status.rStatus);
    op.RegRead16(REG_TTC_STATE, &status.rttcStatus);
    op.RegRead32(REG32_TRIG_PASS, &status.statistic.trigPass);
    op.RegRead32(REG32_TRIG_BLOCK, &status.statistic.trigBlock);
    op.RegRead32(REG32_EVENT_NR, &status.evNum);
    op.RegRead32(REG32_SPILL_NR, &status.spillNum);
    op.RegRead64(REG64_SYSTEM_TS, &status.systemTs);
    op.RegRead32(REG32_BUSY_TIMEOUT_CNT, &status.statistic.busyTimeoutCnt);
    op.RegRead64(REG64_BUSY_TIMER, &status.statistic.busyTimer);
    op.RegRead64(REG64_IDLE_TIMER, &status.statistic.idleTimer);
    op.RegRead64(REG64_BUSY_MSTR_TIMER, &status.statistic.busyMStreamTimer);
    op.RegRead64(REG64_IDLE_MSTR_TIMER, &status.statistic.idleMStreamTimer);
    op.RegRead64(REG64_GLOB_EVENT_NR, &status.globEvNum);
    op.RegRead16(REG16_FIFO_SIZE, &status.fifoSize);

    if(getDeviceFwVersion() >= REV_CNT_LOCK)
        op.RegWrite16(REG_CONTROL, 0);

//    op.RegRead16(REG_CORE0);
    if (!regOpExec(op)) {
        if(isConnected())
            logWarning << "readStatus failed";
        return false;
    }

//    const quint64 TS_FREQ = 125e6;
//    qInfo() << __PRETTY_FUNCTION__
//            << status.systemTs
//            << status.systemTs / TS_FREQ;

/*
    const quint64 TS_FREQ = 125000000;
//    double system_ts_s = system_ts - (system_ts / TS_FREQ) * TS_FREQ;
//    double system_ts_ns = system_ts - system_ts_s;
    double system_ts_real = (double)status.systemTs / (double)TS_FREQ;
    quint64 system_ts_s = (quint64)system_ts_real;
    quint64 system_ts_us_part = (quint64)(1000000 * (system_ts_real - system_ts_s));
    QDateTime date = QDateTime::fromTime_t(system_ts_s);
//    quint32 spill_nr         = 0;
    qDebug() << QString("RSTATUS: clk_sel %1, clk_ok %2")
            .arg((status.rStatus >> 4) & 0xF, 2, 16, QChar('0'))
            .arg((status.rStatus >> 0) & 0xF, 2, 16, QChar('0'))
            ;
    qDebug() << QString("RTTCSTATE: TTL %1, SPILL %2, TRIG %3, BUSY %4, TTC_CLK_OK %5, TRIG_SYNC_OK %6, RESET_SYNC_OK %7")
                .arg((status.rttcStatus >> 0) & 0xF, 2, 16, QChar('0'))
                .arg((status.rttcStatus >> 4) & 0x1)
                .arg((status.rttcStatus >> 5) & 0x1)
                .arg((status.rttcStatus >> 6) & 0x1)
                .arg((status.rttcStatus >> 8) & 0x1)
                .arg((status.rttcStatus >> 9) & 0x1)
                .arg((status.rttcStatus >> 10) & 0x1)
            ;
    qDebug() << QString("TRIG COUNTS: pass %1, block %2 (%3%)")
            .arg(status.trigPass, 10)
            .arg(status.trigBlock, 10)
            .arg(100.0*status.trigPass/(status.trigPass+status.trigBlock), 0, 'f', 2)
            ;
    qDebug() << QString("SPILL # %1, EVENT # %2, TS %3.%4")
                .arg(status.spillNum, 10)
                .arg(status.evNum, 10)
                .arg(date.toString("yyyy-MM-dd hh:mm:ss"))
                .arg(system_ts_us_part,6, 10, QChar('0'))
            ;
*/
    return true;
}

bool HrbDevice::readStatus()
{
    if (!ModularMregDevice::readStatus())
        return false;

    if(!isConnected()) return false;
    if (!readHrbStatus()) return false;
    if (!readStatistics()) return false;
    emit statusUpdated(status);
    return true;
}

QVector<quint32> HrbDevice::readCnt32(quint32 base, int count)
{
    QVector<quint32> cnt(count);
    RegOpVector op;
    for(int i=0; i<count; i++) {
        op.RegRead32(base + 2*i, &cnt[i]);
    }
    if (!regOpExec(op)) {
        logWarning << QString("readCnt32 failed base:0x%1; count:%2")
                      .arg(base,6,16,QChar('0')).arg(count);
        return {};
    }
    return cnt;
}

QVector<quint32> HrbDevice::readHitCounts()
{
    QVector<quint32> hits_dir = readCnt32(REG_DIRECT_HIT_COUNTERS_BASE, CH_COUNT);
    QVector<quint32> hits_win = readCnt32(REG_WIN_HIT_COUNTERS_BASE, CH_COUNT);
    if (false) {
    int n = std::min(hits_dir.size(), hits_win.size());
    for(int i=0; i<n; i++) {
        if (hits_dir[i] > 0)
            logDebug << QString("CNT[%1]: %2  %3")
            .arg(i, 3)
            .arg(hits_dir.at(i), 10)
            .arg(hits_win.at(i), 10);
    }
    }
    return hits_win;
}

int HrbDevice::thresholdConverter(double val)
{
    int newVal = (val+2.5)/5*4096;
    if(newVal>4095) newVal = 4095;
    if(newVal<0) newVal = 0;
    return newVal;
}

void HrbDevice::readHit2DHistogram()
{
    if(!isConnected())
        return;
    const int xSize = (int) pow(2, 10);
    const int ySize = (int) pow(2, 5);
    const uint nBins = xSize * ySize;
    const quint32 MEM_HIST2D_BASE = 0;
    std::vector<quint32> ram;
    if (!memReadBlk(MEM_HIST2D_BASE, nBins, &ram) || (ram.size() != nBins)) {
        logWarning << "readHit2DHistogram failed";
        return;
    }
    QFile f("hits2d.txt");
    f.open(QIODevice::WriteOnly);
    QTextStream ts(&f);
    for (int x=0; x<xSize; x++) {
        for (int y=0; y<ySize; y++) {
            ts << ram[x*ySize + y] << " ";
        }
        ts << "\n";
    }
}

bool HrbDevice::readTTCLinkStatus()
{
    /*
    if(!isConnected())
        return false;
    RegOpVector op;
    op.RegRead16(REG_TTC_CLK_ERR);
    const auto &rxOpt = regOpExec(op);
    if (!rxOpt.has_value()) {
        logWarning << "readTTCLinkStatus failed";
        return false;
    }
    const auto &rx = rxOpt.value();
    quint32 r_ttc_status = rx.at(0).data;
    if (r_ttc_status != 0) {
        logInfo << QString("TTC status: %1").arg(r_ttc_status, 8, 16, QChar('0'));
        if (!(r_ttc_status & 1))
            logWarning << "TTC clock error";
        if (!(r_ttc_status & 2))
            logWarning << "TTC link0 (trig) error";
        if (!(r_ttc_status & 4))
            logWarning << "TTC link1 (ts) error";
//        r.RegWrite16(REG_TTC_CONTROL, 0x1);
//        r.RegWrite16(REG_TTC_CONTROL, 0x0);
    }
*/
    return true;
}

bool HrbDevice::readTTCLinkErrors()
{
    std::array<quint32, REG32_TTC_STATISTICS_NUM>  statCounters;
    RegOpVector op;
    for (int i=0; i<REG32_TTC_STATISTICS_NUM; i++) {
        op.RegRead32(REG32_TTC_STATISTICS_BASE + 2*i, &statCounters[i]);
    }
    if (!regOpExec(op))
        return false;

    HrbStatistic &curStat = status.statistic;
    HrbStatistic newStat = curStat;
    newStat.tsZeroMismatch = statCounters.at(0);
    newStat.ttcTrigSyncLoss = statCounters.at(1);
    newStat.ttcTrigCodeError = statCounters.at(2);
    newStat.ttcResetSyncLoss = statCounters.at(3);
    newStat.ttcResetCodeError = statCounters.at(4);
//    bool link_any_errors = false;
//    QVector<quint32> cnt;
//    for (int i=0; i<REG32_TTC_STATISTICS_NUM; i++) {
//        quint32 value = (quint32)(rx.at(2*i).data) | (((quint32)(rx.at(2*i+1).data)) << 16);
//        cnt.push_back(value);
//        if (value) link_any_errors = true;
//    }

    if(newStat.tsZeroMismatch != curStat.tsZeroMismatch)
        logWarning << QString("Timestamp mismatch count %1").arg(newStat.tsZeroMismatch);
    if (newStat.ttcTrigSyncLoss != curStat.ttcTrigSyncLoss || newStat.ttcTrigCodeError != curStat.ttcTrigCodeError)
        logWarning << QString("Link 0 (TRIG):  sync errors %1, code errors %2")
                      .arg(newStat.ttcTrigSyncLoss).arg(newStat.ttcTrigCodeError);
    if (newStat.ttcResetSyncLoss != curStat.ttcResetSyncLoss || newStat.ttcResetCodeError != curStat.ttcResetCodeError)
        logWarning << QString("Link 0 (RESET):  sync errors %1, code errors %2")
                      .arg(newStat.ttcResetSyncLoss).arg(newStat.ttcResetCodeError);
    curStat = newStat;
    return true;
}

void HrbDevice::updateCountersFile(QVector<quint32> counters)
{
    QFile dataFile(QString("%1_counters.dat").arg(getDeviceSerialNumber()));
    dataFile.open(QIODevice::WriteOnly | QIODevice::Text );
    QTextStream out(&dataFile);
    out << getAddressStr() << '\n';
    out << status.statistic.trigPass << '\n';
    for(int ch=0; ch < counters.size(); ch++){
        out << ch <<' '<<counters[ch] << '\n';
    }
    dataFile.close();
}

void HrbDevice::on_deviceConnected()
{
    ModularMregDevice::on_deviceConnected();
    status.fwStr = getFirmwareVersionStr();
    status.hostStr = getAddressStr();
    emit hrbMStreamVer(getDeviceIndex(),
                       getDeviceFwVersion() >= REV_MSTREAM_22);
    emit connectedStateChanged(getDeviceIndex(), true);
    logInfo << QString("Device online (%1) fw=%2")
               .arg(getAddressStr()).arg(getFirmwareVersionStr());
}

void HrbDevice::on_deviceDisconnected()
{
    ModularMregDevice::on_deviceDisconnected();

    emit connectedStateChanged(getDeviceIndex(), false);
    logWarning << QString("Device HRB offline");
}

void HrbDevice::evNumRequest(int checkIndex)
{
    if(!isConnected()) return;
    quint64 evNum;
    RegOpVector r;
    bool canLock = getDeviceFwVersion() >= REV_CNT_LOCK;
    if(canLock)
        r.RegWrite16(REG_CONTROL, 0x2);
    r.RegRead64(REG64_GLOB_EVENT_NR, &evNum);
    if(canLock)
        r.RegWrite16(REG_CONTROL, 0);
    if (!regOpExec(r))
        return;

    emit evNumResponse(checkIndex, getDeviceIndex(), evNum);
}
