#include "MscDevice.h"

#include <QTimer>

#include "Ad5328/Ad5328Module.h"
#include "Ddr3StreamFifo/Ddr3StreamFifoModule.h"
#include "ModularDeviceConfig.h"
#include "MscCycleCounters/MscCycleCountersModule.h"
#include "MscDacCalibrator.h"
#include "MscDacConfigBuilder.h"
#include "MscDacConverter.h"
#include "MscHist/MscHistModule.h"
#include "MscInput/MscInputModule.h"
#include "MscRegs.h"
#include "MscStream/MscStreamModule.h"
#include "cal/CalResult.h"
#include "data/MscHistData.h"
#include "device-utils/DacConverter.h"

namespace {
Q_DECLARE_LOGGING_CATEGORY(lc_msc)
Q_LOGGING_CATEGORY(lc_msc, "afi.msc")
}

static const int HIST_READOUT_PERIOD_PER_CH = 1000;

MscDevice::MscDevice(const DeviceIndex &index, QObject *parent):
      ModularMregDevice(index, status, parent)
    , status(index)
    , histReadTimer(new QTimer(this))
    , inputMod(createModule<MscInputModule>())
    , histMod(createModule<MscHistModule>())
    , streamMod(createModule<MscStreamModule>())
    , cycleMod(createModule<MscCycleCountersModule>())
    , ddrMod(createModule<Ddr3StreamFifoModule>())
    , dacCalibrator(new MscDacCalibrator(MscDacConfigBuilder::instance()))
{
    histReadTimer->setInterval(HIST_READOUT_PERIOD_PER_CH);
    histReadTimer->setSingleShot(true);
    connect(histReadTimer, &QTimer::timeout, this, &MscDevice::readHists);
    connect(this, &MscDevice::calResultChanged,
            this, &MscDevice::setCalResult); // FIXME: use ConfigDispatcher
    histReadTimer->start();
}

MscDevice::~MscDevice()
{}

bool MscDevice::readStatus()
{
    if (!ModularMregDevice::readStatus())
        return false;
    if (!readMscStatus())
        return false;
    return true;
}

bool MscDevice::readMscStatus()
{
    status.inputStatus.reset();
    if (inputMod) {
        MscInputStatus inputStatus;
        if (!inputMod->readStatus(&inputStatus))
            return {};
        status.inputStatus = inputStatus;
    }

    status.cycleStatus.reset();
    if (cycleMod) {
        MscCycleStatus cycleStatus;
        if (!cycleMod->readStatus(&cycleStatus))
            return {};
        status.cycleStatus = cycleStatus;
    }

    status.histStatus.reset();
    if (histMod) {
        MscHistStatus histStatus;
        if (!histMod->readStatus(&histStatus))
            return {};
        status.histStatus = histStatus;
    }

    status.ddrStatus.reset();
    if (ddrMod) {
        Ddr3StreamFifoStatus ddrStatus;
        if (!ddrMod->readStatus(&ddrStatus))
            return {};
        status.ddrStatus = ddrStatus;
    }

    emit statusUpdated(status);
    if (status.cycleStatus.has_value())
        emit cycleStatusUpdated(getDeviceIndex(), status.cycleStatus.value());
    return true;
}

void MscDevice::enforceConfigConstraints(ModularDeviceConfig *config)
{
    int nch = buildParams.mscInput ? buildParams.mscInput->channels : 16;

    auto dacModule = getFirstModule<Ad5328Module>();
    int index = 0;
    if (dacModule) {
        const auto &chThr = config->mscCore.chThreshold;
        const auto &chInv = config->mscCore.chPolarity;
        // apply Threshold DAC calibration
        QMap<int, int> dacValues;
        for (int ch=0; ch<nch; ch++) {
            auto thr = chThr[ch];
            int sign = chInv[ch] ? -1 : 1;
            double offset = calResult.dacCal_mV[ch];
            dacValues[ch] = MscDacConverter::instance().dacValue(thr * sign + offset);
        }
        // auto index = dacModule->moduleIndex();
        config->ad5328Configs[index++] = MscDacConfigBuilder::instance().build(dacValues);
    }
}

bool MscDevice::writeConfig()
{
    bool ok = ModularMregDevice::writeConfig();

    if(!isConfigCacheValid() || !(config.mscCore.mstream == configCache.mscCore.mstream)) {
        ok &= streamMod->writeConfig(config.mscCore.mstream);
    }

    if(!isConfigCacheValid() || !(config.mscCore.hist == configCache.mscCore.hist)) {
        ok &= histMod->writeConfig(config.mscCore.hist);
    }

    if(!isConfigCacheValid() || !(config.mscCore.input == configCache.mscCore.input)) {
        ok &= inputMod->writeConfig(config.mscCore.input);
    }

    return ok;
}

bool MscDevice::updateInputAndHistStatus()
{
    if (cycleMod)
        status.cycleStatus = cycleMod->readStatus();
    else
        status.cycleStatus.reset();

    if (histMod)
        status.histStatus = histMod->readStatus();
    else
        status.histStatus.reset();
    return status.cycleStatus.has_value() && status.histStatus.has_value();
}

void MscDevice::readHists()
{
    if (!isConnected())
        goto skip;
    if (!buildParams.mscInput.has_value() || !buildParams.mscHist.has_value())
        goto skip;
    if (!updateInputAndHistStatus())
        goto skip;

    {
        MscDevHists devHist;
        for (uint16_t hId = 0; hId < buildParams.mscInput->hists; ++hId) {
            if (!config.mscCore.hist[hId].enable)
                continue;

            const auto tHist = readTimeHist(hId);
            if (!tHist.has_value())
                continue;
            const auto pHist = readPlotHist(hId);
            if (!pHist.has_value())
                continue;
            devHist.hists[hId] = MscHist{.plot = pHist.value(), .time = tHist.value()};
        }
        emit histUpdated(getDeviceIndex(), devHist);
    }

skip:
    histReadTimer->start();
}

std::optional<MscHistData> MscDevice::readTimeHist(quint16 hId)
{
    if (!buildParams.mscHist.has_value())
        return {};
    const auto &mscBuildParams = buildParams.mscHist.value();
    auto hist_step = static_cast<uint16_t>(1<<(mscBuildParams.ramBitsHistTimeNum));
    int time_mem_addr = MEM_HIST_TIME_BASE + hist_step*hId;
    std::vector<quint32> y_std_time;
    if (!memReadBlk(time_mem_addr, hist_step - 1, &y_std_time))
        return {};

    MscHistData hist;
    static constexpr const double TICK_SEC = 8e-9; // 125MHz (8ns)
    static constexpr const double SCALER= 54.0;
    for(uint i=0; i<y_std_time.size(); i++){
        MscHistPoint p;
        p.x = exp(i/SCALER) * TICK_SEC * 1e6;
        auto val = y_std_time[i];
        p.y = val ? val : 0.1;
        hist.append(p);
    }
    return hist;
}

std::optional<MscHistData> MscDevice::readPlotHist(quint16 hId)
{
    if (!status.cycleStatus.has_value())
        return {};
    const auto &cycleStatus = status.cycleStatus.value();

    if (!status.histStatus.has_value())
        return {};
    const auto &histStatus = status.histStatus.value();

    auto curGateSt = cycleStatus.gates[hId];
    auto curRamAddr = histStatus.plot_current_ram_address[hId];
    auto prevIt = prevPlotHists.find(hId);
    if(prevIt != prevPlotHists.end()) {
        auto &chHistory = prevIt.value();
        if(chHistory.ramAddr>curRamAddr) {
            //readout tail of the hist
            //todo: emit this tail

            // reset chHistory to new gate state
            chHistory.ramAddr = 0;
            chHistory.gateCnt = curGateSt.gateCnt;
            chHistory.hist.clear();
        }
    } else {
        prevIt = prevPlotHists.insert(hId,{});
        prevIt->gateCnt = curGateSt.gateCnt;
    }
    auto &chHistory = prevIt.value();
    auto hist_step = static_cast<uint16_t>(1<<(buildParams.mscHist->ramBitsHistPlotNum));
    int startAddr = MEM_HIST_PLOT_BASE + hist_step*hId;
    startAddr += chHistory.ramAddr;
    uint sizeToRead = curRamAddr-chHistory.ramAddr;
    std::vector<quint32> plotData;
    if (!memReadBlk(startAddr, sizeToRead, &plotData))
        return {};

    double plot_slice_sec = 1e-6 * config.mscCore.hist[hId].plotSlice;
    for(uint i=0; i<plotData.size(); i++) {
        // Its rate
        auto val = plotData[i];
        auto pos = i+chHistory.ramAddr;
        if(val) {
            if(!chHistory.minPos.has_value())
                chHistory.minPos = pos;
            chHistory.maxPos = pos;
        }
        MscHistPoint p;
        p.x = pos*plot_slice_sec; //plot slicetime in sec
        p.y = 1.*val/plot_slice_sec;
        chHistory.hist.append(p);
    }

    chHistory.ramAddr = curRamAddr;
    Q_ASSERT(chHistory.ramAddr == chHistory.hist.size());

    auto ret = chHistory.getBoundedHist(curGateSt.lastGateWidth);
    return ret;
}

void MscDevice::on_deviceConnected()
{
    ModularMregDevice::on_deviceConnected();

    emit deviceConnectedFw(getDeviceIndex(), getFirmwareVersionStr());

    buildParams.mscInput = inputMod->getBuildParam();
    buildParams.mscHist = histMod->getBuildParam();

    auto dacModule = getFirstModule<Ad5328Module>();
    dacCalibrator->setModules(inputMod, dacModule);

    emit buildParametersUpdated(getDeviceIndex(), buildParams);

    if (buildParams.mscInput.has_value())
        histMod->setHistNum(buildParams.mscInput->hists);

    startCalibration(config_key);
}

void MscDevice::setCalResult(const CalResult &newCalResult)
{
    if (newCalResult.index != getDeviceIndex())
        return;
    calResult = newCalResult;
    syncConfigDeferred();
}

void MscDevice::startCalibration(const ConfigKey &cal_key)
{
    const auto &calResult = performCalibration();
    if (!calResult.has_value()) {
        qWarning() << "Calibration failed";
        emit calibrateFinished(getDeviceIndex(), ConfigKey());
        return;
    }
    const auto &res = calResult.value();
    emit calResultChanged(res);
    emit calibrateFinished(res.index, isConnected() ? cal_key : ConfigKey());
    syncConfigDeferred();
}

std::optional<CalResult> MscDevice::performCalibration()
{
    if (!isConnected())
        return {};
    QElapsedTimer t;
    t.start();

    CalResult res;
    res.index = getDeviceIndex();

    const auto &calibrateDacResult = calibrateDac();
    if (calibrateDacResult.has_value()) {
        // merge calibration data
        const auto &result = calibrateDacResult.value();
        for (auto it = result.begin(); it != result.end(); ++it)
            res.dacCal_mV[it.key()] = it.value();
        qInfo() << "calibrateDac:" << t.elapsed() << "ms";
    }
    dropConfigCache();
    res.temps = readTemp();
    res.nch = res.dacCal_mV.size();
    return res;
}

std::optional<DacCalResult> MscDevice::calibrateDac()
{
    auto thrMod = getFirstModule<Ad5328Module>();
    if (!thrMod || !inputMod)
        return {};

    // TODO: set polarity
    //    const auto prevHitCoreConfig = config.hitCore;
    //    auto hitCoreConfig = prevHitCoreConfig;
    //    hitCoreConfig.polarity = 0;

    bool ok = true;
    //    ok &= hitCore->writeConfig(hitCoreConfig);
    auto ret = dacCalibrator->calibrateDac();
    // ok &= hitCore->writeConfig(prevHitCoreConfig); // writeConfig() is called later
    if (!ok)
        goto err;
    return ret;
err:
    qWarning() << __func__ << "I/O error";
    return {};
}
