#include "MscHistModel.h"

#include <QDateTime>
#include <QDebug>
#include <QColor>

#include "base/TaiTime.h"

#include "config/MscConfigDispatcher.h"

#include "MscModelItem.h"

namespace {
TaiTime getColTime(MscHistColumns col, const MscGateStatus &st)
{
    return col==MscHistColumns::COL_LAST_START
        ? st.lastGateStart()
        : st.lastGateEnd();
}

QString nsToString(int64_t v) {
    if(v<0)
        return "err: <0";
    QString ret = QString::number(1e-9*static_cast<double>(v), 'f', 3);
    return ret;
}

quint16 MscInputChConfig::* getChLutPtr(MscHistColumns col){
    switch(col) {
    case MscHistColumns::COL_CE_LUT: return &MscInputChConfig::ceLut;
    case MscHistColumns::COL_HIST_LUT: return &MscInputChConfig::histLut;
    default:
        return nullptr;
    }
}
}

MscHistModel::MscHistModel(QObject *parent)
    : QAbstractItemModel(parent)
    , rootItem(new MscModelItem())
{
    qRegisterMetaType<MscCycleStatus>();
    auto configDispatcher = &MscConfigDispatcher::instance();
    configDispatcher->subscribeConfig(this);
}

QVariant MscHistModel::data(const QModelIndex &index, int role) const
{
    if(!index.isValid())
        return QVariant();
    switch (role) {
    case Qt::DisplayRole:
        return displayData(index);
    case Qt::EditRole:
        return editData(index);
    case Qt::CheckStateRole:
        return checkStateData(index);
    case Qt::ForegroundRole:
        return foregroundRoleData(index);
    }
    return QVariant();
}

bool MscHistModel::setData(const QModelIndex &model_index,
                           const QVariant &value,
                           int role)
{
    if(!model_index.isValid())
        return false;

    const auto item = getItemByIndex(model_index);
    if(!item)
        return false;
    auto hId = item->item_row;

    if (!config.contains(item->index))
        return false;
    auto &devCfg = config.findDevice(item->index).mscCore;
    MscInputConfig &input = devCfg.input;
    MscHistConfig &hist = devCfg.hist;


    const auto col = getColumnByIndex(model_index);
    if(col==MscHistColumns::COL_EN) {
        if(role != Qt::CheckStateRole)
            return false;
        if(hId < 0) {
                for(int i : hist.keys()){
                    hist[i].enable = value.toBool();
                }
        } else {
                hist[hId].enable = value.toBool();
        }
        columnDataChanged(col);
        emit configChanged(config);
        return true;
    }

    if(role != Qt::EditRole)
        return false;

    if(hId < 0) {
        switch (col) {
        case MscHistColumns::COL_CE_LUT:
        case MscHistColumns::COL_HIST_LUT:
        {
            auto ptr = getChLutPtr(col);
            for(int i : input.keys())
                input[i].*ptr = static_cast<quint16>(value.toUInt());
            break;
        }
        case MscHistColumns::COL_SLICETIME:
            for(int i : hist.keys()){
                hist[i].plotSlice =  value.toUInt();
            }
            break;
        default:;
        }
    } else {
        switch (col) {
        case MscHistColumns::COL_HIST_NAME:
            hist[hId].name = value.toString();
            break;
        case MscHistColumns::COL_CE_LUT:
        case MscHistColumns::COL_HIST_LUT:
        {
            auto ptr = getChLutPtr(col);
            input[hId].*ptr = static_cast<quint16>(value.toUInt());
            break;
        }
        case MscHistColumns::COL_CH:
            input[hId].histCh = static_cast<quint16>(value.toUInt());
            break;
        case MscHistColumns::COL_SLICETIME:
            hist[hId].plotSlice = value.toUInt();
            break;
        default:;
        }
    }
    columnDataChanged(col);
    emit configChanged(config);
    return true;
}

Qt::ItemFlags MscHistModel::flags(const QModelIndex &index) const
{
    auto f = QAbstractItemModel::flags(index);
    const MscModelItem *childItem =
        static_cast<MscModelItem*>(index.internalPointer());
    auto col = getColumnByIndex(index);
    if(childItem) {
        if(col == MscHistColumns::COL_HIST_NAME) {
            f |= Qt::ItemIsEditable;
        }
    }
    if(col == MscHistColumns::COL_EN)
        f |= Qt::ItemIsUserCheckable;

    return f;
}

QVariant MscHistModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole) {
        switch (static_cast<MscHistColumns>(section)) {
        case MscHistColumns::COL_HIST_NAME: return "Histogram name    ";
        case MscHistColumns::COL_EN: return "En";
        case MscHistColumns::COL_CE_LUT: return "CE LUT";
        case MscHistColumns::COL_HIST_LUT: return "Gate LUT";
        case MscHistColumns::COL_CH: return "Ch     ";
        case MscHistColumns::COL_SLICETIME: return "Plot Slicetime, us";
        case MscHistColumns::COL_CUR_ST: return "Cur st.";
        case MscHistColumns::COL_ACTIVE_CNT: return "Gate cnt";
        case MscHistColumns::COL_FULL_CNT: return "Cycle cnt";
        case MscHistColumns::COL_LAST_START: return "Last/current start";
        case MscHistColumns::COL_LAST_END: return   "Last/current end  ";
        case MscHistColumns::COL_LAST_WIDTH: return   "Gate width, s";
        case MscHistColumns::COL_LAST_PERIOD: return   "Cycle period, s";
        case MscHistColumns::COL_GATE_NUM: return "Gate number";
        default: return {};
        }
    }
    // TODO return DeviceIndexTableModel::headerData(section, orientation, role);
    return {};
}

QModelIndex MscHistModel::index(int row, int column, const QModelIndex &parent) const
{
    if(!hasIndex(row, column, parent))
        return QModelIndex();
    const MscModelItem *parentItem = parent.isValid() ?
        static_cast<MscModelItem*>(parent.internalPointer()) :
        rootItem;
    MscModelItem *childItem = parentItem->child(row);
    if(childItem) {
        return createIndex(row, column, childItem);
    }
    return QModelIndex();
}

QModelIndex MscHistModel::parent(const QModelIndex &index) const
{
    if(!index.isValid())
        return QModelIndex();

    const MscModelItem *childItem =
        static_cast<MscModelItem*>(index.internalPointer());
    if(!childItem)
        return QModelIndex();
    MscModelItem *parentItem = childItem->parent;
    if(parentItem == rootItem)
        return QModelIndex();

    return createIndex(parentItem->row(), 0, parentItem);
}

int MscHistModel::rowCount(const QModelIndex &parent) const
{
    const auto *parentItem = parent.isValid() ?
        static_cast<MscModelItem*>(parent.internalPointer()) :
        rootItem;
    return parentItem->childCount();
}

void MscHistModel::cycleStatusUpdated(const DeviceIndex &devIndex,
                                      const MscCycleStatus &newSt)
{
    status[devIndex] = newSt;
    columnDataChanged(MscHistColumns::COL_CUR_ST);
    columnDataChanged(MscHistColumns::COL_ACTIVE_CNT);
    columnDataChanged(MscHistColumns::COL_FULL_CNT);
    columnDataChanged(MscHistColumns::COL_LAST_START);
    columnDataChanged(MscHistColumns::COL_LAST_END);
    columnDataChanged(MscHistColumns::COL_LAST_WIDTH);
    columnDataChanged(MscHistColumns::COL_LAST_PERIOD);
    columnDataChanged(MscHistColumns::COL_GATE_NUM);
}

void MscHistModel::setConfig(const ConfigKey &config_key, const BaseDeviceAppConfig &newConfig)
{
    Q_UNUSED(config_key)
    if(newConfig == config){
        return;
    }

    const auto &oldSet = config.getDeviceIndexEnabledMap().getEnabled();
    const auto &newSet = newConfig.getDeviceIndexEnabledMap().getEnabled();
    config = newConfig;

    QMap<DeviceIndex, int> histNum_new;
    for (const auto &index : config.deviceKeys()) {
        histNum_new[index] = config.findDevice(index).mscCore.hist.size();
    }

    if((newSet != oldSet)||(histNum_new != histNum)){
        histNum = histNum_new;
        reset();
    }
    else{
        auto top = index(0,0);
        auto lastDev = index(rowCount(top)-1, columnCount(top)-1, top);
        auto bottom = index(rowCount(lastDev)-1, columnCount(lastDev)-1, lastDev);
        emit dataChanged(top, bottom, {Qt::DisplayRole});
    }
}

QVariant MscHistModel::displayData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    if(!item) return QVariant();
    const auto col = getColumnByIndex(index);

    auto hId = item->item_row;
    auto cycleSt = status.value(item->index);
    auto gates = cycleSt.gates.value(hId);
    const auto &dev_config = config.findDevice(item->index).mscCore;
    const auto &input = dev_config.input;
    if(hId < 0){
        switch (col) {
        case MscHistColumns::COL_HIST_NAME:
            return item->index.getSerialStr();
        case MscHistColumns::COL_CE_LUT:
        case MscHistColumns::COL_HIST_LUT:
            return input[0].*getChLutPtr(col);
        default:;
        }
    }
    else {
        switch (col) {
        case MscHistColumns::COL_HIST_NAME: {
            return getMscHistDisplayName(dev_config.hist, hId);
        }
        case MscHistColumns::COL_CUR_ST:
            return QChar(cycleSt.isInGate(hId) ? 0x2713 : 0x2715);
        case MscHistColumns::COL_ACTIVE_CNT: return gates.activeGateCnt;
        case MscHistColumns::COL_FULL_CNT: return gates.fullGateCnt;
        case MscHistColumns::COL_LAST_START:
        case MscHistColumns::COL_LAST_END:
            return QDateTime::fromTime_t(static_cast<uint>(getColTime(col, gates).sec))
                    .toString("dd/MM/yy hh:mm:ss");
        case MscHistColumns::COL_LAST_WIDTH:
        case MscHistColumns::COL_LAST_PERIOD:
        {
            const auto v = col==MscHistColumns::COL_LAST_WIDTH
                    ? gates.lastGateWidth : gates.lastGatePeriod;
            return v.has_value() ? nsToString(v.value()) : "-";

        }
        case MscHistColumns::COL_GATE_NUM:
            return gates.gateCnt;
        default:;
        }
    }
    return QVariant();
}

QVariant MscHistModel::editData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    const auto col = getColumnByIndex(index);
    if(!item)
        return QVariant();

    auto hId = item->item_row;
    const auto &dev_config = config.findDevice(item->index).mscCore;
    const auto &input = dev_config.input;

    if(hId >= 0) {
        if(!dev_config.hist.contains(hId))
            return QVariant();
        const auto &hist = dev_config.hist[hId];
        switch (col) {
        case MscHistColumns::COL_HIST_NAME: return hist.name;
        case MscHistColumns::COL_CE_LUT:
        case MscHistColumns::COL_HIST_LUT:
            return input[hId].*getChLutPtr(col);
        case MscHistColumns::COL_CH: return input[hId].histCh;
        case MscHistColumns::COL_SLICETIME: return hist.plotSlice;
        default:;
        }
    }
    else {
        const auto &hist_0 = dev_config.hist[0];
        switch (col) {
        case MscHistColumns::COL_HIST_NAME: return hist_0.name;
        case MscHistColumns::COL_CE_LUT:
        case MscHistColumns::COL_HIST_LUT:
            return input[0].*getChLutPtr(col);
        case MscHistColumns::COL_SLICETIME: return hist_0.plotSlice;
        default:;
        }
    }
    return QVariant();
}

QVariant MscHistModel::checkStateData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    if(!item) return QVariant();

    const auto col = getColumnByIndex(index);
    if(col != MscHistColumns::COL_EN)
        return QVariant();

    const auto &dev_config = config.findDevice(item->index).mscCore;
    auto hId = item->item_row;
    bool st;
    if(hId<0) {
        st = dev_config.hist[0].enable;
        for(auto v : dev_config.hist) {
            if(st != v.enable)
                return Qt::PartiallyChecked;
        }
    } else
        st = dev_config.hist[hId].enable;

    return st ? Qt::Checked : Qt::Unchecked;
}

QVariant MscHistModel::foregroundRoleData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    const auto col = getColumnByIndex(index);
    if(!item || item->item_row<0)
        return QVariant();

    auto cycleSt = status.value(item->index);
    auto gSt = cycleSt.gates.value(item->item_row);
    switch (col) {
    case MscHistColumns::COL_CUR_ST:
        return cycleSt.isInGate(index.row()) ? QColor(Qt::darkGreen) : QColor();
    case MscHistColumns::COL_LAST_START:
    case MscHistColumns::COL_LAST_END:
        return QColor((getColTime(col, gSt).valid)
                      ? Qt::darkGreen : Qt::red);
    default: return QVariant();
    }
}

MscModelItem *MscHistModel::getItemByIndex(const QModelIndex &index) const
{
    return static_cast<MscModelItem*>(index.internalPointer());
}

MscHistColumns MscHistModel::getColumnByIndex(const QModelIndex &index) const
{
    return static_cast<MscHistColumns>(index.column());
}

void MscHistModel::columnDataChanged(MscHistColumns col)
{
    QVector<int> roles;
    switch (col) {
    case MscHistColumns::COL_EN:
        roles = {Qt::CheckStateRole};
        break;
    case MscHistColumns::COL_HIST_NAME:
        roles = {Qt::DisplayRole, Qt::EditRole};
        break;
    case MscHistColumns::COL_CE_LUT:
    case MscHistColumns::COL_HIST_LUT:
    case MscHistColumns::COL_CH:
    case MscHistColumns::COL_SLICETIME:
        roles = {Qt::EditRole};
        break;
    case MscHistColumns::COL_CUR_ST:
    case MscHistColumns::COL_ACTIVE_CNT:
    case MscHistColumns::COL_FULL_CNT:
    case MscHistColumns::COL_LAST_START:
    case MscHistColumns::COL_LAST_END:
    case MscHistColumns::COL_LAST_WIDTH:
    case MscHistColumns::COL_LAST_PERIOD:
    case MscHistColumns::COL_GATE_NUM:
        roles = {Qt::DisplayRole};
        if(col==MscHistColumns::COL_CUR_ST)
            roles << Qt::ForegroundRole;
        break;
    default: return;
    }

    auto devs = rowCount();
    for(int d=0; d<devs; ++d) {
        auto dev = index(d, static_cast<int>(col));
        emit dataChanged(dev, dev, roles);
        auto chNum = rowCount(dev);
        for(int ch=0; ch<chNum; ++ch) {
            auto chItem = index(ch, static_cast<int>(col), dev);
            emit dataChanged(chItem, chItem, roles);
        }
    }
}

void MscHistModel::reset()
{
    beginResetModel();
    rootItem->children.clear();
    for (const auto &key : config.deviceKeys()) {
        auto devItem = new MscModelItem(rootItem);
        devItem->index = key;
        for(int hId=0; hId<histNum[devItem->index]; ++hId) {
            auto chItem = new MscModelItem(devItem);
            chItem->index = key;
            chItem->item_row = hId;
            devItem->addChild(chItem);
        }
        rootItem->addChild(devItem);
    }
    endResetModel();
    return;
}


