#include "TcpClient.h"

#include <QtNetwork>
#include <QDebug>
#include <QSemaphore>

#if QT_VERSION < 0x050000
Q_DECLARE_METATYPE(QAbstractSocket::SocketError)
#endif

namespace {
const int PACK_SIZE_LIMIT = 2*1024*1024;
}

TcpClient::TcpClient(QSemaphore *_analyzerSem, QSemaphore *_writerSem, QObject *parent) :
    QObject(parent),
    analyzerSem(_analyzerSem),
    writerSem(_writerSem),
    enabled(false),
    tcpSocket(new QTcpSocket(this)),
    hostPort(0),
    reconnectTimer(this),
    delayReadTimer(this),
    statusTimer(this),
    connectingAttempts(0),
    lastDataRecieved(0),
    dataRecieved(0),
    rawWriteEnabled(false)
{
    connect(tcpSocket, &QTcpSocket::readyRead, this, &TcpClient::readData, Qt::QueuedConnection);
    connect(tcpSocket, &QTcpSocket::connected, this, &TcpClient::connected, Qt::QueuedConnection);
    connect(tcpSocket, &QTcpSocket::disconnected, this, &TcpClient::disconnected, Qt::QueuedConnection);
    qRegisterMetaType<QAbstractSocket::SocketError>();

    connect(tcpSocket, QOverload<QAbstractSocket::SocketError>::of(&QAbstractSocket::error),
                this, &TcpClient::displayError);

    connect(&reconnectTimer, &QTimer::timeout, this, &TcpClient::requestNewData, Qt::QueuedConnection);
    reconnectTimer.setSingleShot(true);

    delayReadTimer.setSingleShot(true);
    delayReadTimer.setInterval(1000);
    connect(&delayReadTimer, &QTimer::timeout, this, &TcpClient::readData, Qt::QueuedConnection);

    statusTimer.setSingleShot(false);
    connect(&statusTimer, &QTimer::timeout, this, &TcpClient::sendStatus);
    statusTimer.start(1000);
    statusPeriod.start();
}


TcpClient::~TcpClient()
{
    stopReadOut();
    delete tcpSocket;
}


void TcpClient::requestNewData()
{
    try{
    if(!enabled) return;

    if(tcpSocket->isValid() &&
            (tcpSocket->state() == QTcpSocket::ConnectedState ||
             tcpSocket->state() == QTcpSocket::ConnectingState) )
        return;
    tcpSocket->abort();
    tcpSocket->connectToHost(hostAddr, hostPort);
    ++connectingAttempts;
    if(tcpSocket->waitForConnected(1000)){
        qInfo().noquote() << "[TcpClient]: connected.";
    } else {
        if(connectingAttempts<=1){
            QString mess = QString("[TcpClient]: can't connected.");
            qInfo().noquote() << mess;
            addLog(mess);
            reconnectTimer.start(1000);
        } else
            reconnectTimer.start(5000);
    }
    } catch(...){
        qInfo().noquote() << "[TcpClient]: Exception in TcpClient::requestNewData()";
    }
}


void TcpClient::readData()
{
    try {
    if(!enabled) return;
    int acquireCnt = 0;
    while(!analyzerSem->tryAcquire(1, 100)) {
        if(acquireCnt==0) {
            qWarning()<<QString("[%1] tcp read delayed (analyzer is slow)")
                     .arg(QTime::currentTime().toString("hh:mm:ss:zzz"));
        }
        if(++acquireCnt>10) {
            delayReadTimer.start(200);
            return;
        }
    }
    if(rawWriteEnabled) {
        while(!writerSem->tryAcquire(1, 100)){
            if(acquireCnt==0)
                qWarning()<<QString("[%1] tcp read delayed (writer is slow)")
                         .arg(QTime::currentTime().toString("hh:mm:ss:zzz"));
            if(++acquireCnt>10) {
                delayReadTimer.start(200);
                return;
            }
        }
    }

    qint64 avail = tcpSocket->bytesAvailable();
    if(!avail) {
        analyzerSem->release();
        if(rawWriteEnabled) writerSem->release();
//        qInfo()<<"no data "<<queryFreeSize.available();
        // readData() will be invoqed by readyRead() signal
        return;
    }

    QVector<QByteArray> dataArray;
    int sumSize=0;
    while(avail) {
        QByteArray reciveData = tcpSocket->readAll();
        if(reciveData.size() < avail){
            qWarning() << "We got data lesser that expected.";
        }
        dataArray.push_back(reciveData);
        sumSize += reciveData.size();
        if(sumSize > PACK_SIZE_LIMIT)
            break;

        avail = tcpSocket->bytesAvailable();
//        if(avail){
//            qInfo()<<"double recive";
//        }
    }
    dataRecieved += sumSize;
//    if(++counter%10==0){
//        qInfo() <<QString("analyzer=%1   writer=%2").arg(analyzerQueueSize->available(),3).arg(writerQueueSize->available(),3);
//    }

    if(dataArray.size()==1) {
        emit dataAccuired(dataArray[0], rawWriteEnabled);
    } else {
        QByteArray summedData;
        for (QByteArray data : dataArray) {
            summedData.append(data);
        }
        emit dataAccuired(summedData, rawWriteEnabled);
    }

    } catch(...) {
        qWarning()<<"Exception in TcpClient::readData()";
        qWarning()<<"Possible loss of semaphores in TcpClient::readData";
    }
}


void TcpClient::displayError(QAbstractSocket::SocketError socketError)
{
    switch (socketError) {
    case QAbstractSocket::RemoteHostClosedError:
        qInfo() << "Remote hoste closed connection.";
        addLog("Remote hoste closed connection.");
        break;
    case QAbstractSocket::HostNotFoundError:
        qInfo() << "The host was not found. Please check the "
                    "host name and port settings.";
        addLog("The host was not found. Please check the "
               "host name and port settings.");
        break;
    case QAbstractSocket::ConnectionRefusedError:
        if(connectingAttempts<=1){
            qInfo() << tr("The connection was refused by the peer. "
                           "Make sure the fortune server is running, "
                           "and check that the host name and port "
                           "settings are correct.");
            addLog("The connection was refused by the peer. "
                   "Make sure the fortune server is running, "
                   "and check that the host name and port "
                   "settings are correct.");
        }
        break;
    default:
        qInfo() << tr("The following error occurred: %1.")
                    .arg(tcpSocket->errorString());
        addLog(tr("The following error occurred: %1.")
               .arg(tcpSocket->errorString()));
    }
}

void TcpClient::setAddress(QHostAddress addr, int port)
{
    if(hostAddr != addr || hostPort != port) {
        if(tcpSocket->isValid() &&
                (tcpSocket->state() == QTcpSocket::ConnectedState ||
                 tcpSocket->state() == QTcpSocket::ConnectingState) ) {
            return;
        }

        qInfo().noquote() << "[TcpClient]: setAddress: "<<addr.toString()<<port;
        stopReadOut();
        hostAddr = addr;
        hostPort = port;
        startReadOut();
    }
}

void TcpClient::startReadOut()
{
    enabled = true;
//    requestNewData();
    reconnectTimer.start(0);
}


void TcpClient::stopReadOut()
{
    enabled = false;
    if(tcpSocket->state() != QTcpSocket::UnconnectedState)
        tcpSocket->disconnectFromHost();
    reconnectTimer.stop();
}


void TcpClient::connected()
{
    emit connectedStateChanged(true);
}


void TcpClient::disconnected()
{
    emit connectedStateChanged(false);
    if(enabled){
        qInfo()<<"Client was disconnected, trying to reconnect";
        connectingAttempts=0;
        reconnectTimer.start(1000);
    } else
        qInfo()<<"Client disconnected.";
}

void TcpClient::startWrite(QString, bool rawData)
{
    rawWriteEnabled = rawData;
}

void TcpClient::stopWrite()
{
    rawWriteEnabled = false;
}

void TcpClient::sendStatus()
{
    const bool connOk = enabled && tcpSocket!=nullptr && tcpSocket->state() == QTcpSocket::ConnectedState;
//    if(connOk)
//        qInfo() << "TcpClient::bytesAvailable" << 1e-6*tcpSocket->bytesAvailable() << "MB";
    double rate = 1e3*(dataRecieved-lastDataRecieved)/statusPeriod.restart();
    lastDataRecieved=dataRecieved;
    emit statusUpdated(connOk, rate);
//    qInfo() << "TcpClient: status "<< (connOk?"connected":"disconnected") << dataRecieved;
}
