#include "TqdcDevice.h"

#include <QDateTime>
#include <QElapsedTimer>
#include <QFile>
#include <QTextStream>
#include <QVector>
#include <math.h>

#include "Ad5328/Ad5328Module.h"
#include "AdcSpiLtm9011/AdcSpiLtm9011Module.h"
#include "BoardGpio/BoardGpioModule.h"
#include "DeserializersControl/DeserializersControlModule.h"
#include "HitCore/HitCoreModule.h"
#include "LegacyDeviceModule.h"
#include "RunLogic/RunLogicModule.h"
#include "TqdcAd5328ConfigBuilder.h"
#include "TqdcCore/TqdcHardwareDef.h"
#include "TqdcDacCalibrator.h"
#include "TqdcDacConverter.h"
#include "TqdcRegs.h"
#include "TrigCtrlStat/TrigCtrlStatModule.h"
#include "TdcCore/TdcCoreModule.h"
#include "TrigInfo/TrigInfoModule.h"
#include "TtlIo/TtlIoModule.h"
#include "WaveBlc/WaveBlcModule.h"
#include "WaveRec/WaveRecModule.h"
#include "base/DeviceId.h"
#include "regio/regop.h"

namespace {
int hptdcSetupIndex = 0;
const int CALIBRATE_LIMIT = 1000;

//const double DAC_V_REF = 3.3 * 82.5 / (82.5 + 130); // in V (1.2812)
//const quint16 DAC_RANGE = (1<<12) - 1;
//const int DAC_DEFAULT_ZERO = 1920;
constexpr fw_version_t REV_MAX_SAMPLES_CFG =fw_version_t(2, 0, 24223);

const int TQDC_CH_TO_DAC_CH_MAP[16] = {
    3,2,1,0, 4,5,6,7, 0,1,2,3, 7,6,5,4
};
const int tqdc_map_chtdc2inp[2][32] = {
    {15,-1,-1,-1,14,-1,-1,-1,13,-1,-1,-1,12,-1,-1,-1,11,-1,-1,-1,10,-1,-1,-1, 9,-1,-1,-1, 8,-1,-1,-1 },
    { 7,-1,-1,-1, 6,-1,-1,-1, 5,-1,-1,-1, 4,-1,-1,-1, 3,-1,-1,-1, 2,-1,-1,-1, 1,-1,-1,-1, 0,-1,-1,-1 }
};

static const double HIT_COUNTERS_TIMER_FREQUENCY = 125e6;

} // namespace

TqdcDevice::TqdcDevice(const DeviceIndex &index, QObject *parent)
    : ModularMregDevice(index, statusCache, parent)
    , statusCache(index)
    , legacyDevice (createModule<LegacyDeviceModule>())
    , dacCalibrator(new TqdcDacCalibrator(TqdcAd5328ConfigBuilder::instance()))
{
}

TqdcDevice::~TqdcDevice()
{
    delete dacCalibrator;
}

void TqdcDevice::enforceConfigConstraints(ModularDeviceConfig *config)
{
    // enforce config constraints
    config->tqdcCore.blc.staticOffset = 0;

    config->tqdcCore.makeTdcCoreConfig(&config->tdcCore, getDeviceId());
    config->tqdcCore.makeWaveRecConfig(&config->waveRec, getDeviceId());
    config->tqdcCore.makeWaveTrigConfig(&config->waveTrig);
    config->tqdcCore.makeBoardGpioConfig(&config->boardGpio);
    config->tqdcCore.makeHitCoreConfig(&config->hitCore);

    quint16 chTrigMask = 0;
    for(quint16 ch=0; ch<getNumberOfChannels(); ch++) {
        if(config->tqdcCore.chTrigEn.value(ch, config->tqdcCore.DEF_TRIG_EN)) {
            chTrigMask |= quint16(1) << ch;
        }
    }
    config->trigCtrlStat.analog_thr_channel_en = chTrigMask;

    // apply ADC baseline calibration
    const auto adcBaseline = config->tqdcCore.isGain4() ? calResult.baseline_x4 : calResult.baseline_x1;
    for (auto it = adcBaseline.begin(); it != adcBaseline.end(); ++it) {
        auto ch = it.key();
        auto baseline = it.value();
        auto channel = config->waveBlc.getChannel(ch);
        channel.staticOffset = -baseline.mean;
        config->waveBlc.setChannel(ch, channel);
    }
    auto dacModule = getFirstModule<Ad5328Module>();
    int index = 0;
    if (dacModule) {
        // apply Threshold DAC calibration
        QMap<int, int> dacValues;
        auto nch = TqdcCoreConfig::getNumberOfChannels(getDeviceId());
        for (int ch=0; ch<nch; ch++) {
            double thr = config->tqdcCore.chThr[ch];
            double offset = calResult.dacCal_mV[ch];
            dacValues[ch] = TqdcDacConverter::instance().dacValue(thr + offset);
        }
        //auto index = dacModule->moduleIndex();
        config->ad5328Configs[index++] = TqdcAd5328ConfigBuilder::instance().build(dacValues);
    }
}

bool TqdcDevice::writeConfig()
{
    const auto setupCacheValid = isConfigCacheValid();
    bool ok = true;
    if (!config.isRun()) {
        ok &= StopRun();
    }

    ok &= ModularMregDevice::writeConfig();

    if (!setupCacheValid || (config.isRun() && !configCache.isRun())) {
        ok &= preRunActions();
    }

    if (config.isRun()) {
        ok &= StartRun();
    }

    return ok;
}

void TqdcDevice::writeConfigWithPause()
{
    if (!isConnected())
        return;

    StopRun();

    if (!waitEndOfRun()) {
        if (isConnected()) {
            qWarning("waitEndOfRun timeout");
        } else {
            return;
        }
    }

    syncConfigDeferred();

    clearStatistics();
}

void TqdcDevice::writeSetupForce()
{
    StopRun();
    dropConfigCache();
    on_deviceConnected();
//    writeSetup();
}

std::optional<CalResult> TqdcDevice::performTqdcCalibration()
{
    if (!isConnected())
        return {};
    QElapsedTimer t;
    t.start();

    CalResult res;
    res.index = getDeviceIndex();

    auto boardGpioModule = getFirstModule<BoardGpioModule>();
    if (!boardGpioModule)
        return {};

    auto tmpConfig = config;
    tmpConfig.tqdcCore.setGain(TqdcCoreConfig::Gain1);
    enforceConfigConstraints(&tmpConfig);
    if (!boardGpioModule->writeConfig(tmpConfig.boardGpio))
        return {};

    if (!calibrateBaseline(&res.baseline_x1))
        return {};

    tmpConfig.tqdcCore.setGain(TqdcCoreConfig::Gain4);
    enforceConfigConstraints(&tmpConfig);
    if (!boardGpioModule->writeConfig(tmpConfig.boardGpio))
        return {};
    if (!calibrateBaseline(&res.baseline_x4))
        return {};

    qInfo() << "calibrateBaseline:" << t.restart() << "ms";
    const auto &calibrateDacResult = calibrateDac();
    if (calibrateDacResult.has_value()) {
        // merge calibration data
        const auto &result = calibrateDacResult.value();
        for (auto it = result.begin(); it != result.end(); ++it)
            res.dacCal_mV[it.key()] = it.value();
        qInfo() << "calibrateDac:" << t.elapsed() << "ms";
    }
    dropConfigCache();
    res.temps = readTemp();
    res.nch = getNumberOfChannels();
    return res;
}

void TqdcDevice::startTqdcCalibration(const ConfigKey &cal_key)
{
    const auto &calResult = performTqdcCalibration();
    if (!calResult.has_value()) {
        qWarning() << "Calibration failed";
        emit calibrateFinished(getDeviceIndex(), ConfigKey());
        return;
    }
    const auto &res = calResult.value();
    emit calResultChanged(res);
    emit calibrateFinished(res.index, isConnected() ? cal_key : ConfigKey());
    writeConfigWithPause();
}

bool TqdcDevice::softReset()
{
    if(!isConnected())
        return {};
    dropConfigCache();
    RegOpVector r;
    r.RegWrite16(REG_DEVICE_CTRL, CTRL_BIT_SOFTCLEAR);
    r.RegWrite16(REG_DEVICE_CTRL, 0);
    return legacyDevice->regOpExecRebased(r);
}

bool TqdcDevice::readStatus()
{
    if (!ModularMregDevice::readStatus())
        return {};

    emit statusUpdated(statusCache);
    return true;
}

void TqdcDevice::on_deviceConnected()
{
    ModularMregDevice::on_deviceConnected();

    auto adcSpi = getAdcSpiModule();
    auto dacModule = getFirstModule<Ad5328Module>();
    auto hitCore = getFirstModule<HitCoreModule>();
    dacCalibrator->setModules(hitCore, dacModule);

    setupAdc();

    auto waveRec = getFirstModule<WaveRecModule>();
    if (waveRec) {
        if (adcSpi) {
            if(auto bp=waveRec->getBuildParams())
                adcSpi->setAdcChips(bp->adcChips);
        }
        auto buildParam = waveRec->getBuildParams();
        if (buildParam && buildParam->isValid()) {
            int maxSamplesWithOffset = buildParam->adcMaxMatchWin - 1 - tqdc16vs_adc_lat_offset;
            maxSamplesWithOffset /= tqdc16vs_adc_scale;
            maxSamplesWithOffset *= tqdc16vs_adc_scale;
            emit deviceInfoUpdated(getDeviceIndex(), maxSamplesWithOffset);
        } else {
            qWarning() << "Invalid AdcBuildParams";
        }
    }

    // writeConfigWithPause(); // already done in syncConfigDeferred

    auto ttlModule = getFirstModule<TtlIoModule>();
    if (ttlModule)
        emit ttlIoStartupCfg(ttlModule->readConfig());
}

bool TqdcDevice::StartRun()
{
    return writeSoftRun(true);
}

bool TqdcDevice::StopRun()
{
    return writeSoftRun(false);
}

int TqdcDevice::getNumberOfChannels() const
{
    return TqdcCoreConfig::getNumberOfChannels(getDeviceId());
}

bool TqdcDevice::setupAdc()
{
    bool ok = true;
    auto desModule = getFirstModule<DeserializersControlModule>();
    if (desModule)
        ok &= desModule->setDeserReset(true);

    auto adcModules = getModules<AdcSpiLtm9011Module>();
    for (auto adcModule : adcModules) {
        adcModule->setupAdc();
    }

    if (desModule)
        ok &= desModule->setDeserReset(false);
//    ResetAdc();
    return ok;
}

//QString TQDCDevice::adcStatusStr(int status)
//{
//    quint8 adc_locked = status & 0xFF;
//    quint8 adc_ramp_error = (status >> 8) & 0xFF;
//    return QString("ADC status: error=%1, lock=%2").arg(adc_ramp_error, 8, 2, QChar('0')).arg(adc_locked, 8, 2, QChar('0'));
//}

bool TqdcDevice::waitEndOfRun()
{
    QElapsedTimer timer;
    timer.start();
    while (true) {
        auto trigCtrlStat = getFirstModule<TrigCtrlStatModule>();
        if (!trigCtrlStat)
            return false;
        auto status = trigCtrlStat->readStatus();
        if (!status)
            return false;
        if ((status->trigQueue & 0xF) == 0)
            break;
        if (timer.elapsed() > 2000) {
            qWarning() << QString("regRead(REG_TRIG_QUEUE): %1").arg(status->trigQueue, 4, 16, QChar('0'));
            return false;
        }
    }
    return true;
}

bool TqdcDevice::calibrateBaseline(MultichannelBaselineStats *result)
{
    auto waveBlc = getFirstModule<WaveBlcModule>();
    if (!waveBlc)
        return {};
    // auto waveRec = getFirstModule<WaveRecModule>();
    bool ok = true;
    // ok &= waveRec->readBaseline(result, CALIBRATE_LIMIT);
    // qInfo().noquote() << "Waveform Recorder (" + getIdent() + ")\n" + result->toString();
    ok &= waveBlc->readBaseline(result, CALIBRATE_LIMIT);
    qInfo().noquote() << "Waveform BLC (" + getIdent() + ")\n" + result->toString();
    return ok;
}

std::optional<DacCalResult> TqdcDevice::calibrateDac()
{
    auto thrMod = getFirstModule<Ad5328Module>();
    auto hitCore = getFirstModule<HitCoreModule>();
    if (!thrMod || !hitCore)
        return {};

    const auto prevHitCoreConfig = config.hitCore;
    auto hitCoreConfig = prevHitCoreConfig;
    hitCoreConfig.polarity = 0;

    bool ok = true;
    ok &= hitCore->writeConfig(hitCoreConfig);
    auto ret = dacCalibrator->calibrateDac();
    // ok &= hitCore->writeConfig(prevHitCoreConfig); // writeConfig() is called later
    if (!ok)
        goto err;
    return ret;
err:
    qWarning() << __func__ << "I/O error";
    return {};
}

void TqdcDevice::clearStatistics()
{
    // statusCache.statistic.clear();
}

void TqdcDevice::setCalResult(const ConfigKey &new_config_key, const CalResult &newCalResult)
{
    if (newCalResult.index != getDeviceIndex())
        return;
    calResult = newCalResult;
    writeConfigWithPause();
}

QString TqdcDevice::networkPortName(int portIndex)
{
    switch(portIndex) {
    case NetworkPortType::SFP0: return "sfp0";
    case NetworkPortType::SFP1: return "sfp1";
    case NetworkPortType::SFP2: return "sfp2";
    }
    return {};
}

QSharedPointer<AdcSpiBaseModule> TqdcDevice::getAdcSpiModule() const
{
    return getFirstModule<AdcSpiLtm9011Module>();
}
