#include "TrcDeviceGlobal.h"
#include "TrcDevice.h"

#include <QJsonDocument>
#include <QJsonValue>
#include <QThread>

#include "TrcDevice.h"
#include "TrcSignalRouter.h"
#include "base/AbstractConfigConverter.h"
#include "mstream-lib/MStreamDump.h"

namespace {
static constexpr const int TRC_CH = 16;

QJsonObject getInputObject(const BaseDevCounters &cnt) {
    QJsonObject ret;
    for(auto it=cnt.cnt.cbegin(); it!=cnt.cnt.cend(); ++it) {
        QJsonArray chArr;
        for(int ch=0; ch<TRC_CH; ++ch) {
            chArr.append(QJsonValue(static_cast<qint64>(it->value(ch))));
        }
        auto key = QString("%1: %2").arg(it.key())
                .arg(getDevCountersType(DevCountersTrcIn, it.key()));
        ret[key] = chArr;
    }
    return ret;
}
}

static bool allWrSaveTimeIsValid(const WrStatus &wrStatus)
{
    // Check for:
    // 0-CurTime, 2-SpillStart, 3-SpillEnd
    for (auto i : {0, 2, 3})
        if (wrStatus.savedTimeCodeRaw.at(i) == 0)
            return false;
    return true;
}

TrcDeviceGlobal::TrcDeviceGlobal(const DeviceIndex &index, QObject *parent)
    : MstreamMregDevGlobal(new TrcDevice(index), trcDevStatus, parent)
{
    AbstractConfigConverter<TaiTime>::registerTypes();

    TrcSignalRouter::instance().connect_TrcDeviceGlobal(this);
}

void TrcDeviceGlobal::deviceStatusUpdated(const TrcStatus &status)
{
    Q_ASSERT(getDeviceIndex() == status.getDeviceIndex());
    if (getDeviceIndex() != status.getDeviceIndex())
        return;

    auto inSpill = [](const TrcStatus &s)->bool {return s.phyTrig.spill.isReadoutPhase(); };
    if(trcDevStatus.phyTrig.spill.count != status.phyTrig.spill.count
        || trcDevStatus.phyTrig.spill.lastSpillTai() != status.phyTrig.spill.lastSpillTai()
            || inSpill(trcDevStatus) != inSpill(status)
            || pendingStatusSend) {
        if (status.wrStatus.has_value() && allWrSaveTimeIsValid(status.wrStatus.value())) {
            onSpillStateUpdated(status);
            pendingStatusSend = false;
        } else {
            pendingStatusSend = true;
        }
    }
    trcDevStatus = status;
}

void TrcDeviceGlobal::postAddMStreamDump(MStreamDump *mstream)
{
        connect(this, &TrcDeviceGlobal::sendJsonBlock,
                mstream, &MStreamDump::sendJsonBlock);
}

void TrcDeviceGlobal::onSpillStateUpdated(const TrcStatus &devSt)
{
    QVariantMap devRunStatus;

    {
        QVariantMap spillStatus;
        auto spill = devSt.phyTrig.spill;
        spillStatus["count"] = spill.count;
        spillStatus["phase"] = spill.status.reg.bit.phase;
        spillStatus["curSpillTimeMs"] = spill.curSpillTimeMs;
        spillStatus["lastSpillTai"] = QVariant::fromValue(spill.lastSpillTai())
                .value<QJsonObject>();

        if (devSt.wrStatus.has_value()) {
            auto spillStartTime = devSt.wrStatus->savedTimeCode(2);
            spillStatus["spillStartTaiWrStat"] = QVariant::fromValue(spillStartTime)
                                                     .value<QJsonObject>();
            auto spillStopTime = devSt.wrStatus->savedTimeCode(3);
            spillStatus["spillStopTaiWrStat"] = QVariant::fromValue(spillStopTime)
                                                    .value<QJsonObject>();
        }
        devRunStatus["spill"] = spillStatus;
    }

    {
        const auto &temps = devSt.getTemps();
        QVariantMap tempStatus;
        for(auto it=temps.cbegin(); it!=temps.cend(); ++it) {
            auto id = it.key();
            tempStatus[id] = it.value();
        }
        devRunStatus["temp"] = tempStatus;
    }

    {
        QVariantMap cntStatus;
        {
            QVector<qint64> bVec, aVec; // before err, after err
            const auto &cntBaErr = devSt.phyTrig.rcCounters[DevCountersTrcBa].cnt;

            for(int ch=0; ch<TRC_CH; ++ch) {
                bVec.push_back(cntBaErr[0][ch]);
                aVec.push_back(cntBaErr[1][ch]);
            }
            cntStatus["input"] = getInputObject(devSt.phyTrig.rcCounters[DevCountersTrcIn]);
            cntStatus["input_afterReduc"] = getInputObject(devSt.phyTrig.rcCounters[DevCountersTrcReduc]);
            cntStatus["befor_err"] = ConfigConverterUtil::toJsonValue(bVec);
            cntStatus["after_err"] = ConfigConverterUtil::toJsonValue(aVec);
        }
        devRunStatus["counters"] = cntStatus;
    }

    {
        QVariantMap timeStatus;
        timeStatus["app"] = QDateTime::currentDateTime().toTime_t();

        // WrStatus SavedTime in TRC module:
        // 0-CurTime, 1-RunStart, 2-SpillStart, 3-SpillEnd
        if (devSt.wrStatus.has_value()) {
            auto curDevTime = devSt.wrStatus->savedTimeCode(0);
            if(curDevTime.isNull())
                curDevTime = devSt.wrStatus->timeCode(); // timeCode - doesn't locked on readout
            timeStatus["WR"] = QVariant::fromValue(curDevTime)
                                   .value<QJsonObject>();
        }
        devRunStatus["time"] = timeStatus;
    }

    QVariantMap devStatus{{"status",QVariantMap{{"runTime", devRunStatus}}}};
    auto jsonObj = AbstractConfigConverter<int>::fromVariantMap(devStatus);

//    qInfo().noquote() << QJsonDocument(jsonObj).toJson(QJsonDocument::Indented);
    emit sendJsonBlock(getDeviceIndex(), QJsonDocument(jsonObj).toJson(QJsonDocument::Compact));
}
