#include "WaveformViewer.h"
#include "ui_WaveformViewer.h"

#include "base/gui_utils.h"
#include "base/AbstractConfigConverter.h"
#include "base/DeviceId.h"
#include "mregdev/qtmregdevice.h"

#include <QTimer>
#include <optional>

int getDeviceChNum(const DeviceIndex &index)
{

        switch (index.getDevId()) {
        case DEVICE_ID_TDC72VXS:
            return 72;
        case DEVICE_ID_ADC64:
        case DEVICE_ID_ADC64S2:
        case DEVICE_ID_ADC64VE:
        case DEVICE_ID_ADC64VE2:
        case DEVICE_ID_ADC64VE_XGE:
        case DEVICE_ID_ADC64ECAL:
        case DEVICE_ID_TDC64VHLE:
        case DEVICE_ID_ADC64WR:
            return 64;
        case DEVICE_ID_PHOS_FEC:
            return 64;
        case DEVICE_ID_TRC: // 16 phy + 16 possible aux
            return 32; // use this method for tdc ch num
        case DEVICE_ID_TQDC16VS_E:
        case DEVICE_ID_TQDC16VS_DIG:
            return 16;
        default:
            return -1;
        }
}

WaveformViewer::WaveformViewer(QWidget *parent)
    : QWidget(parent)
    , ui(new Ui::WaveformViewer)
    , refreshTimer(new QTimer(this))
{
    ui->setupUi(this);
    AbstractConfigConverter<DeviceIndex>::registerTypes();

    reset();
    connect(refreshTimer, &QTimer::timeout, [=](){this->refreshPlot();});
    refreshTimer->setSingleShot(true);
}

WaveformViewer::~WaveformViewer()
{
    refreshTimer->stop();
    delete ui;
}

void WaveformViewer::deviceDataUpdated(const DecodedData &data, bool writeDecodedEnabled)
{
    Q_UNUSED(writeDecodedEnabled)
    pendingData = data;
    hasPendingData = true;
    if(!refreshTimer->isActive())
        refreshTimer->start(500);
}

void WaveformViewer::refreshPlot(bool force)
{
    if(!hasPendingData && !force)
        return;
    if(hasPendingData) {
        lastData = pendingData;
        hasPendingData = false;
    }
    if(lastData.events.isEmpty())
        return;
    showWaveform(lastData.events.last());
    if(!force)
        refreshTimer->start(500);
}

void WaveformViewer::setWaveformViewerConfig(const WaveformViewerConfig &config)
{
    if(cfg == config)
        return;

    cfg = config;
    reset();
}

void WaveformViewer::reset()
{
    lastData.clear();
    pendingData.clear();
    hasPendingData = false;
    addedDevices.clear();
    hasSeveralDev = false;
    hasSeveralTypes = false;
    ui->widgetWaveform->clear();
    wfRange.reset();
    devMaxCh.clear();
    initGui();
}

void WaveformViewer::on_checkBoxEn_toggled(bool checked)
{
    cfg.en = checked;
    ui->comboBoxDev->setEnabled(cfg.en);
    ui->widgetWaveform->setEnabled(cfg.en);
    emit waveformViewerConfigChanged(cfg);
    refreshPlot(true);
}

void WaveformViewer::on_comboBoxDev_currentIndexChanged(int index)
{
    cfg.selectedDevice = ui->comboBoxDev->itemData(index).value<DeviceIndex>();
    updateMaxMinLimits();
    emit waveformViewerConfigChanged(cfg);
    refreshPlot(true);
}

void WaveformViewer::on_spinBoxChMin_valueChanged(int arg1)
{
    cfg.minCh = arg1;
    ui->spinBoxChMax->setMinimum(cfg.minCh);
    emit waveformViewerConfigChanged(cfg);
    refreshPlot(true);
}

void WaveformViewer::on_spinBoxChMax_valueChanged(int arg1)
{
    cfg.maxCh = arg1;
    ui->spinBoxChMin->setMaximum(cfg.maxCh);
    emit waveformViewerConfigChanged(cfg);
    refreshPlot(true);
}

void WaveformViewer::on_spinBoxSparseMarks_valueChanged(int p)
{
    cfg.sparsePeriod = p;
    emit waveformViewerConfigChanged(cfg);
    updateMarkers();
}

void WaveformViewer::initGui()
{
    RecursiveWidgetSignalBlocker blocker(this);
    ui->checkBoxEn->setChecked(cfg.en);
    ui->comboBoxDev->clear();
    ui->comboBoxDev->addItem(" All ");
    if(!cfg.selectedDevice.isNull()) {
        addDevice(cfg.selectedDevice);
        ui->comboBoxDev->setCurrentIndex(1);
    }

    updateMaxMinLimits();
    ui->spinBoxChMin->setValue(cfg.minCh);
    ui->spinBoxChMax->setValue(cfg.maxCh);

    ui->spinBoxSparseMarks->setValue(cfg.sparsePeriod);
}

void WaveformViewer::addDevice(const DeviceIndex &index)
{
    if(addedDevices.contains(index))
        return;

    auto indexVar = QVariant::fromValue(index);
    if(ui->comboBoxDev->findData(indexVar) == -1) {
        ui->comboBoxDev->addItem(index.getIdent(), indexVar);
        ui->comboBoxDev->model()->sort(0);
    }

    auto maxCh = getDeviceChNum(index)-1;
    if(index.getDevId() == DEVICE_ID_PHOS_FEC)
        maxCh = 64-1;

    addedDevices << index;
    DeviceIndex i;
    for(auto j : addedDevices) {
        if(i.isNull()) {
            i = j;
        } else {
            hasSeveralDev = true;
            if(i.getDevId() != j.getDevId())
                hasSeveralTypes = true;
        }
    }
    if(cfg.selectedDevice.isNull() || cfg.selectedDevice == index)
        updateMaxMinLimits();
}

void WaveformViewer::showWaveform(const CompleteEvent &ev)
{
    if(!cfg.en)
        return;

    for(auto devIt = ev.tqdcEv.cbegin(); devIt != ev.tqdcEv.cend(); ++devIt) {
        auto &index = devIt.key();
        if(addedDevices.contains(index)) continue;
        for(auto hitIt=devIt.value().hits.cbegin(); hitIt != devIt.value().hits.cend(); ++hitIt) {
            if(!hitIt.value().adcHits.isEmpty()) {
                addDevice(index);
                break;
            }
        }
    }
    for(auto devIt = ev.trcEv.cbegin(); devIt != ev.trcEv.cend(); ++devIt) {
        auto &index = devIt.key();
        if(addedDevices.contains(index)) continue;
        if(!devIt.value().wf.isEmpty()) {
            addDevice(index);
        }
    }

    WaveformDataAll mvv;
    int chOffset = 0;

    for(auto devIt = ev.trcEv.cbegin(); devIt != ev.trcEv.cend(); ++devIt) {
        if(!cfg.selectedDevice.isNull() && devIt.key() != cfg.selectedDevice)
            continue;
        const auto &chWf = devIt.value().wf;
        uint wfSize = static_cast<uint>(chWf.size());
        QMap<quint16, WaveformData> wfMap;
        for(uint step = 0; step<wfSize; ++step) {
            quint32 slice = chWf[step];
            for(quint16 ch=0; ch<32; ++ch){
                if(slice & (1<<ch)) {
                    auto it = wfMap.find(ch);
                    if(it != wfMap.end()) {
                        it.value().p.push_back({double(step), double(ch)});
                    } else {
                        wfMap[ch] = WaveformData(QString("TRC #%1").arg(ch).toStdString(), chOffset++);
                    }
                } else {
                    auto it = wfMap.find(ch);
                    if(it != wfMap.end()) {
                        mvv.push_back(*it);
                        wfMap.erase(it);
                    }
                }
            }
        }
        for(auto it = wfMap.cbegin(); it != wfMap.cend(); ++it) {
            mvv.push_back(*it);
        }
    }

    bool devMaxChUpdated = false;
    std::optional<QPair<double,double>> wfRangeNew;
    for(auto devIt = ev.tqdcEv.cbegin(); devIt != ev.tqdcEv.cend(); ++devIt) {
        if(!cfg.selectedDevice.isNull() && devIt.key() != cfg.selectedDevice)
            continue;
        const auto &chHits = devIt.value().hits;

        int curDevMaxCh = 0;
        for(auto chIt = chHits.cbegin(); chIt != chHits.cend(); ++chIt) {
            int ch = chIt.key();
            curDevMaxCh = std::max(curDevMaxCh, ch);
            if((cfg.minCh!=-1 && ch<cfg.minCh) ||
                    (cfg.maxCh!=-1 && ch>cfg.maxCh))
                continue;
            const auto &adcHits = chIt.value().adcHits;
            for(auto &adc : adcHits) {
                cfg.selectedDevice.getIdent();
                QString chName = QString::number(ch);
                if(hasSeveralDev)
                    chName.prepend(devIt.key().getSerialStr()+" ");
                if(hasSeveralTypes)
                    chName.prepend(devIt.key().getDeviceIdName()+" ");
//                        .arg((config.drawRelativeToBaseline && baseLine) ? "(b)": "")

                if(adc.adcTs) {
                    std::vector<double> vx = adc.wf;
                    for(uint i=0; i<adc.wf.size(); ++i)
                        vx[i] = i+adc.adcTs;
                    WaveformData wd(vx, adc.wf, chName.toStdString(), ch+chOffset);
                    mvv.push_back(wd);
                } else {
                    WaveformData wd(adc.wf, chName.toStdString(), ch+chOffset);
                    mvv.push_back(wd);
                }
                if(adc.wf.size()) {
                    if(wfRangeNew) {
                        wfRangeNew->first = qMin(wfRangeNew->first, double(adc.adcTs));
                        wfRangeNew->second = qMax(wfRangeNew->second, double(adc.adcTs+adc.wf.size()));
                    } else {
                        wfRangeNew = {adc.adcTs, adc.adcTs+adc.wf.size()};
                    }
                }
            }
        }
        {
            if(devMaxCh.value(devIt.key(), -1) < curDevMaxCh) {
                devMaxCh[devIt.key()] = curDevMaxCh;
                devMaxChUpdated = true;
            }
        }
        chOffset += curDevMaxCh;
    }

    if(devMaxChUpdated) {
        updateMaxMinLimits();
    }

//    lastEventData = data;
//    EventData tmpData = data;

//    int chOffset=0;
//    for (quint32 serial : tmpData.devices.keys()) {
//        if(selectedSerial && selectedSerial != serial)
//            continue;
//        DeviceEvent &de = tmpData.devices[serial];
//        if(!adcCalData.contains(de.serial)){
//            updateAdc64Setup(de.devId, de.serial);
//        }
//        QMap<int, int> &baseLines = adcCalData[de.serial];
//        std::map<int, AdcChEvent>::iterator it=de.adcEvents.begin();
//        while(it!=de.adcEvents.end()) {
//            AdcChEvent &adcEv = it->second;
//            if(adcEv.ch<chDisplayLow || adcEv.ch>chDisplayHi){
//                it++;
//                continue;
//            }
//            const int &baseLine = baseLines[adcEv.ch];
//            const QString chName = QString("%1:%2%3")
//                    .arg(de.serial, 4, 16, QChar('0'))
//                    .arg(it->first+1)
//                    .arg((config.drawRelativeToBaseline && baseLine) ? "(b)": "");
//            std::vector<double>::iterator wfIt=adcEv.wf.begin();
//            while(wfIt!=adcEv.wf.end()){
//                if(config.drawRelativeToBaseline)
//                    (*wfIt) -= baseLine;
//                ++wfIt;
//            }
//            WaveformData wd(adcEv.wf, chName.toStdString(), it->first+chOffset);
//            mvv.push_back(wd);
//            it++;
//        }
//        chOffset += 64;
//    }

//    lastWaveformTime.start();
//    ui->widgetWaveform->replot();

    std::optional<bool> wasAutoReplot;
    if(wfRange != wfRangeNew) {
        wasAutoReplot = ui->widgetWaveform->autoReplot();
        ui->widgetWaveform->setAutoReplot(false);
        wfRange = wfRangeNew;
        if(hasSparseMarkers())
            updateMarkers();
    }
    ui->widgetWaveform->setAllData(mvv);
    if(wasAutoReplot)
        ui->widgetWaveform->setAutoReplot(*wasAutoReplot);
    //    ui->widget->setData(mvv);
}

bool WaveformViewer::hasSparseMarkers() const
{
    return cfg.sparsePeriod > 1;
}

void WaveformViewer::updateMarkers()
{
    auto plot = ui->widgetWaveform;
    bool wasAutoReplot = plot->autoReplot();
    plot->setAutoReplot(false);

    for(auto m : sparseMarkers) {
        delete m;
    }
    sparseMarkers.clear();

    if(!hasSparseMarkers()) {
        plot->setAutoReplot(wasAutoReplot);
        return;
    }

    const static auto initMarker = [=]() {
        auto m = new QwtPlotMarker();
        m->setLineStyle(QwtPlotMarker::VLine);
        auto l = m->label();
        auto font = l.font();
        font.setItalic(true);
        l.setFont(font);
        m->setLabel(l);
        m->setLabelOrientation(Qt::Vertical);
        m->setLabelAlignment(Qt::AlignRight | Qt::AlignBottom);
        return m;
    };

    if(wfRange) {
        for(int i=wfRange->first/cfg.sparsePeriod; ; ++i) {
            auto pos = i*cfg.sparsePeriod;
            if(pos>wfRange->second)
                break;
            QwtPlotMarker* marker = initMarker();
            auto label = marker->label();
            label.setText(QString("sparse: %1").arg(pos));
            marker->setLabel(label);
            marker->setXValue(pos);
            marker->attach(plot);
            sparseMarkers.push_back(marker);
        }
    }
    plot->setAutoReplot(wasAutoReplot);
}

int WaveformViewer::getMaxCh(const DeviceIndex &i) const
{
    if(i.isNull()) {
        int ch = 0;
        for(auto it=devMaxCh.cbegin(); it!=devMaxCh.cend(); ++it)
            ch = std::max(ch, *it);
        return ch;
    } else {
        return devMaxCh.value(i, std::max(getDeviceChNum(i)-1, 0));
    }
}

void WaveformViewer::updateMaxMinLimits()
{
    ui->spinBoxChMin->setMaximum(cfg.maxCh);
    ui->spinBoxChMax->setMinimum(cfg.minCh);

    auto maxCh = qMax(qMax(cfg.minCh, cfg.maxCh), 0);
    maxCh = std::max(maxCh, getMaxCh(cfg.selectedDevice));
    if(ui->spinBoxChMax->maximum() == maxCh)
        return;

    ui->spinBoxChMax->setMaximum(maxCh);
}
