#include "BaseFsm.h"

#include <QDebug>
#include <QLoggingCategory>
#include <QTimer>

namespace {
Q_DECLARE_LOGGING_CATEGORY(fsm)
Q_LOGGING_CATEGORY(fsm, "afi.fsm")
} // namespace

BaseFsm::BaseFsm(QObject *parent) : QObject(parent)
{
    qRegisterMetaType<FsmState>();
    stateTime.start();
    auto t = new QTimer(this);
    connect(t, &QTimer::timeout, this, &BaseFsm::evaluate);
    t->start(100);
    setObjectName("BaseFSM");
    recreateInput(new BaseFsmInput);
}

bool BaseFsm::setRequestState(const FsmState &nextState)
{
    if (nextState == FsmState::Init || state == nextState)
        return false;
    if(input->requestState == nextState)
        return false;
    qDebug(fsm) << objectName() << "Request new state:" << nextState;
    input->requestState = nextState;
    input->requestTime.start();
    return evaluate();
}

bool BaseFsm::isTransitionAllowed(const FsmState &toState) const
{
    // Error allowed from any state
    if (toState == FsmState::Error)
        return true;

    // Unknown is prohibited
    if (toState == FsmState::Unknown)
        return false;

    static const QMap<QPair<FsmState, FsmState>, bool> map
        = {
            // basic Start-Stop sequence
            {{FsmState::Ready, FsmState::Starting}, true},
            {{FsmState::Starting, FsmState::Run}, true},
            {{FsmState::Run, FsmState::Stopping}, true},
            {{FsmState::Stopping, FsmState::Ready}, true},
            // TODO: other states
        };
    return map.value({getState(), toState}, false);
}

void BaseFsm::printFsmInput() const
{
    auto qinfo = qInfo().noquote().nospace();
    qinfo << QString("[%1] ").arg(objectName());
    qinfo << "curState:" << getState() <<"; ";
    printInputString(qinfo);
}

bool BaseFsm::evaluate()
{
    if(stateTimeout()) {
        auto dbg = qDebug(fsm) << objectName() << "Timeout error. Input:";
        printInputString(dbg);
        setState(FsmState::Error);
        return true;
    }

    if(requestTimeout()) {
        auto dbg = qDebug(fsm) << objectName() << "Timeout error. Input:";
        printInputString(dbg);
        setState(FsmState::Error);
        return true;
    }

    switch (state) {
    case FsmState::Init:
        setState(FsmState::Error);
        return true;
    default:;
    }

    if(hasRequestState() && isTransitionAllowed(getRequestState())) {
        acceptRequestedState();
        return true;
    }

    return false;
}

void BaseFsm::printInputString(QDebug dbg) const
{
    dbg << *getInput();
}

qint64 BaseFsm::stateTimeoutMs(const FsmState &state) const
{
    switch (state) {
    case FsmState::Starting: return 3000;
    case FsmState::Stopping: return 3000;
    default: return -1;
    }
}

qint64 BaseFsm::requestTimeoutMs() const
{
    return -1;
}

void BaseFsm::setState(const FsmState &nextState)
{
    if (nextState == state)
        return;
    const auto &message = QString("[%1] state change %2 to %3 (time: %4%5)")
            .arg(objectName())
            .arg(fsmStateToString(state))
            .arg(fsmStateToString(nextState))
            .arg(stateTime.elapsed())
            .arg(stateTimeout() ? " TIMEOUT" : "");
    if (nextState == FsmState::Error)
        qWarning(fsm) << message;
    else
        qDebug(fsm) << message;
    auto dbg = qDebug(fsm) << objectName() << " input:";
    printInputString(dbg);

    const auto oldState = state;
    state = nextState;
    stateTime.restart();
    if(state == FsmState::Error && input->requestState != FsmState::Init) {
        qWarning(fsm) << objectName() << "RejectOnError requested state " << input->requestState;
        input->clearRequests();
    }
    emit stateChanged(state, oldState);
}

void BaseFsm::acceptRequestedState()
{
    const auto nextState = input->requestState;
    input->clearRequests();
    setState(nextState);
}

void BaseFsm::recreateInput(BaseFsmInput *p)
{
    input = p;
}

bool BaseFsm::stateTimeout() const
{
    const auto &timeout = stateTimeoutMs(state);
    if(timeout == -1)
        return false;

    const bool timeouted = stateTime.hasExpired(timeout);
    if(!timeouted)
        return false;

    auto dbg = qDebug(fsm)
            << QString("%1 Got timeout at state: %2 reqSt: %3 (timeInState:%4)")
               .arg(objectName(), fsmStateToString(state), fsmStateToString(getRequestState()))
               .arg(stateTime.elapsed());
    printInputString(dbg);
    return true;
}

bool BaseFsm::requestTimeout() const
{
    const auto &timeout = requestTimeoutMs();
    if(timeout == -1)
        return false;

    const bool timeouted = input->requestTime.hasExpired(timeout);
    if(!timeouted)
        return false;

    auto dbg = qDebug(fsm)
            << QString("%1 Got timeout at state transition request %2->%3 (timeInRequest:%4)")
               .arg(objectName(), fsmStateToString(state), fsmStateToString(getRequestState()))
               .arg(input->requestTime.elapsed());
    printInputString(dbg);
    return true;
}

QDebug operator<<(QDebug dbg, const BaseFsmInput &r)
{
    dbg << "requestState:";
    if (r.requestState == FsmState::Init)
        dbg << "(none)";
    else
        dbg << r.requestState;

    dbg << "; ";
    return dbg;
}
