#include "DeviceMgrFsm.h"

#include "base/ConfigKey.h"
#include "base/DeviceIndexEnabled.h"

#include <QDebug>
#include <QLoggingCategory>

namespace {
Q_DECLARE_LOGGING_CATEGORY(fsm)
Q_LOGGING_CATEGORY(fsm, "afi.fsm.device")
}

DeviceMgrFsm::DeviceMgrFsm(QObject *parent):
    BaseFsm(parent)
{
    setObjectName("DevMgrFSM");
    recreateInput(new DeviceMgrFsmInput());
}

bool DeviceMgrFsm::isTransitionAllowed(const FsmState &toState) const
{
    switch (getState()) {
    case FsmState::Ready:
        if(toState == FsmState::Starting) return allDevicesWritten();
        break;
    case FsmState::Starting:
        if(toState == FsmState::Run) return allDevicesWritten();
        break;
    case FsmState::Stopping:
        if(toState == FsmState::Ready) return allDevicesWritten();
        break;
    default:;
    }
    return BaseFsm::isTransitionAllowed(toState);
}

void DeviceMgrFsm::setDeviceList(const DeviceIndexEnabledMap &newList)
{
    qDebug(fsm) << __PRETTY_FUNCTION__ << objectName() << newList;
    getInput()->devSet = newList.getEnabled();
    evaluate();
}

void DeviceMgrFsm::setMainSetupKey(const ConfigKey &key)
{
    qDebug(fsm) << __PRETTY_FUNCTION__ << objectName() << key;

    if(getInput()->setupKey == key)
        return;
    getInput()->setupKey = key;
    getInput()->lastSetupKeyUpdate.start();
    evaluate();
}

void DeviceMgrFsm::setDevSetupKey(const DeviceIndex &index, const ConfigKey &key)
{
    auto &keys = getInput()->deviceSetupKeys;
    auto it = keys.find(index);
    if(it == keys.end())
        keys[index] = key;
    else if(*it != key)
        *it = key;
    else
        return;

    qDebug(fsm) << __PRETTY_FUNCTION__ << objectName() << index << key;
    evaluate();
}

bool DeviceMgrFsm::evaluate()
{
    if(BaseFsm::evaluate())
        return true;

    switch (getState()) {
    case FsmState::Error:
        if (allDevicesWritten()) {
            setState(FsmState::Ready);
            return true;
        }
        break;
    case FsmState::Ready:
    case FsmState::Run:
        if(!allDevicesWritten() && setupKeyGuargExpired()) {
            qDebug(fsm) << objectName() << "Go to Error state (some dev cfg is invalid)";
            setState(FsmState::Error);
            return true;
        }
        break;
    default:;
    }

    return false;
}

void DeviceMgrFsm::printInputString(QDebug dbg) const
{
    dbg << *getInput();
}

qint64 DeviceMgrFsm::stateTimeoutMs(const FsmState &state) const
{
    switch (state) {
    case FsmState::Starting:
    case FsmState::Stopping:
        return DEVICE_CONFIG_WRITE_TIMEOUT;
    default:
        return BaseFsm::stateTimeoutMs(state);
    }
}

bool DeviceMgrFsm::allDevicesWritten() const
{
    return checkDevKey(getInput()->setupKey, getInput()->deviceSetupKeys);
}

bool DeviceMgrFsm::setupKeyGuargExpired() const
{
    const auto &t = getInput()->lastSetupKeyUpdate;
    return !t.isValid() || t.elapsed() > DEVICE_CONFIG_WRITE_TIMEOUT;
}

bool DeviceMgrFsm::checkDevKey(const ConfigKey &mainKey, const QMap<DeviceIndex, ConfigKey> &keyMap) const
{
    for(auto key : getInput()->devSet)
        if(keyMap.value(key) != mainKey)
            return false;
    return true;
}

QDebug operator<<(QDebug dbg, const DeviceMgrFsmInput &in)
{
    dbg << BaseFsmInput(in);

    dbg << "setupKey:" << in.setupKey << "; ";
    dbg << "devSet:" << in.devSet << "; ";
    dbg << "deviceSetupKeys:" << in.deviceSetupKeys << "; ";

    return dbg;
}
