/*
**    Copyright 2014-2020 Ivan Filippov
**    Copyright 2014-2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "DiscoverDialog_p.h"

#include "BaseDeviceTableModel.h"
#include "DeviceTableEntry.h"
#include "DeviceTableModel.h"
#include "DeviceTableView.h"
#include "DiscoverDialog.h"
#include "ManualAddDialog.h"
#include "ProxyChain.h"
#include "SubnetListWidget.h"
#include "ui_DiscoverDialog.h"

DiscoverDialogPrivate::DiscoverDialogPrivate(DiscoverDialog *q)
    :
      q_ptr(q),
      ui(new Ui::DiscoverDialog),
      modelFactory(new ProxyChain(q)),
      model(new DeviceTableModel(q))

{
    ui->setupUi(q);

    q->connect(q, &DiscoverDialog::lldpCacheUpdated,
            model, &DeviceTableModel::lldpCacheUpdated);

    auto modelFactory = new ProxyChain(q);

    q->connect(q, &DiscoverDialog::setDeviceIdFilter,
                     modelFactory, &ProxyChain::setDeviceIdFilter);

    q->connect(ui->listWidgetSubnet, &SubnetListWidget::setNetworkFilter,
            modelFactory, &ProxyChain::setNetworkFilter);
    q->connect(q, &DiscoverDialog::lldpCacheUpdated,
            ui->listWidgetSubnet, &SubnetListWidget::lldpCacheUpdated);

    q->connect(ui->listWidgetTypes, &DeviceIdListWidget::setDeviceIdFilter,
            modelFactory, &ProxyChain::setDeviceIdFilter);
    q->connect(q, &DiscoverDialog::lldpCacheUpdated,
            ui->listWidgetTypes, &DeviceIdListWidget::lldpCacheUpdated);

    // upper table
    auto availProxy = modelFactory->createProxyChain(
                model,
                ProxyChain::EnableDeviceIdFilter,
                ProxyChain::DisableDeviceFilter,
                ProxyChain::DynamicDeviceStateFilter,
                ProxyChain::EnableNetworkFilter
                );
    ui->tableViewAvailable->setModel(availProxy);
    ui->tableViewAvailable->hideColumn(BaseDeviceTableModel::COL_ENABLE);
    ui->tableViewAvailable->hideColumn(BaseDeviceTableModel::COL_ONLINE);
    ui->tableViewAvailable->hideColumn(BaseDeviceTableModel::COL_DYNAMIC);
    ui->tableViewAvailable->hideColumn(BaseDeviceTableModel::COL_DEVICE_INDEX);

    // lower table
    auto selectedProxy = modelFactory->createProxyChain(
                model,
                ProxyChain::DisableDeviceIdFilter,
                ProxyChain::DisableDeviceFilter,
                ProxyChain::StaticDeviceStateFilter,
                ProxyChain::DisableNetworkFilter
                );
    ui->tableViewSelected->setModel(selectedProxy);
    ui->tableViewSelected->hideColumn(BaseDeviceTableModel::COL_DYNAMIC);
    ui->tableViewSelected->hideColumn(BaseDeviceTableModel::COL_DEVICE_INDEX);

    on_checkBoxShowMac_toggled(ui->checkBoxShowMac->isChecked());

    QObject::connect(ui->pushButtonManualAdd, &QPushButton::clicked,
                     [this] { on_pushButtonManualAdd_clicked(); });

    QObject::connect(ui->pushButtonAdd, &QPushButton::clicked,
                     [this] { on_pushButtonAdd_clicked(); });

    QObject::connect(ui->pushButtonRemove, &QPushButton::clicked,
                     [this] { on_pushButtonRemove_clicked(); });

    QObject::connect(ui->checkBoxShowMac, &QCheckBox::toggled,
                     [this](bool enable) { on_checkBoxShowMac_toggled(enable); });

    QObject::connect(ui->listWidgetTypes, &QListWidget::itemSelectionChanged,
                     [this] { on_listWidgetTypes_itemSelectionChanged(); });

    QObject::connect(ui->tableViewAvailable, &QTableView::doubleClicked,
                     [this](const QModelIndex &i) { on_tableViewAvailable_doubleClicked(i); });

    QObject::connect(ui->tableViewSelected, &QTableView::doubleClicked,
                     [this](const QModelIndex &i) { on_tableViewSelected_doubleClicked(i); });

    QObject::connect(ui->buttonBox, &QDialogButtonBox::clicked,
                     [this](QAbstractButton *b) { on_buttonBox_clicked(b); });
}

DiscoverDialogPrivate::~DiscoverDialogPrivate()
{
    delete ui;
}

void DiscoverDialogPrivate::setDeviceList(const DeviceIndexEnabledMap &list)
{
    model->addStaticDevices(DeviceTable::staticFromMap(list));
    for(auto i : list)
        ui->listWidgetTypes->addDeviceId(i.index.getDevId());
}

void DiscoverDialogPrivate::apply()
{
    Q_Q(DiscoverDialog);
    const auto &result = DeviceTable::staticToMap(model->getDevices());
    emit q->deviceListChanged(DeviceIndexEnabledMap(result));
}

void DiscoverDialogPrivate::on_pushButtonManualAdd_clicked()
{
    Q_Q(DiscoverDialog);
    ManualAddDialog dialog(q);
    if(dialog.exec() == QDialog::Accepted) {
        DeviceIndex index(static_cast<DeviceId>(dialog.getDeviceId()), dialog.getSerial());
        DeviceIndexEnabledMap map;
        map[index] = DeviceIndexEnabled{index, true};
        model->addStaticDevices(DeviceTable::staticFromMap(map));

        const auto &result = DeviceTable::staticToMap(model->getDevices());
        emit q->deviceListChanged(DeviceIndexEnabledMap(result));
    }
}

void DiscoverDialogPrivate::on_pushButtonAdd_clicked()
{
    QItemSelectionModel *select = ui->tableViewAvailable->selectionModel();
    while (select->hasSelection())
        moveToStatic(select->selectedRows().first());
}

void DiscoverDialogPrivate::on_pushButtonRemove_clicked()
{
    QItemSelectionModel *select = ui->tableViewSelected->selectionModel();
    while (select->hasSelection())
        moveToDynamic(select->selectedRows().first());
}

void DiscoverDialogPrivate::on_checkBoxShowMac_toggled(bool checked)
{
    ui->tableViewAvailable->setColumnHidden(DeviceTableModel::COL_MAC, !checked);
    ui->tableViewSelected->setColumnHidden(DeviceTableModel::COL_MAC, !checked);
}

void DiscoverDialogPrivate::on_listWidgetTypes_itemSelectionChanged()
{
}

void DiscoverDialogPrivate::on_tableViewAvailable_doubleClicked(const QModelIndex &index)
{
    moveToStatic(index);
}

void DiscoverDialogPrivate::on_tableViewSelected_doubleClicked(const QModelIndex &index)
{
    if(index.column() == DeviceTableModel::COL_ENABLE)
        return;
    moveToDynamic(index);
}

void DiscoverDialogPrivate::on_buttonBox_clicked(QAbstractButton *button)
{
    switch(ui->buttonBox->standardButton(button)) {
    case QDialogButtonBox::Apply:
        apply();
        break;
    default:;
    }
}

DeviceIndex DiscoverDialogPrivate::deviceIndexFromAvail(const QModelIndex &index) const
{
    const QModelIndex &dataIndex = modelFactory->mapToDataSource(
                ui->tableViewAvailable->model(), index);
    return model->deviceIndexOf(dataIndex.row());
}

DeviceIndex DiscoverDialogPrivate::deviceIndexFromSelected(const QModelIndex &index) const
{
    const QModelIndex &dataIndex = modelFactory->mapToDataSource(
                ui->tableViewSelected->model(), index);
     return model->deviceIndexOf(dataIndex.row());
}

void DiscoverDialogPrivate::moveToStatic(const QModelIndex &index) const
{
    const QModelIndex &dataIndex = modelFactory->mapToDataSource(
                ui->tableViewAvailable->model(), index);
    model->setData(dataIndex.sibling(dataIndex.row(), DeviceTableModel::COL_DYNAMIC),
                   DeviceTableEntry::Static, Qt::DisplayRole);
}

void DiscoverDialogPrivate::moveToDynamic(const QModelIndex &index) const
{
    const QModelIndex &dataIndex = modelFactory->mapToDataSource(
                ui->tableViewSelected->model(), index);
    model->setData(dataIndex.sibling(dataIndex.row(), DeviceTableModel::COL_DYNAMIC),
                   DeviceTableEntry::DynamicDefault, Qt::DisplayRole);
    model->setData(dataIndex.sibling(dataIndex.row(), DeviceTableModel::COL_ENABLE),
                   DeviceTableEntry::EnableDefault, Qt::DisplayRole);
}

