/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef LLDPINFO_H
#define LLDPINFO_H

#include <QDeadlineTimer>
#include <QDebug>
#include <QMetaType>
#include <QPair>
#include <QString>

#include "mldp.h"
#include "base/HwipSocketAddress.h"

class DeviceIndex;

struct LldpduCommon {
    friend bool operator==(const LldpduCommon &l, const LldpduCommon &r) = default;
    // LLDP
    quint64 haddr = 0; // chassis address (MAC)
    QString portID; // id for the port from which this LLDPDU was transmitted: "2"
    quint32 ttl = 0; // Time-to-live, seconds
    quint32 ip_addr = 0; // management address
};

struct LldpduMed {
    friend bool operator==(const LldpduMed &l, const LldpduMed &r) = default;
    // LLDP-MED
    //    QString hw;
    QString hw_str; // 5 Hardware Revision, formatted: "1.0"
    QString fw_str; // 6 Firmware Revision, formatted: "2.0.19676"
    QString sn_str; // 8 Serial Number: "0000-0569-59F2"
    QString mn_str; // 9 Manufacturer Name: "JINR AFI"
    QString modelname_str; // 10 Model Name: "        U40VE-RC"
};

struct LldpduExt {
    friend bool operator==(const LldpduExt &l, const LldpduExt &r) = default;
    // OUI_AFI - 1
    quint16 device_id = 0;
    quint64 serial_id = 0;
    quint16 chassis_slot = 0;

    // OUI_AFI - 2
    HwipSocketAddress master;
    HwipSocketAddress mstream;
    quint64 master_mac = 0;
    quint64 mstream_mac = 0;
    bool master_locked = false;
    bool mstream_locked = false;
};

struct LldpInfo
{
    bool isNull() const { return !valid; }
    int msToExpire() const { return timer.remainingTime(); }
    bool hasExpired() const { return timer.hasExpired(); }
    static QString formatVersion(const QString &input);

    friend bool operator==(const LldpInfo &l, const LldpInfo &r) = default;

    DeviceIndex getDeviceIndex() const;
    QString getIdent() const;
    QString getSerialStr() const;
    QString getModelStr() const;
    QString getModelVerStr() const; // "BOARD-1.2"
    QString getFirmWareStr() const;
    QPair<int,int> getHwVer() const;
    static QPair<int,int> getHwVer(const QString &str);

    LldpduCommon common;
    LldpduMed med;
    LldpduExt ext;

    void startTimer();
    bool valid = false;
private:
    QDeadlineTimer timer;
};

Q_DECLARE_METATYPE(LldpInfo)

QDebug operator<<(QDebug dbg, const LldpInfo &r);

#endif // LLDPINFO_H
