#include "AdcSerDesCalibration.h"

#include "AdcDataTest/AdcDataTestModule.h"
#include "AdcDataTest/AdcDataTestStatus.h"
#include "AdcDeserCheckStatus.h"
#include "AdcSpiBaseModule.h"
#include "DeserializersControl/DeserializersControlModule.h"
#include "MlinkCsr/MlinkCsrModule.h"
#include "ModularMregDevice.h"

#include <QDebug>
#include <QRandomGenerator>

namespace {
QString getDelayPointStr(QPair<int, int> shift) {
    return "at point delay with shift " +
        QString::asprintf("clock:%+d data:%+d", shift.first, shift.second);
}
}

AdcSerDesCalibration::AdcSerDesCalibration(ModularMregDevice *parent)
    : QObject(parent)
    , dev (parent)
{
    qRegisterMetaType<AdcDeserCheckStatus>();
}

void AdcSerDesCalibration::init()
{
    adcDataTest = dev->getFirstModule<AdcDataTestModule>();
    adcSpi = dev->getAdcSpiModule();
    desCtrl = dev->getFirstModule<DeserializersControlModule>();
    mlinkCsrModule = dev->getFirstModule<MlinkCsrModule>();
    prbsMode = (adcSpi && adcDataTest)
        ? (adcSpi->supportPRBS() && adcDataTest->supportPRBS())
        : false;

    if (!adcDataTest || !adcSpi || !desCtrl || !mlinkCsrModule) {
        deinit();
        return;
    }
    auto bp = desCtrl->getBuildParams();
    if (!bp.has_value()) {
        qCritical().noquote() << QString("[%1] Empty deserializer's buildparams").arg(desCtrl->getDeviceIndex().getIdent());
        return;
    }
    desCnt = bp->desCnt;
    desCh = bp->chPerChip;
}

void AdcSerDesCalibration::deinit()
{
    adcDataTest = nullptr;
    adcSpi = nullptr;
    desCtrl = nullptr;
}

quint8 AdcSerDesCalibration::adcBits() const
{
    if (adcSpi)
        return adcSpi->getDataBits();
    return 16;
}

bool AdcSerDesCalibration::calibrate(bool hardMode, const QString &resDir)
{
    if (!adcSpi || !dev || !desCtrl)
        return {};

    hardPatternTestEnabled = hardMode;
    generateSpiPatternSet();

    if(prbsMode)
        adcSpi->setPRBSMode(true);
    else
        adcSpi->setCustomPatternMode(true);
    delayStatus = AdcClkStatus();

    // [desIndex][clkDelay][dataDelay] = true/false
    QMap<int, QMap<int, QMap<int, bool > > > resultsPattern;
    QMap<int, QMap<int, QMap<int, bool > > > resultsAdcStatus;
    QElapsedTimer timer, startTime;
    timer.start();
    startTime = timer;
    int goodPoints=0, badPoints=0;
    for(clkDelay=0; clkDelay<32; clkDelay++) {
        desCtrl->setClockDelay(clkDelay);
        for(dataDelay=0; dataDelay<32; dataDelay++) {
            const int &taskProgres = 100*(clkDelay*32+dataDelay+1)/(32*32);
            emit progressUpdated(taskProgres);
            if(timer.elapsed() > 5000) {
                qInfo("%s - %3d %% Good-%d Bad-%d", __func__, taskProgres,
                      goodPoints, badPoints);
                timer.start();
            }

            // Set delays
            desCtrl->setDataDelay(dataDelay);
            desCtrl->ResetDeser();
            desCtrl->ResetAdcStatus();
            QThread::msleep(1);

            // Get des lock status
            const auto desStOptional = desCtrl->readStatus();
            if (!desStOptional.has_value()) {
                qCritical().noquote() << QString("[%1] Error reading deserializer status").arg(desCtrl->getDeviceIndex().getIdent());
                return {};
            }

            // Get channels pattern test
            auto chPatternRes = patternTestByChannel();

            // Save des lock status
            QMap<int, bool> adcDesStatus;
            const auto desSt = desStOptional.value();
            for(int desId=0; desId<desCnt; ++desId) {
                adcDesStatus[desId] = desSt.isDesOk(desId);
                resultsAdcStatus[desId][clkDelay][dataDelay] = desSt.isDesOk(desId);
            }

            // Mask reults with ch. pattern test results
            bool allChRes=true;
            for(int ch : chPatternRes.keys()) {
                bool &v = chPatternRes[ch];
                allChRes &= v;
                resultsPattern[ch][clkDelay][dataDelay] = v;
                const int desId = ch/desCh;
                adcDesStatus[desId] &= v;
            }

            for(auto desId : adcDesStatus.keys()) {
                delayStatus.data[desId][clkDelay][dataDelay] = adcDesStatus[desId];
            }
            if(allChRes && desSt.isOk()) {
                ++goodPoints;
            } else {
                ++badPoints;
            }
        }
    }

    qInfo("%s passed in %lldms", __func__, startTime.elapsed());

    qInfo().noquote() << AdcClkStatus::getLegend();
    qInfo().noquote() << delayStatus.toString();
    delayStatus.temp = readTemp();
    delayStatus.fwStr = getFirmwareVersionStr();

    if(prbsMode)
        adcSpi->setPRBSMode(false);
    else
        adcSpi->setCustomPatternMode(false);

    saveResultsToFile(dev->getDeviceIndex(), delayStatus, resDir);
    return true;
}

bool AdcSerDesCalibration::checkAdcDeserConfig(const AdcSerDesDelayMap &delayBase, bool hardMode, bool printFaliedCh)
{
    if(!adcSpi || !dev || !desCtrl)
        return false;

    hardPatternTestEnabled = hardMode;
    generateSpiPatternSet();

    const QPair<int, int> NULL_POINT(0,0);
    const static QList< QPair<int,int> > checkPoints{
        {-1,-1}, {-1,1}, {1,-1}, {1,1},
        NULL_POINT};

    if(prbsMode)
        adcSpi->setPRBSMode(true);
    else
        adcSpi->setCustomPatternMode(true);

    AdcDeserCheckStatus checkSt;
    auto &failedDes = checkSt.failedDes;
    QMap<QPair<int,int>, QList<int> > failedChPerPoint;

    for(auto &shift : checkPoints) {
        if(dev->hasDeleteFlag())
            return {};
        if(!dev->isConnected())
            break;

        auto shiftedDelay = delayBase;
        for (auto &adcDelay : shiftedDelay.map) {
            adcDelay.clock += shift.first;
            adcDelay.data += shift.second;
        }

        if (!desCtrl->setSerDesDelay(shiftedDelay)) {
            qWarning() << dev->getIdent() << "Error writing deserializer delays in" << __func__;
            for (int id=0; id<desCnt; ++id)
                failedDes[shift] << id;
            continue;
        }

        auto desStOptional = desCtrl->readStatus();
        if (!desStOptional.has_value()) {
            qCritical().noquote() << QString("[%1] Error reading deserializer status").arg(desCtrl->getDeviceIndex().getIdent());
            return {};
        }

        for(int id=0; id<desCnt; ++id) {
            if(!desStOptional->isDesOk(id)) {
                failedDes[shift] << id;
                if(printFaliedCh) {
                    auto absPos = shiftedDelay.map.value(id);
                    qWarning() << __func__ << Qt::hex
                               << QString::asprintf("Des #%d lock fail at point clock/data:{%+d;%+d} ({%d;%d})",
                                                    id, shift.first, shift.second, absPos.clock, absPos.data);
                }
            }
        }

        const auto &chPatternRes = patternTestByChannel();
        for(auto it=chPatternRes.begin(); it!=chPatternRes.end(); ++it) {
            if(!(*it)) {
                if(printFaliedCh)
                    failedChPerPoint[shift] << it.key();
                const int &id = it.key()/desCh;
                failedDes[shift] << id;
            }
        }
    }

    if(prbsMode)
        adcSpi->setPRBSMode(false);
    else
        adcSpi->setCustomPatternMode(false);

    QSet<int> summaryFailedDesId;
    QSet<int> summaryFailedDesIdNull;
    for(auto &shift : checkPoints) {
        const auto &pointRes = failedDes.value(shift);
        const bool &ok = pointRes.isEmpty();
        (shift == NULL_POINT ? summaryFailedDesIdNull : summaryFailedDesId) += pointRes;
        if (ok)
            continue;
        QStringList failedAdcStr;
        for(auto id : pointRes) failedAdcStr << QString::number(id);
        failedAdcStr.sort();
        auto failedChList = failedChPerPoint[shift];
        auto dbg = (ok ? qDebug() : qWarning()).noquote()
                << dev->getIdent() << __func__ << (ok?"passed":"failed")
                << getDelayPointStr(shift)
                << (ok ? "" : QString("for adcId:{%1}").arg(failedAdcStr.join(", ")));

        if(printFaliedCh && !failedChList.isEmpty())
            dbg <<  "ch:" << failedChList;
    }
    emit adcDeserConfigChecked(dev->getDeviceIndex(), checkSt);

    auto ok = checkSt.getStatus() == AdcDeserCheckStatus::Status::Ok;
    (ok ? qDebug() : qWarning().noquote())
            << dev->getIdent() << __func__ << "finished; result:"
            << (ok ? "passed" : "failed");

    return true;
}

void AdcSerDesCalibration::generateSpiPatternSet()
{
    spiPatternSet.clear();

    auto rndGen = QRandomGenerator::securelySeeded();
    const int &rndLimit = hardPatternTestEnabled ? 30 : 1;
    while(spiPatternSet.size()<rndLimit) {
        quint16 val = static_cast<quint16>(rndGen.generate());
        if(val!=0)
            spiPatternSet.push_back(val);
    }
    if(hardPatternTestEnabled) {
        spiPatternSet.push_back(0x5555);
        spiPatternSet.push_back(0xAAAA);

        spiPatternSet.push_back(0x0A95);
        spiPatternSet.push_back(0x056a);
    }
}

QMap<int, bool> AdcSerDesCalibration::patternTestByChannel()
{
    if(prbsMode)
        return TestAdcCustomPatternByChannel(0);

    if (!adcSpi)
        return {};

    quint16 adc_pattern;
    QMap<int,bool> totalChRes;

    for(quint16 spi_pattern : spiPatternSet) {
        // Mask with adcBits
        spi_pattern &= (1<<adcBits())-1;
        // Arrange to 16bit data
        adc_pattern = (spi_pattern << (16-adcBits()));
        //qInfo() << hex << "adc_pattern:" << adc_pattern;

        adcSpi->setCustomPattern(spi_pattern);
        const auto &chRes = TestAdcCustomPatternByChannel(adc_pattern);
        if(totalChRes.isEmpty()) {
            totalChRes = chRes ;
        } else {
            for(int ch : chRes .keys())
                totalChRes[ch]&=chRes [ch];
        }
    }
    return totalChRes;
}

QMap<int, bool> AdcSerDesCalibration::TestAdcCustomPatternByChannel(const quint16 &pattern)
{
    if (!adcDataTest)
        return {};

    QMap<int, bool> test_ok;

    if (!(prbsMode ? adcDataTest->setPRBSEn(true)
                   : adcDataTest->setTestPatternAndEnable(pattern)))
        return {};

    QThread::msleep(hardPatternTestEnabled ? (prbsMode ? 50 : 2) : 1);
    AdcPatternStatus st;
    if (!adcDataTest->readPatternStatus(&st))
        return {};
    for (int ch=0; ch<getNumberOfChannels(); ch++) {
        bool chFail = (ch<64 ? (st.failedChMask_lo>>ch) : (st.failedChMask_hi>>(ch-64))) & 1;
        test_ok[ch] = !chFail;
    }
    return test_ok;
}

double AdcSerDesCalibration::readTemp() const
{
    if (!dev)
        return {};
    auto t = mlinkCsrModule->getStatus().temps;
    if (t.empty())
        return {};
    if (t.contains("1-Wire"))
        return t["1-Wire"];
    return t.first(); // FIXME: use correct temperature sensor
}

QString AdcSerDesCalibration::getFirmwareVersionStr() const
{
    return dev ? dev->getFirmwareVersionStr() : QString();
}

void AdcSerDesCalibration::saveResultsToFile(const DeviceIndex &index, const AdcClkStatus &delayStatus, const QString &resDir)
{
    if(index.isNull())
        return;

    if(resDir.isEmpty())
        return;

    QFile f;
    int loopCnt = 0;
    do {
        QStringList fileNameParts{"devSerDesDelay"};
        fileNameParts << index.getIdent().replace(' ', '_');
        fileNameParts << QString("%1").arg(delayStatus.temp, 2, 'f', 0, QChar('0'));
        if(loopCnt)
            fileNameParts << QString::number(loopCnt);

        const auto &filePath = QString("%1/%2.txt")
                .arg(resDir).arg(fileNameParts.join('_'));

        f.setFileName(filePath);

        if(++loopCnt > 100) {
            qWarning() << "Failed to save adcClkDelay; last file attempt:"
                       << filePath;
            break;
        }
    }while(!f.open(QIODevice::WriteOnly|QIODevice::NewOnly));

    if(!f.isOpen())
        return;

    QTextStream stream(&f);
    stream << QString("=== %1 (f/w = %2; temp = %3) ===")
              .arg(index.getIdent())
              .arg(delayStatus.fwStr)
              .arg(delayStatus.temp, 0, 'f', 1);
    stream << "\n";
    stream << delayStatus.toString();
    f.close();
}
