#include "FeLinkTreeViewFancy.h"

#include <QHeaderView>
#include <QMenu>
#include <QTimer>
#include <QSettings>

#include "FeLinkTreeModel.h"

namespace {
const auto k_FeLinkTreeViewFancy = "FeLinkTreeViewFancy";
const auto k_visible = "visible";
} // namespace

FeLinkTreeViewFancy::FeLinkTreeViewFancy(QWidget *parent)
    : FeLinkTreeView(parent)
    , headerPopupMenu(new QMenu(this))
{
    header()->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(header(), &QHeaderView::customContextMenuRequested,
            this, &FeLinkTreeViewFancy::horizontalHeaderPopup);

    QTimer::singleShot(10, this, &FeLinkTreeViewFancy::deferredInit);
}

void FeLinkTreeViewFancy::setModel(QAbstractItemModel *model)
{
    FeLinkTreeView::setModel(model);
    setDefaultColumnState();
}

void FeLinkTreeViewFancy::setDefaultColumnState()
{
    const auto &states = defaultColumnState();
    for (auto it = states.begin(); it != states.end(); ++it) {
        int column = it.key();
        bool shown = it.value();
        FeLinkTreeView::setColumnHidden(column, !shown);
    }
}

void FeLinkTreeViewFancy::hideColumn(int column)
{
    FeLinkTreeView::hideColumn(column);
    saveColumnState(column, false);
}

void FeLinkTreeViewFancy::showColumn(int column)
{
    FeLinkTreeView::showColumn(column);
    saveColumnState(column, true);
}

void FeLinkTreeViewFancy::horizontalHeaderPopup(const QPoint &pos)
{
    auto index = indexAt(pos);
    if (! index.isValid())
        return;

    headerPopupMenu->clear();
    auto insertFunc = [=](const QString &name, void (FeLinkTreeViewFancy::*pt2ConstMember)(), QVariant data = {}) {
        auto a = headerPopupMenu->addAction(name, this, pt2ConstMember);
        if(data.isValid())
            a->setData(data);
    };

    insertFunc("Hide column", &FeLinkTreeViewFancy::hideColumnAction, index.column());
    insertFunc("Show all columns", &FeLinkTreeViewFancy::showAllColumnsSetAction);
    insertFunc("Column set default", &FeLinkTreeViewFancy::setColumnSetAction,
               QVariant::fromValue(getColumnSetDefault()));

    headerPopupMenu->addSeparator();

    for (int i=0; i<header()->count(); i++) {
        auto hidden = isColumnHidden(i);
        auto data = header()->model()->headerData(i, Qt::Horizontal);
        auto str = data.toString();
        if (str.isEmpty())
            continue;
        auto action = headerPopupMenu->addAction(str, this, &FeLinkTreeViewFancy::toggleColumnAction);
        action->setCheckable(true);
        action->setChecked(!hidden);
        action->setData(i);
    }
    headerPopupMenu->popup(mapToGlobal(pos));
}

void FeLinkTreeViewFancy::hideColumnAction()
{
    auto action = qobject_cast<QAction *>(sender());
    if (!action) return;
    int column = action->data().toInt();
    hideColumn(column);
}

void FeLinkTreeViewFancy::showAllColumnsSetAction()
{
    for(auto c=0; c < model()->columnCount(); ++c)
        showColumn(c);
}

void FeLinkTreeViewFancy::setColumnSetAction()
{
    auto action = qobject_cast<QAction *>(sender());
    if (!action) return;
    auto columns = action->data().value<QSet<int>>();
    for(auto c=0; c < model()->columnCount(); ++c) {
        auto f = columns.contains(c) ? &FeLinkTreeViewFancy::showColumn
                                     : &FeLinkTreeViewFancy::hideColumn;
        (*this.*f)(c);
    }
}

void FeLinkTreeViewFancy::showColumnAction()
{
    auto action = qobject_cast<QAction *>(sender());
    if (!action) return;
    int column = action->data().toInt();
    showColumn(column);
}

void FeLinkTreeViewFancy::toggleColumnAction()
{
    auto action = qobject_cast<QAction *>(sender());
    if (!action) return;
    int column = action->data().toInt();
    if (isColumnHidden(column))
        showColumn(column);
    else
        hideColumn(column);
}

void FeLinkTreeViewFancy::deferredInit()
{
    setDefaultColumnState();

    auto myModel = model();
    if(!myModel)
        return;

    for (int i=0; i<myModel->columnCount(); i++) {
        auto var = loadColumnState(i);
        if (!var.canConvert<bool>())
            continue;
        FeLinkTreeView::setColumnHidden(i, !var.toBool());
    }
}

void FeLinkTreeViewFancy::saveColumnState(int column, bool shown)
{
    auto key = columnSettingsKey(column);
    if (key.isEmpty())
        return;
    QSettings settings;
    settings.setValue(key, shown);
}

QVariant FeLinkTreeViewFancy::loadColumnState(int column)
{
    const QSettings settings;
    return settings.value(columnSettingsKey(column));
}

QString FeLinkTreeViewFancy::columnSettingsKey(int column)
{
    const auto m = model();
    if (!m)
        return {};
    QRegularExpression rx("[^A-za-z0-9]");
    QString modelName = m->objectName();
    QString columnName = m->headerData(column, Qt::Horizontal).toString();
    modelName.replace(rx, "_");
    columnName.replace(rx, "_");
    QString key = QString(k_FeLinkTreeViewFancy) + "/" + modelName + "/" + columnName + "/" + k_visible;
    return key;
}

QMap<int, bool> FeLinkTreeViewFancy::defaultColumnState() const
{
    auto m = qobject_cast<FeLinkTreeModel*>(model());
    QMap<int, bool> ret;
    if(m) {
        for(auto c : {FeLinkTreeColumns::COL_RESET_PHY_BUT,
                       FeLinkTreeColumns::COL_REMOTE_RESET,
                       FeLinkTreeColumns::COL_RECV_TRIG,
                       FeLinkTreeColumns::COL_IS_MASTER,
                       FeLinkTreeColumns::COL_REMOTE_ID,
                       FeLinkTreeColumns::COL_COMMUNICATION_REMOTE_ENV,
                       FeLinkTreeColumns::COL_SFP_STATUS,
                       FeLinkTreeColumns::COL_PHASE_ADJ_EN,
                       FeLinkTreeColumns::COL_RESET_SOFT_PLL,
                       FeLinkTreeColumns::COL_REMOTE_VER}) {
            ret[static_cast<int>(c)] = false;
        }
    }
    return ret;
}

QSet<int> FeLinkTreeViewFancy::getColumnSetDefault() const
{
    const auto &defMap = defaultColumnState();
    QSet<int> set;
    for(auto c=0; c < model()->columnCount(); ++c) {
        if(defMap.value(c, true))
            set << c;
    }
    return set;
}
