#include "SetFirFilterDialog.h"
#include "ui_SetFirFilterDialog.h"

#include "widgets/FileDialog.h"

#include <QAbstractButton>
#include <QDebug>
#include <QFile>
#include <QFileDialog>
#include <QLoggingCategory>
#include <QTableWidgetItem>
#include <QSettings>
#include <limits>

namespace {
Q_DECLARE_LOGGING_CATEGORY(fir)
Q_LOGGING_CATEGORY(fir, "afi.fir")

enum {
    COL_VAL,
    COL_DOUBLE,
    COL_TOTAL
};
}

SetFirFilterDialog::SetFirFilterDialog(const FirFilterConfig &cfg, QWidget *parent)
    : QDialog(parent)
    , ui(new Ui::SetFirFilterDialog)
{
    ui->setupUi(this);
    ui->tableWidget->setColumnCount(COL_TOTAL);
    ui->tableWidget->setHorizontalHeaderLabels({"Value", "Coef"});
    setFirFilterConfig(cfg);
    updateGui();
}

SetFirFilterDialog::~SetFirFilterDialog()
{
    delete ui;
}

void SetFirFilterDialog::setSizeLimit(int size)
{
    ui->spinBoxSize->setMaximum(size);
}

void SetFirFilterDialog::setFirFilterConfig(const FirFilterConfig &newConfig)
{
    if(config == newConfig)
        return;
    config = newConfig;
    updateGui();
}

void SetFirFilterDialog::on_groupBox_toggled(bool en)
{
    if(config.en == en)
        return;
    config.en = en;
    emit firFilterConfigChanged(config);
}

void SetFirFilterDialog::on_spinBoxSize_valueChanged(int size)
{
    config.coef.resize(size);
    emit firFilterConfigChanged(config);
    updateGui();
}

void SetFirFilterDialog::on_tableWidget_itemChanged(QTableWidgetItem *item)
{
    item->row();
}

void SetFirFilterDialog::on_buttonBox_clicked(QAbstractButton *button)
{
    if(ui->buttonBox->standardButton(button) != QDialogButtonBox::RestoreDefaults)
        return;
    setFirFilterConfig({});
}

void SetFirFilterDialog::coefChanged(int val)
{
    for(int r=ui->tableWidget->rowCount()-1; r>=0; --r) {
        if(ui->tableWidget->cellWidget(r, COL_VAL) != sender())
            continue;

        if(config.coef[r] != val) {
            config.coef[r] = val;
            emit firFilterConfigChanged(config);
        }
        updateDouble(r);
        break;
    }
}

void SetFirFilterDialog::updateGui()
{
    int oldSize = ui->tableWidget->rowCount();
    auto size = config.coef.size();

    ui->groupBox->setChecked(config.en);

    ui->spinBoxSize->blockSignals(true);
    ui->spinBoxSize->setValue(size);
    ui->spinBoxSize->blockSignals(false);

    ui->tableWidget->setRowCount(size);
    for(int r=oldSize; r<size; ++r) {
        auto sb = new QSpinBox(ui->tableWidget);
        sb->setMinimum(std::numeric_limits<qint16>::min());
        sb->setMaximum(std::numeric_limits<qint16>::max());
        sb->setValue(config.coef[r]);
        ui->tableWidget->setCellWidget(r, COL_VAL, sb);
        connect(sb, QOverload<int>::of(&QSpinBox::valueChanged),
                this, &SetFirFilterDialog::coefChanged);
        ui->tableWidget->setItem(r, COL_DOUBLE, new QTableWidgetItem());
        updateDouble(r);
    }
    for(int r=0; r<size; ++r) {
        auto sb = qobject_cast<QSpinBox*>(ui->tableWidget->cellWidget(r, COL_VAL));
        if(!sb) {
            Q_ASSERT(sb);
            continue;
        }
        if(sb->value() != config.coef[r])
            sb->setValue(config.coef[r]);
    }
}

void SetFirFilterDialog::updateDouble(int r)
{
    auto val = config.coef.value(r);
    double c = double(val)/std::numeric_limits<qint16>::max();
    ui->tableWidget->item(r, COL_DOUBLE)->setText(QString("%1").arg(c, 0, 'f', 3));
}

void SetFirFilterDialog::on_pushButton_clicked()
{
    QSettings s;
    static const char *key = "SetFirFilter/csvFile";
    auto oldPath = s.value(key).toString();
    QString path = FileDialog::getOpenFileName(this, "Set FIR from csv-file", oldPath );
    if(path.isNull())
        return;
    if(path != oldPath)
        s.setValue(key, path);


    QFile file(path);
    if (!file.open(QIODevice::ReadOnly)) {
        qWarning(fir) << file.errorString();
        return;
    }

    while (!file.atEnd()) {
        QByteArray line = file.readLine();
        auto lineStr = QString::fromLatin1(line).trimmed();
        if(lineStr.isEmpty()) {
            qWarning(fir) << QString("refuse empty line '%1'")
                          .arg(lineStr);
            continue;
        }

        bool ok = true;
        FirPreset coef;
        for(auto i : lineStr.split(',')) {
            auto var = i.toInt(&ok);
            if(!ok)
                break;
            if(var > std::numeric_limits<qint16>::max() ||
                    var < std::numeric_limits<qint16>::min()) {
                qWarning(fir) << QString("refuse line '%1': value %2 is out of range")
                              .arg(lineStr).arg(var);
                ok = false;
                break;
            }
            coef.push_back(var);
        }
        if(ok) {
            FirFilterConfig cfg = config;
            cfg.coef = coef;
            emit firFilterConfigChanged(cfg);
            setFirFilterConfig(cfg);
            return;
        }
    }
}

