/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "TrigCtrlStatWidget.h"
#include "ui_TrigCtrlStatWidget.h"

#include "TrigCtrlStat/TrigCtrlStatStatus.h"
#include "base/DeviceIndexEnabled.h"
#include "base/Globals.h"
#include "base/gui_utils.h"

namespace {
inline const double minTimerFreq = 1.0;
inline const double maxTimerFreq = 100e6;
inline const double scaleRandomMinPeriod = 1e6;
inline const double scaleDeadTime = 1e6;
} // namespace

TrigCtrlStatWidget::TrigCtrlStatWidget(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::TrigCtrlStatWidget)
{
    ui->setupUi(this);

    sourceMap = {
        {ui->checkBoxTrigDigitalThr, TrigCtrlStatConfig::TrigSourceDigThr},
        {ui->checkBoxTrigAnalogThr, TrigCtrlStatConfig::TrigSourceAnalogThr},
        {ui->checkBoxTrigVxsPulse, TrigCtrlStatConfig::TrigSourceVxsPulse},
        {ui->checkBoxTrigVxsFelink, TrigCtrlStatConfig::TrigSourceFelinkVxs},
        {ui->checkBoxTrigFelink, TrigCtrlStatConfig::TrigSourceFelinkFp},
        {ui->checkBoxTrigIn1, TrigCtrlStatConfig::TrigSourceExtPulse},
        {ui->checkBoxTrigIn2, TrigCtrlStatConfig::TrigSourceExtPulse2},
        {ui->checkBoxTrigIn3, TrigCtrlStatConfig::TrigSourceExtPulse3},
        {ui->checkBoxTrigIn4, TrigCtrlStatConfig::TrigSourceExtPulse4},
        {ui->checkBoxTrigPeriod, TrigCtrlStatConfig::TrigSourcePeriodic},
        {ui->checkBoxTrigRandom, TrigCtrlStatConfig::TrigSourceRandom}
    };
    for (auto it = sourceMap.cbegin(); it != sourceMap.cend(); ++it) {
        auto *button = it.key();
        auto index = it.value();
        connect(button, &QAbstractButton::toggled,
                this, &TrigCtrlStatWidget::onTrigSourceButtonToggle);
        button->setChecked(cfg.isEnabled(index));
    }
    restoreGuiFromConfig();
}

TrigCtrlStatWidget::~TrigCtrlStatWidget()
{
    delete ui;
}

void TrigCtrlStatWidget::updateExtTrigGroup()
{
//    bool moreThanOne = config->trigInMoreThatOne();
    bool moreThanOne = false;
    ui->checkBoxTrigIn1->setText(QString("External TTL")
                                .append(moreThanOne ? " 1" : ""));
    ui->checkBoxTrigIn2->setVisible(moreThanOne);
    ui->checkBoxTrigIn3->setVisible(moreThanOne);
    ui->checkBoxTrigIn4->setVisible(moreThanOne);
}

void TrigCtrlStatWidget::restoreGuiFromConfig()
{
    RecursiveWidgetSignalBlocker blocker(this);

    for (auto it = sourceMap.cbegin(); it != sourceMap.cend(); ++it) {
        auto *button = it.key();
        auto index = it.value();
        button->setChecked(cfg.isEnabled(index));
    }

    ui->spinBoxAnalogThrDelay->setRange(1, 0xFFFF); // min should be 0; 1 - due to fw bug
    ui->spinBoxAnalogThrDelay->setValue(cfg.analogThrDelay);
    ui->spinBoxDigitalThrDelay->setRange(1, 0xFFFF); // min should be 0; 1 - due to fw bug
    ui->spinBoxDigitalThrDelay->setValue(cfg.digThrDelay);

    ui->doubleSpinBoxTimerFreq->setRange(minTimerFreq, maxTimerFreq);
    ui->doubleSpinBoxTimerFreq->setValue(cfg.trigTimerFreq);

    ui->doubleSpinBoxRandomFreq->setRange(minTimerFreq, maxTimerFreq);
    ui->doubleSpinBoxRandomFreq->setValue(cfg.randomPulserFreq);

    ui->doubleSpinBoxRandomMinPeriod->setRange(0, 1e6);
    ui->doubleSpinBoxRandomMinPeriod->setValue(cfg.randomPulserMinPeriod * scaleRandomMinPeriod);

    ui->doubleSpinBoxDeadTime->setRange(0, 1e6);
    ui->doubleSpinBoxDeadTime->setValue(cfg.programmedDeadTime * scaleDeadTime);

    updateExtTrigGroup();
    updateEnableState();
    updateLabels();
}

void TrigCtrlStatWidget::updateEnableState()
{
    auto isPeriodicEn =cfg.isEnabled(TrigCtrlStatConfig::TrigSourcePeriodic);
    auto isRandomEn =cfg.isEnabled(TrigCtrlStatConfig::TrigSourceRandom);
    // ui->doubleSpinBoxTimerFreq->setEnabled(isPeriodicEn);
    // ui->doubleSpinBoxRandomFreq->setEnabled(isRandomEn);
    ui->labelTrigSourceRandomMinPeriod->setEnabled(isRandomEn);
    ui->doubleSpinBoxRandomMinPeriod->setEnabled(isRandomEn);

    for (auto it = sourceMap.cbegin(); it != sourceMap.cend(); ++it) {
        auto *button = it.key();
        auto index = it.value();
        button->setVisible((1u << index) & trigSourceCaps);
    }
    bool analogThrCap = (1u << (TrigCtrlStatConfig::TrigSourceAnalogThr)) & trigSourceCaps;
    bool digThrCap = (1u << (TrigCtrlStatConfig::TrigSourceDigThr)) & trigSourceCaps;
    ui->labelAnalogThrDelay->setVisible(analogThrCap);
    ui->labelDigitalThrDelay->setVisible(digThrCap);
    ui->spinBoxAnalogThrDelay->setVisible(analogThrCap);
    ui->spinBoxDigitalThrDelay->setVisible(digThrCap);
    ui->doubleSpinBoxTimerFreq->setVisible((1u << (TrigCtrlStatConfig::TrigSourcePeriodic)) & trigSourceCaps);
    ui->doubleSpinBoxRandomFreq->setVisible((1u << (TrigCtrlStatConfig::TrigSourceRandom)) & trigSourceCaps);
}

void TrigCtrlStatWidget::updateLabels()
{
    auto f = [](QLabel *l, quint16 delay, std::optional<double> step) {
        l->setText(step ? QString("%L1 ns").arg(round(delay * (*step) / 1e-9)) : "n/a");
        l->setToolTip(step ? "" : "Can't define TrigCtrlStatModule clock");
        auto p = l->palette();
        p.setColor(l->foregroundRole(), step ? QColor() : QColor(Qt::red));
        l->setPalette(p);
    };
    f(ui->labelAnalogThrDelay, cfg.analogThrDelay, trigModClockStep);
    f(ui->labelDigitalThrDelay, cfg.digThrDelay, trigModClockStep);
}

void TrigCtrlStatWidget::updateTrigSourceCaps()
{
    quint32 caps = ~0u;
    for(auto i : devSet) {
        caps &= devBuildParams.value(i).sourceCaps;
    }
    if(trigSourceCaps != caps) {
        trigSourceCaps = caps;
        updateEnableState();
    }
}

void TrigCtrlStatWidget::updateTrigModClock()
{
    auto clockStep = getClockStep();

    if(trigModClockStep != clockStep) {
        trigModClockStep = clockStep;
        updateLabels();
    }
}

std::optional<double> TrigCtrlStatWidget::getClockStep() const
{
    std::optional<quint32> clock;
    for(auto i : devSet) {
        auto it = devBuildParams.constFind(i);
        if(it != devBuildParams.cend()) {
            if(clock) {
                if(*clock != it->timerClock)
                    return {};
            } else {
                clock = it->timerClock;
            }
        }
    }
    if(clock) {
        return 1./(*clock);
    }
    return {};
}

void TrigCtrlStatWidget::setTrigCtrlStatConfig(const TrigCtrlStatConfig &newConfig)
{
    if(cfg == newConfig)
        return;
    cfg = newConfig;

    restoreGuiFromConfig();
}

void TrigCtrlStatWidget::setTrigCtrlBp(const DeviceIndex &index, const TrigCtrlBuildParams &bp)
{
    devBuildParams[index] = bp;
    updateTrigSourceCaps();
    updateTrigModClock();
}

void TrigCtrlStatWidget::setDeviceList(const DeviceIndexEnabledMap &map)
{
    auto set = map.getEnabled();
    if(devSet == set)
        return;
    devSet = set;
    updateTrigSourceCaps();
    updateTrigModClock();
}

void TrigCtrlStatWidget::onFsmStateChange(const FsmState &state)
{
    bool en = state==FsmState::Ready || state==FsmState::Error
            || Globals::getDebugMode();
    setEnabled(en);
}

void TrigCtrlStatWidget::on_doubleSpinBoxTimerFreq_valueChanged(double arg1)
{
    cfg.trigTimerFreq = arg1;
    emit trigCtrlStatConfigChanged(cfg);
}

void TrigCtrlStatWidget::on_doubleSpinBoxRandomFreq_valueChanged(double arg1)
{
    cfg.randomPulserFreq = arg1;
    emit trigCtrlStatConfigChanged(cfg);
}

void TrigCtrlStatWidget::on_doubleSpinBoxRandomMinPeriod_valueChanged(double arg1)
{
    cfg.randomPulserMinPeriod = arg1 / scaleRandomMinPeriod;
    emit trigCtrlStatConfigChanged(cfg);
}

void TrigCtrlStatWidget::on_doubleSpinBoxDeadTime_valueChanged(double arg1)
{
    cfg.programmedDeadTime = arg1 / scaleDeadTime;
    emit trigCtrlStatConfigChanged(cfg);
}

void TrigCtrlStatWidget::onTrigSourceButtonToggle(bool checked)
{
    for (auto it = sourceMap.cbegin(); it != sourceMap.cend(); ++it) {
        auto *button = it.key();
        if (button == sender()) {
            auto index = it.value();
            cfg.setEnabled(index, checked);
            updateEnableState();
            emit trigCtrlStatConfigChanged(cfg);
            return;
        }
    }
}

void TrigCtrlStatWidget::on_spinBoxAnalogThrDelay_valueChanged(int arg1)
{
    cfg.analogThrDelay = arg1;
    updateLabels();
    emit trigCtrlStatConfigChanged(cfg);
}

void TrigCtrlStatWidget::on_spinBoxDigitalThrDelay_valueChanged(int arg1)
{
    cfg.digThrDelay = arg1;
    updateLabels();
    emit trigCtrlStatConfigChanged(cfg);
}
