#include "DataBase.h"

#include <iostream>

#include <QDebug>
#include <QFileInfo>
#include <QHostInfo>
#include <QMessageBox>
#include <QRegularExpression>
#include <QSettings>
#include <QSqlError>
#include <QSqlQuery>

namespace {
constexpr const auto testHost = "afi-db.he.jinr.ru";
constexpr const auto testDatabase = "equipment";
constexpr const auto testUser = "DevFlashProg";
constexpr const auto testPasswd = "RK9JxnoWHUU7KlvO";

QString getSerialStr(quint64 serial)
{
    QString serialStr = QString("%1-%2")
                            .arg((serial >> 16) & 0xFFFFu, 4, 16, QChar('0'))
                            .arg(serial & 0xFFFFu, 4, 16, QChar('0'));
    if (serial >> 32) {
        serialStr.prepend(QString("%1-").arg((serial >> 32) & 0xFFFFu, 4, 16, QChar('0')));
    }
    if (serial >> 48) {
        serialStr.prepend(QString("%1-").arg((serial >> 48) & 0xFFFFu, 4, 16, QChar('0')));
    }
    return serialStr.toUpper();
}
} // namespace

DataBase::DataBase(QWidget *p) : QObject(p)
{
    if (!QSqlDatabase::drivers().contains("QMYSQL", Qt::CaseInsensitive)) {
        QMessageBox::critical(
            p,
            "QMYSQL not loaded",
            "Database driver QMYSQL not loaded.<br/>Add missing dll (qsqlmysql4.dll).");
        emit setStatus("Database driver QMYSQL not loaded", DB_STATUS_ERROR);
        return;
    }

    QSettings settings;
    foreach (QString section, settings.value("MainWindow/databases").toStringList()) {
        if (section.isEmpty())
            continue;

        settings.beginGroup(section);
        if (settings.contains("DbName") && settings.contains("login")) {
            QSqlDatabase db = QSqlDatabase::addDatabase("QMYSQL", section);
            db.setHostName(settings.value("HostName").toString());
            db.setDatabaseName(settings.value("DbName").toString());
            db.setUserName(settings.value("login").toString());
            db.setPassword(settings.value("passwd").toString());
            db.setConnectOptions("MYSQL_OPT_RECONNECT=1");
            if (!db.isValid()) {
                qCritical() << "Database driver not loaded.";
            }
            if (db.databaseName().isEmpty() || db.userName().isEmpty())
                QSqlDatabase::removeDatabase(section);
        }
        settings.endGroup();
    }

    if (!getConnection().isOpen()) {
        QString section = "test";
        QSqlDatabase db = QSqlDatabase::addDatabase("QMYSQL", section);
        db.setHostName(testHost);
        db.setDatabaseName(testDatabase);
        db.setUserName(testUser);
        db.setPassword(testPasswd);
        db.setConnectOptions("MYSQL_OPT_RECONNECT=1");
        if (!db.isValid()) {
            qCritical() << "Database driver not loaded.";
        }
        if (db.databaseName().isEmpty() || db.userName().isEmpty())
            QSqlDatabase::removeDatabase(section);
        selectedConnection = section;
    }

    emit setStatus("Not connected", DB_STATUS_OK);
}

QSqlDatabase DataBase::getConnection()
{
    // Check previous connection
    if (!selectedConnection.isEmpty()) {
        QSqlDatabase db = QSqlDatabase::database(selectedConnection);
        if (db.isOpen())
            return db;
    }

    // Search other connections
    foreach (QString connection, QSqlDatabase::connectionNames()) {
        if (connection == selectedConnection)
            continue;
        QSqlDatabase db = QSqlDatabase::database(connection);
        if (db.isOpen()) {
            selectedConnection = connection;
            return db;
        }
    }

    // No database available
    selectedConnection.clear();
    return QSqlDatabase::database();
}

bool DataBase::getDataBaseStatus()
{
    return getConnection().isOpen();
}

QDateTime DataBase::getRecordTimeStamp(const int &id)
{
    QSqlDatabase db = getConnection();
    if (db.isOpen()) {
        QSqlQuery query(db);
        query.prepare("SELECT datetime FROM current where id=:id");
        query.bindValue(":id", id);
        if (query.exec()) {
            if (query.next()) {
                return query.value(0).toDateTime();
            }
        }
    }
    return QDateTime();
}

QSqlError DataBase::getLastDbError()
{
    return QSqlDatabase::database(selectedConnection, false).lastError();
}

QString DataBase::toString()
{
    QSqlDatabase db = QSqlDatabase::database(selectedConnection, false);
    return QString("%1: %2@%3").arg(selectedConnection, db.databaseName(), db.hostName());
}

void DataBase::writeInfo(const quint64 &serial, bool isGold, QString imgPath)
{
    QVariant fw_rev = QVariant::Int, fw_maj = QVariant::Int, fw_min = QVariant::Int;
    QVariant git_hash = QVariant::String;
    bool svnRevOk = false, gitRevOk = false;

    QRegularExpression rx_git("^.*(\\d+)\\.(\\d+)-(\\d+)-g([\\da-fA-F]+)\\.bit$");
    QRegularExpressionMatch match_git = rx_git.match(imgPath);
    if (match_git.hasMatch()) {
        auto groups = match_git.capturedTexts();
        Q_ASSERT(groups.size() == 5);

        bool ok1, ok2, ok3;
        int maj = static_cast<quint8>(groups[1].toUShort(&ok1));
        int min = static_cast<quint8>(groups[2].toUShort(&ok2));
        int rev = groups[3].toUShort(&ok3);

        gitRevOk = ok1 & ok2 & ok3;
        if (gitRevOk) {
            fw_rev = rev;
            fw_maj = maj;
            fw_min = min;
            git_hash = groups[4];
        }
    }

    QRegularExpression rx_svn("^.*-r(\\d+).bit$");
    QRegularExpressionMatch match_svn = rx_svn.match(imgPath);
    if (!gitRevOk && match_svn.hasMatch()) {
        uint rev = match_svn.capturedTexts()[1].toUInt(&svnRevOk);
        if (svnRevOk)
            fw_rev = rev;
    }

    QSqlDatabase db = QSqlDatabase::database(selectedConnection);
    if (db.isOpen()) {
        QSqlQuery query(db);
        query.prepare(QString("INSERT INTO hw_firmware ("
                              " fw_maj, fw_min, git_hash, revision,"
                              " base_name, path,"
                              " fromHost, goldImg, serial, serialHex)"
                              " VALUES ("
                              ":fw_maj, :fw_min, :git_hash, :rev,"
                              " :base_name, :path,"
                              " :host, :isGold, :serial, :serialHex)"));

        query.bindValue(":fw_maj", fw_maj);
        query.bindValue(":fw_min", fw_min);
        query.bindValue(":rev", fw_rev);
        query.bindValue(":git_hash", git_hash);

        query.bindValue(":base_name", QFileInfo(imgPath).fileName());
        query.bindValue(":path", imgPath);

        query.bindValue(":host", QHostInfo::localHostName());
        query.bindValue(":isGold", isGold);
        query.bindValue(":serial", serial);
        query.bindValue(":serialHex", getSerialStr(serial));

        if (query.exec()) {
            qInfo() << "Record parameters were saved to DataBase";
            emit setStatus("Record parameters were saved to DataBase", DB_STATUS_OK);
        } else {
            qWarning() << "DB error:" << db.lastError().text();
            emit setStatus(QString("Failed to save parameters into DataBase<br/>")
                               .append(query.lastError().text()),
                           DB_STATUS_ERROR);
        }
        //        qInfo() << "insert query: " << query.lastQuery();
        //        qInfo() << query.boundValues();
    } else {
        qWarning() << db.lastError().text();
        emit setStatus(QString("Can't open DataBase to save parameter<br/>").append(db.lastError().text()),
                       DB_STATUS_ERROR);
    }
}

void DataBase::readInfo(quint64 devSerial)
{
    QSqlDatabase db = getConnection();
    if (db.isOpen()) {
        QSqlQuery query(db);
        const QString sqlBase = "SELECT id, serialHex, path, datetime, goldImg, fw_maj, fw_min, "
                                "revision, git_hash FROM hw_firmware %1 ORDER BY id";
        if (devSerial) {
            query.prepare(sqlBase.arg(" WHERE serial=:serial"));
            query.bindValue(":serial", devSerial);
        } else
            query.prepare(sqlBase.arg(""));

        if (query.exec()) {
            QList<Record> recordList;
            uint lastGoldId = -1, lastNormId = -1;
            while (query.next()) {
                QFileInfo fi(query.value(2).toString());

                Record r;
                r.id = query.value(0).toUInt();
                r.serialHexStr = query.value(1).toString();
                r.dateStr = query.value(3).toString();
                r.isGold = query.value(4).toBool();
                r.fw_maj = query.value(5).toInt();
                r.fw_min = query.value(6).toInt();
                r.revision = query.value(7).toUInt();
                r.isGit = !query.value(8).isNull();
                r.git_hash = query.value(8).toString();
                r.fileName = fi.fileName();
                r.filePath = fi.filePath();

                if (r.isGold)
                    lastGoldId = r.id;
                else
                    lastNormId = r.id;

                recordList << r;
            }
            for (const Record &r : recordList) {
                QDebug deb = qInfo().noquote();
                deb << r.serialHexStr << r.dateStr.toStdString().c_str();

                if (r.id == lastGoldId)
                    deb << "\033[31m";
                else if (r.id == lastNormId)
                    deb << "\033[32m";
                else
                    deb << "";

                deb << (r.isGold ? " Gold " : "Second")
                    << QString("%1 %2 %3 %4")
                           .arg(r.fw_maj)
                           .arg(r.fw_min)
                           .arg(r.revision, 5)
                           .arg(r.git_hash.isNull() ? "---" : r.git_hash, 7)
                    << (r.id == lastGoldId || r.id == lastNormId ? "\033[0m" : "") << r.fileName
                    << r.filePath;
            }
        } else {
            qDebug() << db.lastError().text();
            emit setStatus(QString("Failed to read freeze parameters from DataBase<br/>")
                               .append(query.lastError().text()),
                           DB_STATUS_ERROR);
        }
    }
}
