/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "FlashDevTask.h"
#include "FpgaBitFile.h"

#include <netinet/in.h>

#include <QDebug>

namespace {
// See UG470 "7 Series FPGAs Configuration"
// Type-1 packet headers
constexpr const uint32_t BITSTREAM_WRITE_TIMER   = 0x30022001;
constexpr const uint32_t BITSTREAM_WRITE_WBSTAR  = 0x30020001;
constexpr const uint32_t BITSTREAM_WRITE_CMD     = 0x30008001;
// Type-1 packet data
constexpr const uint32_t BITSTREAM_TIMER_CFG_MON = 0x40000000;
constexpr const uint32_t BITSTREAM_CMD_IPROG     = 0x0000000F;
// TIMER clock
constexpr const double timer_freq_min = 125e3;
constexpr const double timer_freq_max = 380e3;
} // namespace

bool FlashDevTask::patchGoldenImage(QByteArray *bitstream, uint32_t bootAddr, bool useAddr4bMode)
{
    auto syncPos = FpgaBitFile::findBitstreamSync(*bitstream);
    if (syncPos < 0 || (syncPos + 4*12) > bitstream->size())
        return false;

    bitstream->remove(0, syncPos);
    syncPos = 0;

    uint32_t *data = reinterpret_cast<uint32_t *>(bitstream->data() + syncPos);

    auto p_timer = data + 8;
    auto p_wbstar = data + 10;
    auto p_cmd = data + 12;

    if (ntohl(*(p_timer-1)) != BITSTREAM_WRITE_TIMER) {
        qWarning() << "bitstream patch failed: write TIMER not found";
        goto err;
    }
    if (ntohl(*(p_wbstar-1)) != BITSTREAM_WRITE_WBSTAR) {
        qWarning() << "bitstream patch: write WBSTAR not found";
        goto err;
    }
    if (ntohl(*(p_cmd-1)) != BITSTREAM_WRITE_CMD) {
        qWarning() << "bitstream patch: write CMD not found";
        goto err;
    }

    {
        const uint32_t timer_ticks = 0x10000;
        const uint32_t data_timer = BITSTREAM_TIMER_CFG_MON | timer_ticks; // Enable watchdog during configuration
        const uint32_t timer_value = data_timer & 0x3FFFFFFF;
        const double timer_ms_min = timer_value / timer_freq_max * 1e3;
        const double timer_ms_max = timer_value / timer_freq_min * 1e3;

        qDebug().noquote() << QString("Change TIMER from %1 to %2 (%3~%4 ms)")
                              .arg(ntohl(*p_timer), 8, 16, QChar('0'))
                              .arg(data_timer, 8, 16, QChar('0'))
                              .arg(timer_ms_min).arg(timer_ms_max);
        *p_timer = htonl(data_timer);
    }

    {
        const uint32_t data_wbstar = bootAddr >> (useAddr4bMode ? 8 : 0);

        qDebug().noquote() << QString("Change WBSTAR from %1 to %2")
                              .arg(ntohl(*p_wbstar), 8, 16, QChar('0'))
                              .arg(data_wbstar, 8, 16, QChar('0'));
        *p_wbstar = htonl(data_wbstar);
    }

    {
        const uint32_t data_cmd = BITSTREAM_CMD_IPROG;

        qDebug().noquote() << QString("Change CMD from %1 to %2")
                              .arg(ntohl(*p_cmd), 8, 16, QChar('0'))
                              .arg(data_cmd, 8, 16, QChar('0'));
        *p_cmd = htonl(data_cmd);
    }
    return true;
err:
    FpgaBitFile::dumpBitstream(*bitstream);
    return false;
}

std::vector<uint16_t> FlashDevTask::getBlockData() const
{
    return getImageData(flashIdData.getBlockSize());
}

std::vector<uint16_t> FlashDevTask::getSectorData() const
{
    return getImageData(flashIdData.getSectorSize());
}

std::vector<uint16_t> FlashDevTask::getImageData(int chunkSize) const
{
    if (chunkSize <= 0)
        return {};
    const auto &count = chunkSize / sizeof(uint16_t); // in 16-bit words
    std::vector<uint16_t> blockData(count, 0);
    auto begin = pos.offset();
    auto end = std::min(begin + chunkSize, int(flashImageData.size()));
    int dataSize = std::min(chunkSize, end - begin); // bytes
    memcpy(blockData.data(), flashImageData.data() + begin, dataSize);

    return blockData;
}

uint32_t FlashDevTask::startAddr() const
{
    switch (image.type) {
    case FlashImageType::Golden: return 0;
    case FlashImageType::Regular: return flashIdData.getFlashSizeBytes() / 2; // duplicate of FlashDev::getImageOffset
    }
    return 0;
}

QByteArray FlashDevTask::makeMultibootImage() const
{
    switch (image.type) {
    case FlashImageType::Golden: {
        QByteArray buf(image.data);
        uint32_t bootAddr = flashIdData.getFlashSizeBytes() / 2;
        patchGoldenImage(&buf, bootAddr, useAddr4bMode);
        return buf;
    }
    case FlashImageType::Regular: {
        return image.data;
    }
    }
    return {};
}

QString FlashDevTask::getProgressStr() const
{
    if (pos.isComplete())
        return QString("%1 bytes")
            .arg(pos.size());
    return QString("%1 of %2 bytes [flash:%3]")
        .arg(pos.offset())
        .arg(pos.size())
        .arg(startAddr() + pos.offset(), 8, 16, QChar('0'));
}

QString FlashDevTask::taskStatusStr() const
{
    QString s;
    s += QString("%1 %2").arg(flashOpStr(op)).arg(stateStr());
    if (isAborted()) {
        s += " at " + getProgressStr();
    }
    if (isComplete()) {
        s += " " + getProgressStr();
    }
    return s;
}
