#ifndef HISTWORKER_H
#define HISTWORKER_H

#include <QObject>
#include <optional>

#include "base/DeviceIndex.h"

#include "Histogramm.h"
#include "HistMngrCfg.h"
#include "DeviceCalibration.h"
#include "DeviceData.h"
#include "WorkerStat.h"

class QTime;
class QTimer;
class QSemaphore;
class QwtPlotCurve;

// [DevIndex] = <chOffset, maxDetectedChNum>
struct DevChMap : public QHash<DeviceIndex, QPair<int, int>>
{
    int getMaxChNum() const
    {
        int maxChNum = 0;
        auto it = cbegin();
        while(it != cend()) {
            maxChNum = std::max(maxChNum, it.value().first + it.value().second);
            ++it;
        }
        return maxChNum;
    }
};
Q_DECLARE_METATYPE(DevChMap)

struct TestDp {
    TestDp() { clear(); }
    void clear() {
        dt=0;
        cnt=0;
        temp.clear();
        startTai = {};
        started=false;
    }
    bool expired(TaiTime tai) {
        double dt = (tai.sec-startTai.sec)*1e9;
        dt += tai.ns-startTai.ns;
        return dt > 1e9;
    }

    double dt;
    int cnt;
    QMap<DeviceIndex, QMap<int, QPair<double, int> >> temp; // per [devIndex][tempId]={sum_temp, sum_cnt}
    TaiTime startTai;
    bool started;
};

class HistWorker : public QObject
{
    Q_OBJECT
public:
    explicit HistWorker(QSemaphore *_histSem);
    ~HistWorker() override;

signals:
    void devListUpdated(const DevChMap &tdcChMap,
                        const QHash<HistKey, Histogramm> &histogramms) const;
    void inlResultsUpdated(const InlResults &, bool allRes=false) const;
    void hist_updated(QHash<HistKey, Histogramm> histogramms, bool force) const;
    void histConfigChanged(HistMngrCfg config_) const;
    void workerStatUpdated(const WorkerStat &stats);
    void taiReperUpdated(HistKey hKey, const TaiTime &taiReperSec) const;

private slots :
    void handle_data();

public slots:
    void clearAll();
    void setHistConfig(HistMngrCfg config_);
    void clearHist(const HistKey &hist_key);
    void dataUpdated(const DecodedData &data);
    void inlConfigUpdated(InlResults inl);
    void tdcTtvxMapUpdated(TdcTtvxsMapType ttMap);
    void requestInlResults(bool allRes=false) const;

private:
    void corrFillCurDevMap();
    void refreshCurrentPlot(bool force=false);
    void recreateTaiHist(HistKey histKey, std::optional<TaiTime> reper);
    void clearTaiHist(HistKey histKey);
    void shiftTaiHist(HistKey histKey, uint shift=1);
    int updateDevMaxCh(const HistId &hId, const DeviceIndex &index, int ch);
    int getDevChOffset(const HistId &hId, const DeviceIndex &index, int curCh);
    void addBaseHist(HistId id);
    void addDeviceChHist(HistKey key);
    void addDeviceChHist(HistId id, const DeviceIndex &index, int ch);
    void dataProcessingSingleEv(CompleteEvent &ev,
                                int &tdcHitsPerEv,
                                int &adcHitsPerEv);
    void dataProcessingTqdc(TqdcEvent &ev,
                           int &tdcHitsPerEv,
                           int &adcHitsPerEv,
                           TqdcEvent &tqdcStopEvent,
                           QVector<TdcChHit> &stopHits);
    void dataProcessingTdc(TqdcEvent &ev,
                           const int &ch,
                           QVector<TdcChHit> &tdcHits,
                           int &tdcHitsPerEv);
    void dataProcessingAdc(TqdcEvent &ev,
                           const int &ch,
                           QVector<AdcChHit> &adcHits,
                           int &adcHitsPerEv);
    void dataProcessingTai(const TqdcEvent &ev);
    void dataProcessingMsc(const DeviceIndex &index, const MscData &ev);
    void grabStatInfo(TqdcEvent &ev);
    void errorsProcessing(TqdcEvent &ev);
    void fillDoublePulse(HistKey hist_key,
                         TqdcEvent &tdcStopEvent,
                         QVector<TdcChHit> &stopHits,
                         TqdcEvent &ev,
                         QVector<TdcChHit> &tdcEvents);
    QVector<double> &getInl(const TqdcEvent &ev, const TdcChHit &tdcEv);
    QVector<double> &getInl(const DeviceIndex &index, uint ch);

    QMap<DeviceIndex, QMap<uint, uint> > getDevChCntMap(const HistId &hId);
    void fillDevChHist(const HistId &hId, QMap<DeviceIndex, QMap<uint, uint> > map);
    void trashAllData();
    bool isHistGroupEn(HistId groupId) const;

    QSemaphore * histSem;
    QHash<HistKey, std::optional<TaiTime>> taiReperNew;
    InlResults curDevCorrMap; // INL from config (formed from fullDevCorrMap for knownDevices)
    InlResults fullDevCorrMap; // INL from config
    InlResults inlResults; // collecting data for new INL
    QMap<DeviceIndex, QMap<uint, quint16> > devErrors;
    bool data_processing = false;
    bool destroy_flg = false;
    QList<DecodedData> data_list;
    TdcTtvxsMapType tdcTtvxsMap;
    const CompleteEvent *curEv = nullptr;
    TestDp testDp;
    HistMngrCfg config;
    DevChMap tdcChMap; // use for corrFillCurDevMap() and hist HistId::tdcCh
    DevChMap adcChMap;
    QHash<HistKey, Histogramm> histogramms;
    WorkerStat stat;
    QTimer * const refreshHistTimer;
};

#endif // HISTWORKER_H

