//
//    Copyright 2016 Ivan Filippov
//    Copyright 2016 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "TcpMetricSender.h"

#include <QtNetwork>
#include <QDebug>
#include <QSemaphore>
#include <QTimer>

#include "MetricSenderHelper.h"

namespace {
const int CONNECT_TIMEOUT_MS = 1000;
const int RECONNECT_TIMEOUT_MS = 1000;
const int RECOVER_TIMEOUT_MS = 3 * RECONNECT_TIMEOUT_MS;
// queue max 1000 metrics per second
const int QUEUE_SIZE = 100;
const int RUN_QUEUE_TIMER_MS = 10;
}

#if QT_VERSION < 0x050000
Q_DECLARE_METATYPE(QAbstractSocket::SocketError)
#endif

TcpMetricSender::TcpMetricSender(QObject *parent) :
    QObject(parent),
    helper(new MetricSenderHelper(this)),
    tcpSocket(new QTcpSocket(this)),
    reconnectTimer(new QTimer(this)),
    runQueueTimer(new QTimer(this)),
    connectionAttempts(0)
{
    connect(tcpSocket, &QTcpSocket::connected,
            this, &TcpMetricSender::connected);
    connect(tcpSocket, &QTcpSocket::disconnected,
            this, &TcpMetricSender::disconnected);
    qRegisterMetaType<QAbstractSocket::SocketError>();

    connect(tcpSocket, QOverload<QAbstractSocket::SocketError>::of(&QTcpSocket::error),
        this, &TcpMetricSender::displayError);

    connect(reconnectTimer, &QTimer::timeout, this, &TcpMetricSender::reconnectToServer);
    reconnectTimer->setSingleShot(true);
    reconnectTimer->start(0);

    connect(runQueueTimer, &QTimer::timeout, this, &TcpMetricSender::runQueue);
    runQueueTimer->setSingleShot(true);
    runQueueTimer->start(RUN_QUEUE_TIMER_MS);
}

TcpMetricSender::~TcpMetricSender()
{
    reconnectTimer->stop();
    tcpSocket->abort();
    delete helper;
}

void TcpMetricSender::setBaseTags(const MetricTags &s)
{
    helper->setBaseTags(s);
}

void TcpMetricSender::sendMetric(const MetricEntity &ent)
{
    queue.push_back(helper->formatMetricLine(ent));
//    sendQueue(s);
}

void TcpMetricSender::reconnectToServer()
{
    if (tcpSocket->isValid() &&
            (tcpSocket->state() == QTcpSocket::ConnectedState ||
             tcpSocket->state() == QTcpSocket::ConnectingState) ) {
        qInfo() << "reconnectToServer ignored";
        return;
    }

    const QHostAddress &hostAddr = helper->hostAddr();
    quint16 hostPort = helper->hostPort();
    if (hostAddr.isNull()) {
        // may be a network error
        qWarning() << QString("TcpMetricSender: no address for %1").arg(helper->getUrl().toDisplayString());
        reconnectTimer->start(RECOVER_TIMEOUT_MS);
        return;
    }
// TODO
//    if (connectionAttempts == 0) {
//        qInfo() << QString("TcpMetricSender: connecting to %1 (%2:%3)")
//                   .arg(host)
//                   .arg(hostAddr.toString())
//                   .arg(hostPort);
//    }
    try{
        tcpSocket->abort();
        tcpSocket->connectToHost(hostAddr, hostPort);
        ++connectionAttempts;
        if (!tcpSocket->waitForConnected(CONNECT_TIMEOUT_MS)) {
            if (connectionAttempts<=1) {
                QString mess = QString("TcpMetricSender can't connect to %1 (%2:%3): %4")
                        .arg(helper->getUrl().toDisplayString())
                        .arg(hostAddr.toString())
                        .arg(hostPort)
                        .arg(tcpSocket->errorString());
                qWarning()<<mess;
                reconnectTimer->start(RECONNECT_TIMEOUT_MS);
            } else
                reconnectTimer->start(RECOVER_TIMEOUT_MS);
        }
    } catch(...) {
        qWarning()<<"TcpMetricSender::reconnectToServer() exception caught";
        reconnectTimer->start(RECOVER_TIMEOUT_MS);
    }
}

void TcpMetricSender::displayError(QAbstractSocket::SocketError socketError)
{
    switch (socketError) {
    case QAbstractSocket::RemoteHostClosedError:
//        qInfo() << "Remote hoste closed connection.";
        break;
    case QAbstractSocket::HostNotFoundError:
//        qInfo() << "The host was not found. Please check the "
//                    "host name and port settings.";
        break;
    case QAbstractSocket::ConnectionRefusedError:
        if(connectionAttempts<=1){
//            qInfo() << tr("The connection was refused by the peer. "
//                           "Make sure the fortune server is running, "
//                           "and check that the host name and port "
//                           "settings are correct.");
        }
        break;
    default:;
//        qInfo() << tr("The following error occurred: %1.")
//                    .arg(tcpSocket->errorString());
    }
}

void TcpMetricSender::deleteClient()
{
    deleteLater();
}


void TcpMetricSender::connected()
{
    qInfo().noquote() << QString("[TcpMetricSender]: connected to %1")
               .arg(helper->getUrl().toDisplayString());
    connectionAttempts=0;
    sem.release(QUEUE_SIZE);
}

void TcpMetricSender::disconnected()
{
    qInfo().noquote() << "[TcpMetricSender]: disconnected, trying to reconnect";
    connectionAttempts=0;
    reconnectTimer->start(RECONNECT_TIMEOUT_MS);
    while (sem.tryAcquire()) {}
}

void TcpMetricSender::runQueue()
{
    //    qInfo() << "TcpMetricSender::runQueue()" << sem.available();
    if (sendQueue())
        runQueueTimer->start(1);
    else
        runQueueTimer->start(RUN_QUEUE_TIMER_MS);
}

bool TcpMetricSender::sendQueue()
{
    if (queue.isEmpty())
        return false;
    if (!helper->isEnabled())
        return false;
    try {
    if (!tcpSocket->isValid() || tcpSocket->state() != QTcpSocket::ConnectedState) {
        return false;
    }
    if (!tcpSocket->isWritable()) {
        return false;
    }
    Q_FOREACH(QString mess, queue) {
        QByteArray messData = mess.toUtf8();
        int written = tcpSocket->write(messData);
        if (written != messData.size()) {
            qWarning() << QString("Failed to send '%1': %2").arg(mess).arg(tcpSocket->errorString());
            return false;
        }
    }
    int n = queue.size();
    queue.clear();
    sem.release(n);
    }
    catch (...) {
        return false;
    }
    return true;
}

bool TcpMetricSender::isConnectionAlive() const
{
    if(tcpSocket->isValid() &&
            (tcpSocket->state() == QTcpSocket::ConnectedState ||
             tcpSocket->state() == QTcpSocket::ConnectingState) )
        return true;
    return false;
}
