#include "PostgresDB.h"

#include "base/ConfigDocument.h"
#include "base/ConfigSelector.h"
#include "base/ConfigSelectorConverter.h"
#include "base/Globals.h"
#include "base/RcRunId.h"
#include "base/ServiceConfigUrl.h"

#include <QCoreApplication>
#include <QDateTime>
#include <QElapsedTimer>
#include <QJsonObject>
#include <QJsonDocument>
#include <QJsonParseError>
#include <QSettings>
#include <QSqlError>
#include <QSqlQuery>
#include <QUrl>
#include <optional>

namespace {
constexpr const char *tableCur = "current_config";
constexpr const char *tableRun = "run_history";

constexpr const char *serviceName = "postgres";
constexpr const char *driverName = "QPSQL";
constexpr const char *connectionName = "PostgresDB";

constexpr const char *k_id = "id";
constexpr const char *k_cs = "config_selector";
constexpr const char *k_host = "current_host";
constexpr const char *k_phantom = "phantom";
constexpr const char *k_dom = "date_of_modification";
constexpr const char *k_cfg = "config";

constexpr const char *k_runId = "runId";

template<typename T>
std::optional<T> jsonStrToObject(const QString &jsonStr) {
    QJsonParseError jsonErr;
    auto jsonDoc = QJsonDocument::fromJson(jsonStr.toUtf8(), &jsonErr);
    if (QJsonParseError::NoError != jsonErr.error) {
        qWarning() << QString("ERROR: Parsing config json: %1")
                          .arg(jsonErr.errorString());
        return {};
    }
    auto json = jsonDoc.object();
    return QVariant::fromValue(json).value<T>();
}

/*
    query.prepare(
        "INSERT INTO current_config ("
                                    "phantom"
                                    ", config_selector"
                                    ", current_host"
                                    ", date_of_modification"
                                    ", config"
                                    ")"
                                    " VALUES ("
                                    ":phantom"
                                    ", :config_selector"
                                    ", :current_host"
                                    ", :date_of_modification"
                                    ", :config"
                                    ") "
                                    "ON CONFLICT (config_selector) DO UPDATE SET "
                                    "phantom = excluded.phantom"
                                    ", current_host = excluded.current_host"
                                    ", date_of_modification = excluded.date_of_modification"
                                    ", config = excluded.config"
                                    ";");
*/
}

void PostgresDB::connectToDb()
{
    if (QSqlDatabase::isDriverAvailable(driverName)) {
        auto db = getDb();
        alive = db.isValid() && db.isOpen();
    } else {
        alive = false;
        qCritical() << __func__ << "QPSQL not loaded";
        qCritical() << __func__ << "Database driver QPSQL not loaded (qt5-qtbase-postgresql).";
    }
}

ConfigDocument PostgresDB::readConfigDocument(const ConfigSelector &cs)
{
    if(cs.getIndex().isDefault())
        return {};

    auto db = getDb();
    if(!db.isValid() || !db.isOpen())
        return {};

    QSqlQuery query(db);
    query.prepare(
        QString("SELECT %1, %2, %3, %4, %5 FROM %10 WHERE %11=:%11;")
            .arg(k_id, k_host, k_dom, k_phantom, k_cfg,
                 tableCur, k_cs)
        );

    auto csJson = QVariant::fromValue(cs).value<QJsonObject>();
    auto csData = QJsonDocument(csJson).toJson(QJsonDocument::Compact);
    auto csStr = QString::fromUtf8(csData);

    query.bindValue(QString(":")+k_cs, csStr);

    if (query.exec()) {
        while (query.next()) {
            // Doesn't work QVar isn't JSON type (it is QString type)
            // auto jsonDoc = query.value(3).toJsonDocument();
            // ret = jsonDoc.object();

            ConfigOid oid(QString::number(query.value(k_id).toUInt()));
            auto hostStr = query.value(k_host).toString();
            auto dateOfMod = query.value(k_dom).toDateTime();
            auto isPhantom = query.value(k_phantom).toBool();
            auto jsonStr = query.value(k_cfg).toString();
            auto json = jsonStrToObject<QJsonObject>(jsonStr);
            if(!json) {
                qWarning() << QString("[%1] ERROR: Parsing config json for cs='%3' id=%4")
                                  .arg(__func__)
                                  .arg(cs.toString(),
                                       query.value(k_id).toString());
                continue;
            }

            ConfigMeta cm(QHostInfo::fromName(hostStr), dateOfMod, isPhantom);
            ConfigDocument cd(oid, cs, cm, *json);
            return cd;
        }
    } else {
        logError(__func__, query);
    }

    return {};
}

bool PostgresDB::writeConfigDocument(const ConfigDocument &cd)
{
    auto db = getDb();
    if(!db.isValid() || !db.isOpen())
        return {};

    QSqlQuery query(db);
    query.prepare(
        QString("INSERT INTO %1 (%2, %3, %4, %5, %6) VALUES (:%2,:%3,:%4,:%5,:%6) "
                "ON CONFLICT (%2) DO UPDATE SET "
                "%3 = excluded.%3,"
                "%4 = excluded.%4,"
                "%5 = excluded.%5,"
                "%6 = excluded.%6;")
            .arg(tableCur, k_cs, k_host, k_dom, k_phantom, k_cfg)
        );

    auto csJson = QVariant::fromValue(cd.getConfigSelector()).value<QJsonObject>();
    auto csData = QJsonDocument(csJson).toJson(QJsonDocument::Compact);
    auto csStr = QString::fromUtf8(csData);

    auto cfgData = QJsonDocument(cd.getConfig()).toJson(QJsonDocument::Compact);
    auto cfgStr = QString::fromUtf8(cfgData);

    query.bindValue(QString(":")+k_cs, csStr);
    query.bindValue(QString(":")+k_host, cd.getConfigMeta().getCurrentHost().hostName());
    query.bindValue(QString(":")+k_dom, cd.getConfigMeta().getModificationTimestamp());
    query.bindValue(QString(":")+k_phantom, cd.getConfigMeta().isPhantom());
    query.bindValue(QString(":")+k_cfg, cfgStr);

    bool ok = query.exec();
    if (ok) {
        qDebug() << "Program config was saved to DataBase";
    } else {
        logError(__func__, query);
    }

    return ok;
}

bool PostgresDB::writeConfigRun(const RcRunId &runId, const QJsonObject &jsonCfg)
{
    auto db = getDb();
    if(!db.isValid() || !db.isOpen())
        return {};

    QSqlQuery query(db);
    query.prepare(
        QString("INSERT INTO %1 (%2, %3, %4, %5) VALUES (:%2,:%3,:%4, now());")
            .arg(tableRun, k_runId, k_host, k_cfg, k_dom)
        );

    auto runIdJson = QVariant::fromValue(runId).value<QJsonObject>();
    auto runIdData = QJsonDocument(runIdJson).toJson(QJsonDocument::Compact);
    auto runIdStr = QString::fromUtf8(runIdData);

    auto cfgData = QJsonDocument(jsonCfg).toJson(QJsonDocument::Compact);
    auto cfgStr = QString::fromUtf8(cfgData);

    query.bindValue(QString(":")+k_cs, runIdStr);
    query.bindValue(QString(":")+k_host, QHostInfo::localHostName());
    query.bindValue(QString(":")+k_dom, QDateTime::currentDateTime());
    query.bindValue(QString(":")+k_cfg, cfgStr);

    bool ok = query.exec();
    if (ok) {
        qInfo() << "Run config was saved to PostgresDB";
    } else {
        logError(__func__, query);
    }

    return ok;
}

int PostgresDB::delete_config(const ConfigSelector &cs)
{
    auto db = getDb();
    if(!db.isValid() || !db.isOpen())
        return {};

    QSqlQuery query(db);
    query.prepare(
        QString("UPDTAE %1 SET %2 = :%2, %3 = now() WHERE %4 = :%4;")
            .arg(tableCur, k_phantom, k_dom, k_cs)
        );

    auto csJson = QVariant::fromValue(cs).value<QJsonObject>();
    auto csData = QJsonDocument(csJson).toJson(QJsonDocument::Compact);
    auto csStr = QString::fromUtf8(csData);

    query.bindValue(QString(":")+k_cs, csStr);
    query.bindValue(QString(":")+k_phantom, true);

    bool ok = query.exec();
    if (ok) {
        qInfo() << "Config was marked as phantom in PostgresDB";
    } else {
        logError(__func__, query);
    }

    return ok;
}

QMap<ProgramIndex, QMap<ConfigurationName, QDateTime> > PostgresDB::get_config_map(const DaqModule &daq_module)
{
    auto db = getDb();
    if(!db.isValid() || !db.isOpen())
        return {};

    QMap<ProgramIndex, QMap<ConfigurationName, QDateTime> > ret;
    QSqlQuery query(db);
    query.prepare(
        QString("SELECT %1, %2, %3, %4, %5, %6 FROM %10 WHERE %11->>'%12'=:%12;")
            .arg(k_id, k_cs, k_host, k_dom, k_phantom, k_cfg,
                 tableCur, k_cs, k_cs_program_type)
        );

    query.bindValue(QString(":")+k_cs_program_type, daq_module.toString());
    if (query.exec()) {
        while (query.next()) {
            auto dateOfMod = query.value(k_dom).toDateTime();
            auto isPhantom = query.value(k_phantom).toBool();
            auto jsonStr = query.value(k_cfg).toString();

            auto json = jsonStrToObject<QJsonObject>(jsonStr);
            if(!json) {
                qWarning() << QString("[%1] ERROR: Parsing config json for daq_module='%3' id=%4")
                                  .arg(__func__)
                                  .arg(daq_module.toString(),
                                       query.value(k_id).toString());
                continue;
            }

            auto cs = jsonStrToObject<ConfigSelector>(query.value(k_cs).toString());
            if(!cs) {
                qWarning() << QString("[%1] ERROR: Parsing config json for daq_module='%3' id=%4")
                                  .arg(__func__)
                                  .arg(daq_module.toString(),
                                       query.value(k_id).toString());
                continue;
            }
            if(isPhantom)
                continue;

            ret[cs->getIndex()][cs->getConfigurationName()] = dateOfMod;
        }
    } else {
        logError(__func__, query);
    }

    return ret;
}

void PostgresDB::initDb(QSqlDatabase *db)
{
    ServiceConfigUrl servCfgUrl(serviceName, {}, false);
    enabled = servCfgUrl.isEnabled();
    if(!servCfgUrl.isEnabled())
        return;

    const auto &settings = Globals::getSettings();
    auto env = Globals::getEnvironment();
    const QString base_path = env + "/" + serviceName;
    QString databaseName = settings->value(base_path + "_database").toString();

    // db->setConnectOptions(";port=5432;application_name=aaa;connect_timeout=10");
    const auto &k_connect_options = base_path + "_connect_options";
    QString opStr;
    if(settings->contains(k_connect_options))
        opStr = settings->value(k_connect_options).toString();
    else
        opStr = QString("application_name='%1 %2'")
                    .arg(QCoreApplication::applicationName())
                    .arg(QCoreApplication::arguments().mid(1).join(' '));
    db->setConnectOptions(opStr);

    auto url = servCfgUrl.getUrl();
    db->setHostName(url.host());
    db->setDatabaseName(databaseName);
    db->setUserName(url.userName());
    db->setPassword(url.password());
}

void PostgresDB::logError(const char *funcName, const QSqlQuery &query)
{
    static QMap<QString, QElapsedTimer> suppresTimers;
    auto timer = suppresTimers[funcName];
    // Suppres warnings for 1 minute
    if(timer.isValid() && timer.elapsed()<60e3)
        return;

    qWarning().noquote().nospace() << "Query: " << query.lastQuery();
    qWarning().noquote().nospace() << "BoundValues: " << query.boundValues();
    qWarning() << "BoundValues: " << query.boundValues();

    auto queryErr = query.lastError();
    if(queryErr.isValid())
        qWarning() << queryErr.text();
    auto dbErr = getDb().lastError();
    if(dbErr.isValid())
        qWarning() << dbErr.text();
}

QSqlDatabase PostgresDB::getDb()
{
    if(!QSqlDatabase::isDriverAvailable(driverName))
        return {};

    QSqlDatabase db;
    if(QSqlDatabase::contains(connectionName)) {
        db = QSqlDatabase::database(connectionName);
    } else {
        db = QSqlDatabase::addDatabase(driverName, connectionName);
        initDb(&db);
    }

    if(!db.isOpen()) {
        if(db.open()) {
            qInfo() << serviceName << "connected";
        } else {
            qWarning("Can't connect to %s DB (%s). Err:\n%s",
                     serviceName, db.hostName().toStdString().c_str(),
                     db.lastError().text().toStdString().c_str());
        }
    }

    return db;
}
