#include "PnpDiscovery_p.h"

#include <QTimer>

#include "base/ThreadController.h"
#include "PNPServer.h"

namespace {
const qint64 throttleTimeMs = 1000;
}

PnpDiscoveryPrivate::PnpDiscoveryPrivate(QObject *parent)
    : QObject(parent)
    , throttleTimer(new QTimer(this))
{
    qRegisterMetaType<PnpInfoCache>();

    auto pnp_server = new PNPServer();
    connect(pnp_server, &PNPServer::programDiscovered,
            this, &PnpDiscoveryPrivate::receivedProgramDescription);

    connect(pnp_server, &PNPServer::programClosed,
            this, &PnpDiscoveryPrivate::receivedProgramClosed);

    connect(pnp_server, &PNPServer::programClosed,
            this, &PnpDiscoveryPrivate::gotProgramClosed);

    connect(this, &PnpDiscoveryPrivate::updateProgramDescription,
            pnp_server, &PNPServer::updateProgramDescription);

    connect(this, &PnpDiscoveryPrivate::closeMyProgram,
            pnp_server, &PNPServer::closeProgram);

    connect(this, &PnpDiscoveryPrivate::startPnpSearch,
            pnp_server, &PNPServer::enableSearch);

    connect(throttleTimer, &QTimer::timeout,
            this, &PnpDiscoveryPrivate::throttleTimeout);

    new ThreadController(pnp_server, this);
    QTimer::singleShot(0, this, &PnpDiscoveryPrivate::startPnpSearch);

    throttleTimer->setSingleShot(true);

    sendUpdates(); // send initial empty
}

void PnpDiscoveryPrivate::receivedProgramDescription(const ProgramDescription &prog_decsr)
{
    if(prog_decsr.isEmpty())
        return;
    const QUuid &uuid = prog_decsr.uuid;
    pnp_cache.insert(uuid, prog_decsr);
    requestUpdate();
}

void PnpDiscoveryPrivate::receivedProgramClosed(const ProgramDescription &pd)
{
    if(pnp_cache.remove(pd.uuid))
        requestUpdate();
}

void PnpDiscoveryPrivate::sendUpdates()
{
    removeOldDuplicates();
    emit cacheUpdated(pnp_cache);
    updateTime.restart();
}

void PnpDiscoveryPrivate::throttleTimeout()
{
    sendUpdates(); // emit per throttleTimeMs
}

void PnpDiscoveryPrivate::requestUpdate()
{
    qint64 t = updateTime.elapsed() - throttleTimeMs;
    if (t < 0) {
        throttleTimer->start(-t);
    } else {
        throttleTimeout();
    }
}

void PnpDiscoveryPrivate::removeOldDuplicates()
{
    // Normaly we should fill just allClIndexes set and exit on 1st 'return'
    ClientIndexSet allClIndexes;
    ClientIndexSet dupClIndexes;
    for(auto i=pnp_cache.begin(); i!=pnp_cache.end(); ++i) {
        const auto &ci = i->getClientIndex();
        if(allClIndexes.contains(ci))
            dupClIndexes << ci;
        allClIndexes << ci;
    }

    if(dupClIndexes.isEmpty())
        return; // 1st 'return'

    // Futher logic just in case of duplicates
    QMultiMap<ClientIndex, QUuid> ciToUuid;
    for(auto i=pnp_cache.begin(); i!=pnp_cache.end(); ++i) {
        const auto &ci = i->getClientIndex();
        ciToUuid.insert(ci, i.key());
    }

    QSet<QUuid> outOfDateSet;
    QUuid latestUuid;
    QElapsedTimer latestTime, now;
    now.start();
    for(auto i=ciToUuid.cbegin(); i!=ciToUuid.cend(); ++i) {
        auto pdIt = pnp_cache.constFind(*i);
        if(pdIt != pnp_cache.cend()) {
            auto t = pdIt->getDiscoverTimer();
            if(!latestTime.isValid() || latestTime < t) {
                latestTime = t;
                latestUuid = pdIt.key();
            }
            if(pdIt->isOutOfDate(now)) {
                outOfDateSet << pdIt.key();
            }
        }
    }
    for(auto uuid : outOfDateSet) {
        // Skip latest PD
        if(uuid != latestUuid) {
            pnp_cache.remove(uuid);
        }
    }
}
